@@ -18,15 +18,17 @@ import sys
 import time
 from contextlib import contextmanager
 
+from tools import cache
 from tools import shared
 from tools import system_libs
 from tools import ports
 from tools.settings import settings
-import emscripten
+from tools.system_libs import USE_NINJA
 
 
 # Minimal subset of targets used by CI systems to build enough to useful
 MINIMAL_TASKS = [
+    'libbulkmemory',
     'libcompiler_rt',
     'libcompiler_rt-wasm-sjlj',
     'libc',
@@ -36,6 +38,9 @@ MINIMAL_TASKS = [
     'libc++abi',
     'libc++abi-except',
     'libc++abi-noexcept',
+    'libc++abi-debug',
+    'libc++abi-debug-except',
+    'libc++abi-debug-noexcept',
     'libc++',
     'libc++-except',
     'libc++-noexcept',
@@ -56,8 +61,7 @@ MINIMAL_TASKS = [
     'libsockets',
     'libstubs',
     'libstubs-debug',
-    'struct_info',
-    'libstandalonewasm',
+    'libstandalonewasm-nocatch',
     'crt1',
     'crt1_proxy_main',
     'libunwind-except',
@@ -76,11 +80,17 @@ MINIMAL_PIC_TASKS = MINIMAL_TASKS + [
     'libc_optz-mt-debug',
     'libc++abi-mt',
     'libc++abi-mt-noexcept',
+    'libc++abi-debug-mt',
+    'libc++abi-debug-mt-noexcept',
     'libc++-mt',
     'libc++-mt-noexcept',
     'libdlmalloc-mt',
     'libGL-emu',
+    'libGL-emu-webgl2',
     'libGL-mt',
+    'libGL-mt-emu',
+    'libGL-mt-emu-webgl2',
+    'libGL-mt-emu-webgl2-ofb',
     'libsockets_proxy',
     'libsockets-mt',
     'crtbegin',
@@ -89,7 +99,9 @@ MINIMAL_PIC_TASKS = MINIMAL_TASKS + [
     'libwasm_workers_stub-debug',
     'libwebgpu_cpp',
     'libfetch',
+    'libfetch-mt',
     'libwasmfs',
+    'giflib',
 ]
 
 PORTS = sorted(list(ports.ports_by_name.keys()) + list(ports.port_variants.keys()))
@@ -142,11 +154,6 @@ def build_port(port_name):
 def get_system_tasks():
   system_libraries = system_libs.Library.get_all_variations()
   system_tasks = list(system_libraries.keys())
-  # This is needed to build the generated_struct_info.json file.
-  # It is not a system library, but it needs to be built before
-  # running with FROZEN_CACHE.
-  system_tasks += ['struct_info']
-
   return system_libraries, system_tasks
 
 
@@ -166,12 +173,15 @@ def main():
                       help='show build commands')
   parser.add_argument('--wasm64', action='store_true',
                       help='use wasm64 architecture')
-  parser.add_argument('operation', help='currently only "build" and "clear" are supported')
-  parser.add_argument('targets', nargs='+', help='see below')
+  parser.add_argument('operation', choices=['build', 'clear', 'rebuild'])
+  parser.add_argument('targets', nargs='*', help='see below')
   args = parser.parse_args()
 
-  if args.operation not in ('build', 'clear'):
-    shared.exit_with_error('unfamiliar operation: ' + args.operation)
+  if args.operation != 'rebuild' and len(args.targets) == 0:
+    shared.exit_with_error('no build targets specified')
+
+  if args.operation == 'rebuild' and not USE_NINJA:
+    shared.exit_with_error('"rebuild" operation is only valid when using Ninja')
 
   # process flags
 
@@ -198,31 +208,43 @@ def main():
   if args.force:
     do_clear = True
 
+  system_libraries, system_tasks = get_system_tasks()
+
   # process tasks
   auto_tasks = False
-  tasks = args.targets
-  system_libraries, system_tasks = get_system_tasks()
-  if 'SYSTEM' in tasks:
-    tasks = system_tasks
-    auto_tasks = True
-  elif 'USER' in tasks:
-    tasks = PORTS
-    auto_tasks = True
-  elif 'MINIMAL' in tasks:
-    tasks = MINIMAL_TASKS
-    auto_tasks = True
-  elif 'MINIMAL_PIC' in tasks:
-    tasks = MINIMAL_PIC_TASKS
-    auto_tasks = True
-  elif 'ALL' in tasks:
-    tasks = system_tasks + PORTS
-    auto_tasks = True
+  task_targets = dict.fromkeys(args.targets) # use dict to keep targets order
+
+  # subsitute
+  predefined_tasks = {
+    'SYSTEM': system_tasks,
+    'USER': PORTS,
+    'MINIMAL': MINIMAL_TASKS,
+    'MINIMAL_PIC': MINIMAL_PIC_TASKS,
+    'ALL': system_tasks + PORTS,
+  }
+  for name, tasks in predefined_tasks.items():
+    if name in task_targets:
+      task_targets[name] = tasks
+      auto_tasks = True
+
+  # flatten tasks
+  tasks = []
+  for name, targets in task_targets.items():
+    if targets is None:
+      # Use target name as task
+      tasks.append(name)
+    else:
+      # There are some ports that we don't want to build as part
+      # of ALL since the are not well tested or widely used:
+      if 'cocos2d' in targets:
+        targets.remove('cocos2d')
+
+      # Use targets from predefined_tasks
+      tasks.extend(targets)
+
   if auto_tasks:
-    # There are some ports that we don't want to build as part
-    # of ALL since the are not well tested or widely used:
-    skip_tasks = ['cocos2d']
-    tasks = [x for x in tasks if x not in skip_tasks]
     print('Building targets: %s' % ' '.join(tasks))
+
   for what in tasks:
     for old, new in legacy_prefixes.items():
       if what.startswith(old):
@@ -237,17 +259,15 @@ def main():
       if do_clear:
         library.erase()
       if do_build:
-        library.build()
+        if USE_NINJA:
+          library.generate()
+        else:
+          library.build(deterministic_paths=True)
     elif what == 'sysroot':
       if do_clear:
-        shared.Cache.erase_file('sysroot_install.stamp')
+        cache.erase_file('sysroot_install.stamp')
       if do_build:
         system_libs.ensure_sysroot()
-    elif what == 'struct_info':
-      if do_clear:
-        emscripten.clear_struct_info()
-      if do_build:
-        emscripten.generate_struct_info()
     elif what in PORTS:
       if do_clear:
         clear_port(what)
@@ -260,7 +280,10 @@ def main():
     time_taken = time.time() - start_time
     logger.info('...success. Took %s(%.2fs)' % (('%02d:%02d mins ' % (time_taken // 60, time_taken % 60) if time_taken >= 60 else ''), time_taken))
 
-  if len(tasks) > 1:
+  if USE_NINJA and args.operation != 'clear':
+    system_libs.build_deferred()
+
+  if len(tasks) > 1 or USE_NINJA:
     all_build_time_taken = time.time() - all_build_start_time
     logger.info('Built %d targets in %s(%.2fs)' % (len(tasks), ('%02d:%02d mins ' % (all_build_time_taken // 60, all_build_time_taken % 60) if all_build_time_taken >= 60 else ''), all_build_time_taken))
 
