@@ -18,7 +18,7 @@
 /*global simpleReadValueFromPointer, floatReadValueFromPointer, integerReadValueFromPointer, enumReadValueFromPointer, replacePublicSymbol, craftInvokerFunction, tupleRegistrations*/
 /*global finalizationRegistry, attachFinalizer, detachFinalizer, releaseClassHandle, runDestructor*/
 /*global ClassHandle, makeClassHandle, structRegistrations, whenDependentTypesAreResolved, BindingError, deletionQueue, delayFunction:true, upcastPointer*/
-/*global exposePublicSymbol, heap32VectorToArray, new_, RegisteredPointer_getPointee, RegisteredPointer_destructor, RegisteredPointer_deleteObject, char_0, char_9*/
+/*global exposePublicSymbol, heap32VectorToArray, newFunc, RegisteredPointer_getPointee, RegisteredPointer_destructor, RegisteredPointer_deleteObject, char_0, char_9*/
 /*global getInheritedInstanceCount, getLiveInheritedInstances, setDelayFunction, InternalError, runDestructors*/
 /*global requireRegisteredType, unregisterInheritedInstance, registerInheritedInstance, PureVirtualError, throwUnboundTypeError*/
 /*global assert, validateThis, downcastPointer, registeredPointers, RegisteredClass, getInheritedInstance, ClassHandle_isAliasOf, ClassHandle_clone, ClassHandle_isDeleted, ClassHandle_deleteLater*/
@@ -48,19 +48,6 @@ var LibraryEmbind = {
     Module['getLiveInheritedInstances'] = getLiveInheritedInstances;
     Module['flushPendingDeletes'] = flushPendingDeletes;
     Module['setDelayFunction'] = setDelayFunction;
-#if IN_TEST_HARNESS
-#if ASSERTIONS
-    Module['ASSERTIONS'] = true;
-#endif
-#if DYNAMIC_EXECUTION
-    // Without dynamic execution, dynamically created functions will have no
-    // names. This lets the test suite know that.
-    Module['DYNAMIC_EXECUTION'] = true;
-#endif
-#if EMBIND_STD_STRING_IS_UTF8
-    Module['EMBIND_STD_STRING_IS_UTF8'] = true;
-#endif
-#endif
   },
 
   $throwInternalError__deps: ['$InternalError'],
@@ -106,7 +93,7 @@ var LibraryEmbind = {
       proto[methodName] = function() {
         // TODO This check can be removed in -O3 level "unsafe" optimizations.
         if (!proto[methodName].overloadTable.hasOwnProperty(arguments.length)) {
-            throwBindingError("Function '" + humanName + "' called with an invalid number of arguments (" + arguments.length + ") - expects one of (" + proto[methodName].overloadTable + ")!");
+            throwBindingError(`Function '${humanName}' called with an invalid number of arguments (${arguments.length}) - expects one of (${proto[methodName].overloadTable})!`);
         }
         return proto[methodName].overloadTable[arguments.length].apply(this, arguments);
       };
@@ -131,14 +118,14 @@ var LibraryEmbind = {
   $exposePublicSymbol: function(name, value, numArguments) {
     if (Module.hasOwnProperty(name)) {
       if (undefined === numArguments || (undefined !== Module[name].overloadTable && undefined !== Module[name].overloadTable[numArguments])) {
-        throwBindingError("Cannot register public name '" + name + "' twice");
+        throwBindingError(`Cannot register public name '${name}' twice`);
       }
 
       // We are exposing a function with the same name as an existing function. Create an overload table and a function selector
       // that routes between the two.
       ensureOverloadTable(Module, name, name);
       if (Module.hasOwnProperty(numArguments)) {
-          throwBindingError("Cannot register multiple overloads of a function with the same number of arguments (" + numArguments + ")!");
+        throwBindingError(`Cannot register multiple overloads of a function with the same number of arguments (${numArguments})!`);
       }
       // Add the new function into the overload table.
       Module[name].overloadTable[numArguments] = value;
@@ -194,25 +181,16 @@ var LibraryEmbind = {
   },
 
 
-  // from https://github.com/imvu/imvujs/blob/master/src/function.js
   $createNamedFunction__deps: ['$makeLegalFunctionName'],
   $createNamedFunction: function(name, body) {
     name = makeLegalFunctionName(name);
-#if DYNAMIC_EXECUTION == 0
-    return function() {
-      "use strict";
-      return body.apply(this, arguments);
-    };
-#else
-    /*jshint evil:true*/
-    return new Function(
-        "body",
-        "return function " + name + "() {\n" +
-        "    \"use strict\";" +
-        "    return body.apply(this, arguments);\n" +
-        "};\n"
-    )(body);
-#endif
+    // Use an abject with a computed property name to create a new function with
+    // a name specified at runtime, but without using `new Function` or `eval`.
+    return {
+      [name]: function() {
+        return body.apply(this, arguments);
+      }
+    }[name];
   },
 
   $embindRepr: function(v) {
@@ -304,19 +282,19 @@ var LibraryEmbind = {
   $registerType__docs: '/** @param {Object=} options */',
   $registerType: function(rawType, registeredInstance, options = {}) {
     if (!('argPackAdvance' in registeredInstance)) {
-        throw new TypeError('registerType registeredInstance requires argPackAdvance');
+      throw new TypeError('registerType registeredInstance requires argPackAdvance');
     }
 
     var name = registeredInstance.name;
     if (!rawType) {
-        throwBindingError('type "' + name + '" must have a positive integer typeid pointer');
+      throwBindingError(`type "${name}" must have a positive integer typeid pointer`);
     }
     if (registeredTypes.hasOwnProperty(rawType)) {
-        if (options.ignoreDuplicateRegistrations) {
-            return;
-        } else {
-            throwBindingError("Cannot register type '" + name + "' twice");
-        }
+      if (options.ignoreDuplicateRegistrations) {
+        return;
+      } else {
+        throwBindingError(`Cannot register type '${name}' twice`);
+      }
     }
 
     registeredTypes[rawType] = registeredInstance;
@@ -393,7 +371,7 @@ var LibraryEmbind = {
     return ret;
   },
 
-  $getTypeName__deps: ['$readLatin1String', '__getTypeName'],
+  $getTypeName__deps: ['$readLatin1String', '__getTypeName', 'free'],
   $getTypeName: function(type) {
     var ptr = ___getTypeName(type);
     var rv = readLatin1String(ptr);
@@ -421,7 +399,6 @@ var LibraryEmbind = {
     return impl;
   },
 
-  _embind_register_void__sig: 'vpp',
   _embind_register_void__deps: ['$readLatin1String', '$registerType'],
   _embind_register_void: function(rawType, name) {
     name = readLatin1String(name);
@@ -439,7 +416,6 @@ var LibraryEmbind = {
     });
   },
 
-  _embind_register_bool__sig: 'vpppii',
   _embind_register_bool__deps: [
     '$getShiftFromSize', '$readLatin1String', '$registerType'],
   _embind_register_bool: function(rawType, name, size, trueValue, falseValue) {
@@ -546,7 +522,6 @@ var LibraryEmbind = {
 
   // When converting a number from JS to C++ side, the valid range of the number is
   // [minRange, maxRange], inclusive.
-  _embind_register_integer__sig: 'vpppii',
   _embind_register_integer__deps: [
     '$embindRepr', '$getShiftFromSize', '$integerReadValueFromPointer',
     '$readLatin1String', '$registerType'],
@@ -571,10 +546,10 @@ var LibraryEmbind = {
     var checkAssertions = (value, toTypeName) => {
 #if ASSERTIONS
       if (typeof value != "number" && typeof value != "boolean") {
-        throw new TypeError('Cannot convert "' + embindRepr(value) + '" to ' + toTypeName);
+        throw new TypeError(`Cannot convert "${embindRepr(value)}" to ${toTypeName}`);
       }
       if (value < minRange || value > maxRange) {
-        throw new TypeError('Passing a number "' + embindRepr(value) + '" from JS side to C/C++ side to an argument of type "' + name + '", which is outside the valid range [' + minRange + ', ' + maxRange + ']!');
+        throw new TypeError(`Passing a number "${embindRepr(value)}" from JS side to C/C++ side to an argument of type "${name}", which is outside the valid range [${minRange}, ${maxRange}]!`);
       }
 #endif
     }
@@ -603,7 +578,6 @@ var LibraryEmbind = {
   },
 
 #if WASM_BIGINT
-  _embind_register_bigint__sig: 'vpppjj',
   _embind_register_bigint__deps: [
     '$embindRepr', '$readLatin1String', '$registerType', '$integerReadValueFromPointer'],
   _embind_register_bigint: function(primitiveType, name, size, minRange, maxRange) {
@@ -615,7 +589,6 @@ var LibraryEmbind = {
 
     // maxRange comes through as -1 for uint64_t (see issue 13902). Work around that temporarily
     if (isUnsignedType) {
-      // Use string because acorn does recognize bigint literals
       maxRange = (1n << 64n) - 1n;
     }
 
@@ -626,10 +599,10 @@ var LibraryEmbind = {
       },
       'toWireType': function (destructors, value) {
         if (typeof value != "bigint" && typeof value != "number") {
-          throw new TypeError('Cannot convert "' + embindRepr(value) + '" to ' + this.name);
+          throw new TypeError(`Cannot convert "${embindRepr(value)}" to ${this.name}`);
         }
         if (value < minRange || value > maxRange) {
-          throw new TypeError('Passing a number "' + embindRepr(value) + '" from JS side to C/C++ side to an argument of type "' + name + '", which is outside the valid range [' + minRange + ', ' + maxRange + ']!');
+          throw new TypeError(`Passing a number "${embindRepr(value)}" from JS side to C/C++ side to an argument of type "${name}", which is outside the valid range [${minRange}, ${maxRange}]!`);
         }
         return value;
       },
@@ -643,7 +616,6 @@ var LibraryEmbind = {
   _embind_register_bigint: function(primitiveType, name, size, minRange, maxRange) {},
 #endif
 
-  _embind_register_float__sig: 'vppp',
   _embind_register_float__deps: [
     '$embindRepr', '$floatReadValueFromPointer', '$getShiftFromSize',
     '$readLatin1String', '$registerType'],
@@ -658,7 +630,7 @@ var LibraryEmbind = {
       'toWireType': function(destructors, value) {
 #if ASSERTIONS
         if (typeof value != "number" && typeof value != "boolean") {
-          throw new TypeError('Cannot convert "' + embindRepr(value) + '" to ' + this.name);
+          throw new TypeError(`Cannot convert ${embindRepr(value)} to ${this.name}`);
         }
 #endif
         // The VM will perform JS to Wasm value conversion, according to the spec:
@@ -676,10 +648,10 @@ var LibraryEmbind = {
     return this['fromWireType']({{{ makeGetValue('pointer', '0', 'i32') }}});
   },
 
-  _embind_register_std_string__sig: 'vpp',
   _embind_register_std_string__deps: [
     '$readLatin1String', '$registerType',
-    '$simpleReadValueFromPointer', '$throwBindingError'],
+    '$simpleReadValueFromPointer', '$throwBindingError',
+    '$stringToUTF8', '$lengthBytesUTF8', 'malloc', 'free'],
   _embind_register_std_string: function(rawType, name) {
     name = readLatin1String(name);
     var stdStringIsUTF8
@@ -780,7 +752,6 @@ var LibraryEmbind = {
     });
   },
 
-  _embind_register_std_wstring__sig: 'vppp',
   _embind_register_std_wstring__deps: [
     '$readLatin1String', '$registerType', '$simpleReadValueFromPointer',
     '$UTF16ToString', '$stringToUTF16', '$lengthBytesUTF16',
@@ -857,7 +828,6 @@ var LibraryEmbind = {
     });
   },
 
-  _embind_register_emval__sig: 'vpp',
   _embind_register_emval__deps: [
     '_emval_decref', '$Emval',
     '$readLatin1String', '$registerType', '$simpleReadValueFromPointer'],
@@ -882,7 +852,6 @@ var LibraryEmbind = {
     });
   },
 
-  _embind_register_memory_view__sig: 'vpip',
   _embind_register_memory_view__deps: ['$readLatin1String', '$registerType'],
   _embind_register_memory_view: function(rawType, dataTypeIndex, name) {
     var typeMapping = [
@@ -907,7 +876,7 @@ var LibraryEmbind = {
       var heap = HEAPU32;
       var size = heap[handle]; // in elements
       var data = heap[handle + 1]; // byte offset into emscripten heap
-      return new TA(buffer, data, size);
+      return new TA(heap.buffer, data, size);
     }
 
     name = readLatin1String(name);
@@ -929,20 +898,12 @@ var LibraryEmbind = {
     }
   },
 
-  // Function implementation of operator new, per
-  // http://www.ecma-international.org/publications/files/ECMA-ST/Ecma-262.pdf
-  // 13.2.2
-  // ES3
-  $new___deps: ['$createNamedFunction'],
-  $new_: function(constructor, argumentList) {
+#if DYNAMIC_EXECUTION
+  $newFunc__deps: ['$createNamedFunction'],
+  $newFunc: function(constructor, argumentList) {
     if (!(constructor instanceof Function)) {
-      throw new TypeError('new_ called with constructor type ' + typeof(constructor) + " which is not a function");
+      throw new TypeError(`new_ called with constructor type ${typeof(constructor)} which is not a function`);
     }
-#if DYNAMIC_EXECUTION == 0
-    if (constructor === Function) {
-      throw new Error('new_ cannot create a new Function with DYNAMIC_EXECUTION == 0.');
-    }
-#endif
     /*
      * Previously, the following line was just:
      *   function dummy() {};
@@ -960,17 +921,21 @@ var LibraryEmbind = {
     var r = constructor.apply(obj, argumentList);
     return (r instanceof Object) ? r : obj;
   },
+#endif
 
   // The path to interop from JS code to C++ code:
   // (hand-written JS code) -> (autogenerated JS invoker) -> (template-generated C++ invoker) -> (target C++ function)
   // craftInvokerFunction generates the JS invoker function for each function exposed to JS through embind.
   $craftInvokerFunction__deps: [
-    '$makeLegalFunctionName', '$new_', '$runDestructors', '$throwBindingError',
-  #if ASYNCIFY
+    '$makeLegalFunctionName', '$runDestructors', '$throwBindingError',
+#if DYNAMIC_EXECUTION
+    '$newFunc',
+#endif
+#if ASYNCIFY
     '$Asyncify',
-  #endif
+#endif
   ],
-  $craftInvokerFunction: function(humanName, argTypes, classType, cppInvokerFunc, cppTargetFunc) {
+  $craftInvokerFunction: function(humanName, argTypes, classType, cppInvokerFunc, cppTargetFunc, /** boolean= */ isAsync) {
     // humanName: a human-readable string name for the function to be generated.
     // argTypes: An array that contains the embind type objects for all types in the function signature.
     //    argTypes[0] is the type object for the function return value.
@@ -979,12 +944,17 @@ var LibraryEmbind = {
     // classType: The embind type object for the class to be bound, or null if this is not a method of a class.
     // cppInvokerFunc: JS Function object to the C++-side function that interops into C++ code.
     // cppTargetFunc: Function pointer (an integer to FUNCTION_TABLE) to the target C++ function the cppInvokerFunc will end up calling.
+    // isAsync: Optional. If true, returns an async function. Async bindings are only supported with JSPI.
     var argCount = argTypes.length;
 
     if (argCount < 2) {
       throwBindingError("argTypes array size mismatch! Must at least get return value and 'this' types!");
     }
 
+#if ASSERTIONS && ASYNCIFY != 2
+    assert(!isAsync, 'Async bindings are only supported with JSPI.');
+#endif
+
     var isClassMethodFunc = (argTypes[1] !== null && classType !== null);
 
     // Free functions with signature "void function()" do not need an invoker that marshalls between wire types.
@@ -1007,6 +977,12 @@ var LibraryEmbind = {
 
     var returns = (argTypes[0].name !== "void");
 
+#if ASYNCIFY
+    if (isAsync) {
+      cppInvokerFunc = Asyncify.makeAsyncFunction(cppInvokerFunc);
+    }
+#endif
+
 #if DYNAMIC_EXECUTION == 0
     var expectedArgCount = argCount - 2;
     var argsWired = new Array(expectedArgCount);
@@ -1014,9 +990,7 @@ var LibraryEmbind = {
     var destructors = [];
     return function() {
       if (arguments.length !== expectedArgCount) {
-        throwBindingError('function ' + humanName + ' called with ' +
-          arguments.length + ' arguments, expected ' + expectedArgCount +
-          ' args!');
+        throwBindingError(`function ${humanName} called with ${arguments.length} arguments, expected ${expectedArgCount} args!`);
       }
 #if EMSCRIPTEN_TRACING
       Module.emscripten_trace_enter_context('embind::' + humanName);
@@ -1057,10 +1031,14 @@ var LibraryEmbind = {
         }
       }
 
-#if ASYNCIFY
+#if ASYNCIFY == 1
       if (Asyncify.currData) {
         return Asyncify.whenDone().then(onDone);
       }
+#elif ASYNCIFY == 2
+      if (isAsync) {
+        return rv.then(onDone);
+      }
 #endif
 
       return onDone(rv);
@@ -1080,7 +1058,7 @@ var LibraryEmbind = {
         "}\n";
 
 #if EMSCRIPTEN_TRACING
-    invokerFnBody += "Module.emscripten_trace_enter_context('embind::" + humanName + "');\n";
+    invokerFnBody += `Module.emscripten_trace_enter_context('embind::${humanName}');\n`;
 #endif
 
     if (needsDestructorStack) {
@@ -1111,11 +1089,13 @@ var LibraryEmbind = {
     }
 
     invokerFnBody +=
-        (returns?"var rv = ":"") + "invoker(fn"+(argsListWired.length>0?", ":"")+argsListWired+");\n";
+        (returns || isAsync ? "var rv = ":"") + "invoker(fn"+(argsListWired.length>0?", ":"")+argsListWired+");\n";
 
-#if ASYNCIFY
+#if ASYNCIFY == 1
     args1.push("Asyncify");
     args2.push(Asyncify);
+#endif
+#if ASYNCIFY
     invokerFnBody += "function onDone(" + (returns ? "rv" : "") + ") {\n";
 #endif
 
@@ -1144,17 +1124,19 @@ var LibraryEmbind = {
 #endif
     }
 
-#if ASYNCIFY
+#if ASYNCIFY == 1
     invokerFnBody += "}\n";
     invokerFnBody += "return Asyncify.currData ? Asyncify.whenDone().then(onDone) : onDone(" + (returns ? "rv" : "") +");\n"
+#elif ASYNCIFY == 2
+    invokerFnBody += "}\n";
+    invokerFnBody += "return " + (isAsync ? "rv.then(onDone)" : "onDone(rv)") + ";";
 #endif
 
     invokerFnBody += "}\n";
 
     args1.push(invokerFnBody);
 
-    var invokerFunction = new_(Function, args1).apply(null, args2);
-    return invokerFunction;
+    return newFunc(Function, args1).apply(null, args2);
 #endif
   },
 
@@ -1185,36 +1167,34 @@ var LibraryEmbind = {
 
     var fp = makeDynCaller();
     if (typeof fp != "function") {
-        throwBindingError("unknown function pointer with signature " + signature + ": " + rawFunction);
+        throwBindingError(`unknown function pointer with signature ${signature}: ${rawFunction}`);
     }
     return fp;
   },
 
-  _embind_register_function__sig: 'vpipppp',
   _embind_register_function__deps: [
     '$craftInvokerFunction', '$exposePublicSymbol', '$heap32VectorToArray',
     '$readLatin1String', '$replacePublicSymbol', '$embind__requireFunction',
     '$throwUnboundTypeError', '$whenDependentTypesAreResolved'],
-  _embind_register_function: function(name, argCount, rawArgTypesAddr, signature, rawInvoker, fn) {
+  _embind_register_function: function(name, argCount, rawArgTypesAddr, signature, rawInvoker, fn, isAsync) {
     var argTypes = heap32VectorToArray(argCount, rawArgTypesAddr);
     name = readLatin1String(name);
 
     rawInvoker = embind__requireFunction(signature, rawInvoker);
 
     exposePublicSymbol(name, function() {
-      throwUnboundTypeError('Cannot call ' + name + ' due to unbound types', argTypes);
+      throwUnboundTypeError(`Cannot call ${name} due to unbound types`, argTypes);
     }, argCount - 1);
 
     whenDependentTypesAreResolved([], argTypes, function(argTypes) {
       var invokerArgsArray = [argTypes[0] /* return value */, null /* no class 'this'*/].concat(argTypes.slice(1) /* actual params */);
-      replacePublicSymbol(name, craftInvokerFunction(name, invokerArgsArray, null /* no class 'this'*/, rawInvoker, fn), argCount - 1);
+      replacePublicSymbol(name, craftInvokerFunction(name, invokerArgsArray, null /* no class 'this'*/, rawInvoker, fn, isAsync), argCount - 1);
       return [];
     });
   },
 
   $tupleRegistrations: {},
 
-  _embind_register_value_array__sig: 'vpppppp',
   _embind_register_value_array__deps: [
     '$tupleRegistrations', '$readLatin1String', '$embind__requireFunction'],
   _embind_register_value_array: function(
@@ -1233,7 +1213,6 @@ var LibraryEmbind = {
     };
   },
 
-  _embind_register_value_array_element__sig: 'vppppppppp',
   _embind_register_value_array_element__deps: [
     '$tupleRegistrations', '$embind__requireFunction'],
   _embind_register_value_array_element: function(
@@ -1257,7 +1236,6 @@ var LibraryEmbind = {
     });
   },
 
-  _embind_finalize_value_array__sig: 'vp',
   _embind_finalize_value_array__deps: [
     '$tupleRegistrations', '$runDestructors',
     '$simpleReadValueFromPointer', '$whenDependentTypesAreResolved'],
@@ -1302,7 +1280,7 @@ var LibraryEmbind = {
         },
         'toWireType': function(destructors, o) {
           if (elementsLength !== o.length) {
-            throw new TypeError("Incorrect number of tuple elements for " + reg.name + ": expected=" + elementsLength + ", actual=" + o.length);
+            throw new TypeError(`Incorrect number of tuple elements for ${reg.name}: expected=${elementsLength}, actual=${o.length}`);
           }
           var ptr = rawConstructor();
           for (var i = 0; i < elementsLength; ++i) {
@@ -1322,7 +1300,6 @@ var LibraryEmbind = {
 
   $structRegistrations: {},
 
-  _embind_register_value_object__sig: 'viiiiii',
   _embind_register_value_object__deps: [
     '$structRegistrations', '$readLatin1String', '$embind__requireFunction'],
   _embind_register_value_object: function(
@@ -1341,7 +1318,6 @@ var LibraryEmbind = {
     };
   },
 
-  _embind_register_value_object_field__sig: 'viiiiiiiiii',
   _embind_register_value_object_field__deps: [
     '$structRegistrations', '$readLatin1String', '$embind__requireFunction'],
   _embind_register_value_object_field: function(
@@ -1367,7 +1343,6 @@ var LibraryEmbind = {
     });
   },
 
-  _embind_finalize_value_object__sig: 'ii',
   _embind_finalize_value_object__deps: [
     '$structRegistrations', '$runDestructors',
     '$simpleReadValueFromPointer', '$whenDependentTypesAreResolved'],
@@ -1418,7 +1393,7 @@ var LibraryEmbind = {
           // assume all fields are present without checking.
           for (var fieldName in fields) {
             if (!(fieldName in o)) {
-              throw new TypeError('Missing field:  "' + fieldName + '"');
+              throw new TypeError(`Missing field: "${fieldName}"`);
             }
           }
           var ptr = rawConstructor();
@@ -1457,13 +1432,13 @@ var LibraryEmbind = {
     }
 
     if (!handle.$$) {
-      throwBindingError('Cannot pass "' + embindRepr(handle) + '" as a ' + this.name);
+      throwBindingError(`Cannot pass "${embindRepr(handle)}" as a ${this.name}`);
     }
     if (!handle.$$.ptr) {
-      throwBindingError('Cannot pass deleted object as a pointer of type ' + this.name);
+      throwBindingError(`Cannot pass deleted object as a pointer of type ${this.name}`);
     }
     if (!this.isConst && handle.$$.ptrType.isConst) {
-      throwBindingError('Cannot convert argument of type ' + (handle.$$.smartPtrType ? handle.$$.smartPtrType.name : handle.$$.ptrType.name) + ' to parameter type ' + this.name);
+      throwBindingError(`Cannot convert argument of type ${(handle.$$.smartPtrType ? handle.$$.smartPtrType.name : handle.$$.ptrType.name)} to parameter type ${this.name}`);
     }
     var handleClass = handle.$$.ptrType.registeredClass;
     ptr = upcastPointer(handle.$$.ptr, handleClass, this.registeredClass);
@@ -1482,7 +1457,7 @@ var LibraryEmbind = {
           if (handle.$$.smartPtrType === this) {
             ptr = handle.$$.smartPtr;
           } else {
-            throwBindingError('Cannot convert argument of type ' + (handle.$$.smartPtrType ? handle.$$.smartPtrType.name : handle.$$.ptrType.name) + ' to parameter type ' + this.name);
+            throwBindingError(`Cannot convert argument of type ${(handle.$$.smartPtrType ? handle.$$.smartPtrType.name : handle.$$.ptrType.name)} to parameter type ${this.name}`);
           }
           break;
 
@@ -1526,7 +1501,7 @@ var LibraryEmbind = {
     }
 
     if (!handle.$$) {
-      throwBindingError('Cannot pass "' + embindRepr(handle) + '" as a ' + this.name);
+      throwBindingError(`Cannot pass "${embindRepr(handle)}" as a ${this.name}`);
     }
     if (!handle.$$.ptr) {
       throwBindingError('Cannot pass deleted object as a pointer of type ' + this.name);
@@ -1548,13 +1523,13 @@ var LibraryEmbind = {
     }
 
     if (!handle.$$) {
-      throwBindingError('Cannot pass "' + embindRepr(handle) + '" as a ' + this.name);
+      throwBindingError(`Cannot pass "${embindRepr(handle)}" as a ${this.name}`);
     }
     if (!handle.$$.ptr) {
-      throwBindingError('Cannot pass deleted object as a pointer of type ' + this.name);
+      throwBindingError(`Cannot pass deleted object as a pointer of type ${this.name}`);
     }
     if (handle.$$.ptrType.isConst) {
-        throwBindingError('Cannot convert argument of type ' + handle.$$.ptrType.name + ' to parameter type ' + this.name);
+        throwBindingError(`Cannot convert argument of type ${handle.$$.ptrType.name} to parameter type ${this.name}`);
     }
     var handleClass = handle.$$.ptrType.registeredClass;
     var ptr = upcastPointer(handle.$$.ptr, handleClass, this.registeredClass);
@@ -1663,7 +1638,7 @@ var LibraryEmbind = {
 #if MEMORY64
     ptr = bigintToI53Checked(ptr);
 #if ASSERTIONS
-    assert(ptr != NaN);
+    assert(Number.isSafeInteger(ptr));
 #endif
 #endif
 
@@ -1791,7 +1766,7 @@ var LibraryEmbind = {
         // This is more useful than the empty stacktrace of `FinalizationRegistry`
         // callback.
         var cls = $$.ptrType.registeredClass;
-        info.leakWarning = new Error("Embind found a leaked C++ instance " + cls.name + " <" + ptrToString($$.ptr) + ">.\n" +
+        info.leakWarning = new Error(`Embind found a leaked C++ instance ${cls.name} <${ptrToString($$.ptr)}>.\n` +
         "We'll free it automatically in this case, but this functionality is not reliable across various environments.\n" +
         "Make sure to invoke .delete() manually once you're done with the instance instead.\n" +
         "Originally allocated"); // `.stack` will add "at ..." after this sentence
@@ -1992,7 +1967,6 @@ var LibraryEmbind = {
     };
   },
 
-  _embind_register_class__sig: 'vppppppppppppp',
   _embind_register_class__deps: [
     '$BindingError', '$ClassHandle', '$createNamedFunction',
     '$registeredPointers', '$exposePublicSymbol',
@@ -2026,7 +2000,7 @@ var LibraryEmbind = {
 
     exposePublicSymbol(legalFunctionName, function() {
       // this code cannot run if baseClassRawType is zero
-      throwUnboundTypeError('Cannot construct ' + name + ' due to unbound types', [baseClassRawType]);
+      throwUnboundTypeError(`Cannot construct ${name} due to unbound types`, [baseClassRawType]);
     });
 
     whenDependentTypesAreResolved(
@@ -2053,7 +2027,7 @@ var LibraryEmbind = {
           }
           var body = registeredClass.constructor_body[arguments.length];
           if (undefined === body) {
-            throw new BindingError("Tried to invoke ctor of " + name + " with invalid number of parameters (" + arguments.length + ") - expected (" + Object.keys(registeredClass.constructor_body).toString() + ") parameters instead!");
+            throw new BindingError(`Tried to invoke ctor of ${name} with invalid number of parameters (${arguments.length}) - expected (${Object.keys(registeredClass.constructor_body).toString()}) parameters instead!`);
           }
           return body.apply(this, arguments);
         });
@@ -2073,6 +2047,15 @@ var LibraryEmbind = {
                                                   upcast,
                                                   downcast);
 
+        if (registeredClass.baseClass) {
+          // Keep track of class hierarchy. Used to allow sub-classes to inherit class functions.
+          if (registeredClass.baseClass.__derivedClasses === undefined) {
+            registeredClass.baseClass.__derivedClasses = [];
+          }
+
+          registeredClass.baseClass.__derivedClasses.push(registeredClass);
+        }
+
         var referenceConverter = new RegisteredPointer(name,
                                                        registeredClass,
                                                        true,
@@ -2103,7 +2086,6 @@ var LibraryEmbind = {
     );
   },
 
-  _embind_register_class_constructor__sig: 'vpipppp',
   _embind_register_class_constructor__deps: [
     '$heap32VectorToArray', '$embind__requireFunction', '$runDestructors',
     '$throwBindingError', '$whenDependentTypesAreResolved', '$registeredTypes',
@@ -2130,10 +2112,10 @@ var LibraryEmbind = {
         classType.registeredClass.constructor_body = [];
       }
       if (undefined !== classType.registeredClass.constructor_body[argCount - 1]) {
-        throw new BindingError("Cannot register multiple constructors with identical number of parameters (" + (argCount-1) + ") for class '" + classType.name + "'! Overload resolution is currently only performed using the parameter count, not actual type info!");
+        throw new BindingError(`Cannot register multiple constructors with identical number of parameters (${argCount-1}) for class '${classType.name}'! Overload resolution is currently only performed using the parameter count, not actual type info!`);
       }
       classType.registeredClass.constructor_body[argCount - 1] = () => {
-        throwUnboundTypeError('Cannot construct ' + classType.name + ' due to unbound types', rawArgTypes);
+        throwUnboundTypeError(`Cannot construct ${classType.name} due to unbound types`, rawArgTypes);
       };
 
       whenDependentTypesAreResolved([], rawArgTypes, function(argTypes) {
@@ -2165,7 +2147,7 @@ var LibraryEmbind = {
   $upcastPointer: function(ptr, ptrClass, desiredClass) {
     while (ptrClass !== desiredClass) {
       if (!ptrClass.upcast) {
-        throwBindingError("Expected null or instance of " + desiredClass.name + ", got an instance of " + ptrClass.name);
+        throwBindingError(`Expected null or instance of ${desiredClass.name}, got an instance of ${ptrClass.name}`);
       }
       ptr = ptrClass.upcast(ptr);
       ptrClass = ptrClass.baseClass;
@@ -2182,7 +2164,7 @@ var LibraryEmbind = {
       throwBindingError(humanName + ' incompatible with "this" of type ' + this_.constructor.name);
     }
     if (!this_.$$.ptr) {
-      throwBindingError('cannot call emscripten binding method ' + humanName + ' on deleted object');
+      throwBindingError(`cannot call emscripten binding method ${humanName} on deleted object`);
     }
 
     // todo: kill this
@@ -2191,7 +2173,6 @@ var LibraryEmbind = {
                          classType.registeredClass);
   },
 
-  _embind_register_class_function__sig: 'vppippppi',
   _embind_register_class_function__deps: [
     '$craftInvokerFunction', '$heap32VectorToArray', '$readLatin1String',
     '$embind__requireFunction', '$throwUnboundTypeError',
@@ -2203,7 +2184,8 @@ var LibraryEmbind = {
                                             invokerSignature,
                                             rawInvoker,
                                             context,
-                                            isPureVirtual) {
+                                            isPureVirtual,
+                                            isAsync) {
     var rawArgTypes = heap32VectorToArray(argCount, rawArgTypesAddr);
     methodName = readLatin1String(methodName);
     rawInvoker = embind__requireFunction(invokerSignature, rawInvoker);
@@ -2221,7 +2203,7 @@ var LibraryEmbind = {
       }
 
       function unboundTypesHandler() {
-        throwUnboundTypeError('Cannot call ' + humanName + ' due to unbound types', rawArgTypes);
+        throwUnboundTypeError(`Cannot call ${humanName} due to unbound types`, rawArgTypes);
       }
 
       var proto = classType.registeredClass.instancePrototype;
@@ -2240,7 +2222,7 @@ var LibraryEmbind = {
       }
 
       whenDependentTypesAreResolved([], rawArgTypes, function(argTypes) {
-        var memberFunction = craftInvokerFunction(humanName, argTypes, classType, rawInvoker, context);
+        var memberFunction = craftInvokerFunction(humanName, argTypes, classType, rawInvoker, context, isAsync);
 
         // Replace the initial unbound-handler-stub function with the appropriate member function, now that all types
         // are resolved. If multiple overloads are registered for this function, the function goes into an overload table.
@@ -2258,7 +2240,6 @@ var LibraryEmbind = {
     });
   },
 
-  _embind_register_class_property__sig: 'vpppppppppp',
   _embind_register_class_property__deps: [
     '$readLatin1String', '$embind__requireFunction', '$runDestructors',
     '$throwBindingError', '$throwUnboundTypeError',
@@ -2281,14 +2262,14 @@ var LibraryEmbind = {
       var humanName = classType.name + '.' + fieldName;
       var desc = {
         get: function() {
-          throwUnboundTypeError('Cannot access ' + humanName + ' due to unbound types', [getterReturnType, setterArgumentType]);
+          throwUnboundTypeError(`Cannot access ${humanName} due to unbound types`, [getterReturnType, setterArgumentType]);
         },
         enumerable: true,
         configurable: true
       };
       if (setter) {
         desc.set = () => {
-          throwUnboundTypeError('Cannot access ' + humanName + ' due to unbound types', [getterReturnType, setterArgumentType]);
+          throwUnboundTypeError(`Cannot access ${humanName} due to unbound types`, [getterReturnType, setterArgumentType]);
         };
       } else {
         desc.set = (v) => {
@@ -2330,7 +2311,6 @@ var LibraryEmbind = {
     });
   },
 
-  _embind_register_class_class_function__sig: 'vppipppp',
   _embind_register_class_class_function__deps: [
     '$craftInvokerFunction', '$ensureOverloadTable', '$heap32VectorToArray',
     '$readLatin1String', '$embind__requireFunction', '$throwUnboundTypeError',
@@ -2341,7 +2321,8 @@ var LibraryEmbind = {
                                                   rawArgTypesAddr,
                                                   invokerSignature,
                                                   rawInvoker,
-                                                  fn) {
+                                                  fn,
+                                                  isAsync) {
     var rawArgTypes = heap32VectorToArray(argCount, rawArgTypesAddr);
     methodName = readLatin1String(methodName);
     rawInvoker = embind__requireFunction(invokerSignature, rawInvoker);
@@ -2350,7 +2331,7 @@ var LibraryEmbind = {
       var humanName = classType.name + '.' + methodName;
 
       function unboundTypesHandler() {
-        throwUnboundTypeError('Cannot call ' + humanName + ' due to unbound types', rawArgTypes);
+        throwUnboundTypeError(`Cannot call ${humanName} due to unbound types`, rawArgTypes);
       }
 
       if (methodName.startsWith("@@")) {
@@ -2374,20 +2355,29 @@ var LibraryEmbind = {
         // function. If multiple overloads are registered, the function handlers
         // go into an overload table.
         var invokerArgsArray = [argTypes[0] /* return value */, null /* no class 'this'*/].concat(argTypes.slice(1) /* actual params */);
-        var func = craftInvokerFunction(humanName, invokerArgsArray, null /* no class 'this'*/, rawInvoker, fn);
+        var func = craftInvokerFunction(humanName, invokerArgsArray, null /* no class 'this'*/, rawInvoker, fn, isAsync);
         if (undefined === proto[methodName].overloadTable) {
           func.argCount = argCount-1;
           proto[methodName] = func;
         } else {
           proto[methodName].overloadTable[argCount-1] = func;
         }
+
+        if (classType.registeredClass.__derivedClasses) {
+          for (const derivedClass of classType.registeredClass.__derivedClasses) {
+            if (!derivedClass.constructor.hasOwnProperty(methodName)) {
+              // TODO: Add support for overloads
+              derivedClass.constructor[methodName] = func;
+            }
+          }
+        }
+
         return [];
       });
       return [];
     });
   },
 
-  _embind_register_class_class_property__sig: 'vpppppppp',
   _embind_register_class_class_property__deps: [
     '$readLatin1String', '$embind__requireFunction', '$runDestructors',
     '$throwBindingError', '$throwUnboundTypeError',
@@ -2408,18 +2398,18 @@ var LibraryEmbind = {
       var humanName = classType.name + '.' + fieldName;
       var desc = {
         get: function() {
-          throwUnboundTypeError('Cannot access ' + humanName + ' due to unbound types', [rawFieldType]);
+          throwUnboundTypeError(`Cannot access ${humanName} due to unbound types`, [rawFieldType]);
         },
         enumerable: true,
         configurable: true
       };
       if (setter) {
         desc.set = () => {
-          throwUnboundTypeError('Cannot access ' + humanName + ' due to unbound types', [rawFieldType]);
+          throwUnboundTypeError(`Cannot access ${humanName} due to unbound types`, [rawFieldType]);
         };
       } else {
         desc.set = (v) => {
-          throwBindingError(humanName + ' is a read-only property');
+          throwBindingError(`${humanName} is a read-only property`);
         };
       }
 
@@ -2451,7 +2441,6 @@ var LibraryEmbind = {
     });
   },
 
-  _embind_create_inheriting_constructor__sig: 'pppp',
   _embind_create_inheriting_constructor__deps: [
     '$createNamedFunction', '$Emval',
     '$PureVirtualError', '$readLatin1String',
@@ -2473,7 +2462,7 @@ var LibraryEmbind = {
     var ctor = createNamedFunction(constructorName, function() {
       registeredClass.baseClass.pureVirtualFunctions.forEach(function(name) {
           if (this[name] === baseClassPrototype[name]) {
-              throw new PureVirtualError('Pure virtual function ' + name + ' must be implemented in JavaScript');
+              throw new PureVirtualError(`Pure virtual function ${name} must be implemented in JavaScript`);
           }
       }.bind(this));
 
@@ -2535,7 +2524,6 @@ var LibraryEmbind = {
     return name;
   },
 
-  _embind_register_smart_ptr__sig: 'vpppipppppppp',
   _embind_register_smart_ptr__deps: ['$RegisteredPointer', '$embind__requireFunction', '$whenDependentTypesAreResolved'],
   _embind_register_smart_ptr: function(rawType,
                                        rawPointeeType,
@@ -2574,7 +2562,6 @@ var LibraryEmbind = {
     });
   },
 
-  _embind_register_enum__sig: 'vpppi',
   _embind_register_enum__deps: ['$exposePublicSymbol', '$getShiftFromSize', '$enumReadValueFromPointer',
     '$readLatin1String', '$registerType'],
   _embind_register_enum: function(rawType, name, size, isSigned) {
@@ -2600,7 +2587,6 @@ var LibraryEmbind = {
     exposePublicSymbol(name, ctor);
   },
 
-  _embind_register_enum_value__sig: 'vppi',
   _embind_register_enum_value__deps: ['$createNamedFunction', '$readLatin1String', '$requireRegisteredType'],
   _embind_register_enum_value: function(rawEnumType, name, enumValue) {
     var enumType = requireRegisteredType(rawEnumType, 'enum');
@@ -2616,7 +2602,6 @@ var LibraryEmbind = {
     Enum[name] = Value;
   },
 
-  _embind_register_constant__sig: 'vppd',
   _embind_register_constant__deps: ['$readLatin1String', '$whenDependentTypesAreResolved'],
   _embind_register_constant: function(name, type, value) {
     name = readLatin1String(name);
