@@ -22,87 +22,50 @@ mergeInto(LibraryManager.library, {
 #if ASYNCIFY
   $Asyncify__deps: ['$runAndAbortIfError', '$callUserCallback', '$sigToWasmTypes',
 #if !MINIMAL_RUNTIME
-    '$runtimeKeepalivePush', '$runtimeKeepalivePop'
+    '$runtimeKeepalivePush', '$runtimeKeepalivePop',
+#endif
+#if ASYNCIFY == 1
+    // Needed by allocateData and handleSleep respectively
+    'malloc', 'free',
 #endif
   ],
+
   $Asyncify: {
-    State: {
-      Normal: 0,
-      Unwinding: 1,
-      Rewinding: 2,
-      Disabled: 3,
-    },
-    state: 0,
-    StackSize: {{{ ASYNCIFY_STACK_SIZE }}},
-    currData: null,
-    // The return value passed to wakeUp() in
-    // Asyncify.handleSleep(function(wakeUp){...}) is stored here,
-    // so we can return it later from the C function that called
-    // Asyncify.handleSleep() after rewinding finishes.
-    handleSleepReturnValue: 0,
-    // We must track which wasm exports are called into and
-    // exited, so that we know where the call stack began,
-    // which is where we must call to rewind it.
-    exportCallStack: [],
-    callStackNameToId: {},
-    callStackIdToName: {},
-    callStackId: 0,
-    asyncPromiseHandlers: null, // { resolve, reject } pair for when *all* asynchronicity is done
-    sleepCallbacks: [], // functions to call every time we sleep
-
-#if ASYNCIFY == 2
-    // TODO: Stack switching support could be implemented without all the
-    //       asyncify infrastructure, perhaps as code on the size instead of
-    //       ifdefs. That might be cleaner, and it would avoid overhead like the
-    //       malloc() asyncify does (we could ifdef it out, but it's even more
-    //       ifdefing).
-
-    // The promise that is being suspended on in the VM atm, or null.
-    promise: null,
-    // The function we should call to resolve the promise at the right time.
-    promiseResolve: null,
-#endif
-
-    getCallStackId: function(funcName) {
-      var id = Asyncify.callStackNameToId[funcName];
-      if (id === undefined) {
-        id = Asyncify.callStackId++;
-        Asyncify.callStackNameToId[funcName] = id;
-        Asyncify.callStackIdToName[id] = funcName;
-      }
-      return id;
-    },
-
+    //
+    // Asyncify code that is shared between mode 1 (original) and mode 2 (JSPI).
+    //
     instrumentWasmImports: function(imports) {
 #if ASYNCIFY_DEBUG
-      err('asyncify instrumenting imports');
+      dbg('asyncify instrumenting imports');
 #endif
-      var ASYNCIFY_IMPORTS = {{{ JSON.stringify(ASYNCIFY_IMPORTS) }}}.map((x) => x.split('.')[1]);
+      var importPatterns = [{{{ ASYNCIFY_IMPORTS.map(x => '/^' + x.split('.')[1].replace(new RegExp('\\*', 'g'), '.*') + '$/') }}}];
+
       for (var x in imports) {
         (function(x) {
           var original = imports[x];
           var sig = original.sig;
           if (typeof original == 'function') {
-            var isAsyncifyImport = ASYNCIFY_IMPORTS.indexOf(x) >= 0 ||
-                                   x.startsWith('__asyncjs__');
+            var isAsyncifyImport = original.isAsync ||
+                                   importPatterns.some(pattern => !!x.match(pattern));
 #if ASYNCIFY == 2
-            // Wrap all imports with a suspending WebAssembly function.
-            // TODO: wrap only async functions.
+            // Wrap async imports with a suspending WebAssembly function.
+            if (isAsyncifyImport) {
 #if ASSERTIONS
-            assert(sig, 'Missing __sig for ' + x);
+              assert(sig, 'Missing __sig for ' + x);
 #endif
-            var type = sigToWasmTypes(sig, original);
+              var type = sigToWasmTypes(sig);
 #if ASYNCIFY_DEBUG
-            err('asyncify: suspendOnReturnedPromise for', x, original);
+              dbg('asyncify: suspendOnReturnedPromise for', x, original);
 #endif
-            // Add space for the suspender promise that will be used in the
-            // Wasm wrapper function.
-            type.parameters.unshift('externref');
-            imports[x] = original = new WebAssembly.Function(
-              type,
-              original,
-              { suspending: 'first' }
-            );
+              // Add space for the suspender promise that will be used in the
+              // Wasm wrapper function.
+              type.parameters.unshift('externref');
+              imports[x] = original = new WebAssembly.Function(
+                type,
+                original,
+                { suspending: 'first' }
+              );
+            }
 #endif
 #if ASSERTIONS && ASYNCIFY != 2 // We cannot apply assertions with stack switching, as the imports must not be modified from suspender.suspendOnReturnedPromise TODO find a way
             imports[x] = function() {
@@ -144,7 +107,10 @@ mergeInto(LibraryManager.library, {
 
     instrumentWasmExports: function(exports) {
 #if ASYNCIFY_DEBUG
-      err('asyncify instrumenting exports');
+      dbg('asyncify instrumenting exports');
+#endif
+#if ASYNCIFY == 2
+      var ASYNCIFY_EXPORTS = {{{ JSON.stringify(ASYNCIFY_EXPORTS) }}};
 #endif
       var ret = {};
       for (var x in exports) {
@@ -153,40 +119,32 @@ mergeInto(LibraryManager.library, {
           if (typeof original == 'function') {
 #if ASYNCIFY == 2
             // Wrap all exports with a promising WebAssembly function.
-            // TODO: wrap only async functions.
-#if ASYNCIFY_DEBUG
-            err('asyncify: returnPromiseOnSuspend for', x, original);
-#endif
-            var type = WebAssembly.Function.type(original);
-            var parameters = type.parameters;
-            var results = type.results;
-#if ASSERTIONS
-            assert(results.length !== 0, 'There must be a return result')
-#endif
-            // Remove the extern ref;
-            parameters.shift();
-            original = new WebAssembly.Function(
-              { parameters , results: ['externref'] },
-              original,
-              { promising : 'first' });
+            var isAsyncifyExport = ASYNCIFY_EXPORTS.indexOf(x) >= 0;
+            if (isAsyncifyExport) {
+              original = Asyncify.makeAsyncFunction(original);
+            }
 #endif
             ret[x] = function() {
 #if ASYNCIFY_DEBUG >= 2
-              err('ASYNCIFY: ' + '  '.repeat(Asyncify.exportCallStack.length) + ' try ' + x);
+              dbg('ASYNCIFY: ' + '  '.repeat(Asyncify.exportCallStack.length) + ' try ' + x);
 #endif
+#if ASYNCIFY == 1
               Asyncify.exportCallStack.push(x);
               try {
+#endif
                 return original.apply(null, arguments);
+#if ASYNCIFY == 1
               } finally {
                 if (!ABORT) {
                   var y = Asyncify.exportCallStack.pop();
                   assert(y === x);
 #if ASYNCIFY_DEBUG >= 2
-                  err('ASYNCIFY: ' + '  '.repeat(Asyncify.exportCallStack.length) + ' finally ' + x);
+                  dbg('ASYNCIFY: ' + '  '.repeat(Asyncify.exportCallStack.length) + ' finally ' + x);
 #endif
                   Asyncify.maybeStopUnwind();
                 }
               }
+#endif
             };
 #if MAIN_MODULE
             ret[x].orig = original;
@@ -199,9 +157,47 @@ mergeInto(LibraryManager.library, {
       return ret;
     },
 
+#if ASYNCIFY == 1
+    //
+    // Original implementation of Asyncify.
+    //
+    State: {
+      Normal: 0,
+      Unwinding: 1,
+      Rewinding: 2,
+      Disabled: 3,
+    },
+    state: 0,
+    StackSize: {{{ ASYNCIFY_STACK_SIZE }}},
+    currData: null,
+    // The return value passed to wakeUp() in
+    // Asyncify.handleSleep(function(wakeUp){...}) is stored here,
+    // so we can return it later from the C function that called
+    // Asyncify.handleSleep() after rewinding finishes.
+    handleSleepReturnValue: 0,
+    // We must track which wasm exports are called into and
+    // exited, so that we know where the call stack began,
+    // which is where we must call to rewind it.
+    exportCallStack: [],
+    callStackNameToId: {},
+    callStackIdToName: {},
+    callStackId: 0,
+    asyncPromiseHandlers: null, // { resolve, reject } pair for when *all* asynchronicity is done
+    sleepCallbacks: [], // functions to call every time we sleep
+
+    getCallStackId: function(funcName) {
+      var id = Asyncify.callStackNameToId[funcName];
+      if (id === undefined) {
+        id = Asyncify.callStackId++;
+        Asyncify.callStackNameToId[funcName] = id;
+        Asyncify.callStackIdToName[id] = funcName;
+      }
+      return id;
+    },
+
     maybeStopUnwind: function() {
 #if ASYNCIFY_DEBUG
-      err('ASYNCIFY: maybe stop unwind', Asyncify.exportCallStack);
+      dbg('ASYNCIFY: maybe stop unwind', Asyncify.exportCallStack);
 #endif
       if (Asyncify.currData &&
           Asyncify.state === Asyncify.State.Unwinding &&
@@ -209,17 +205,15 @@ mergeInto(LibraryManager.library, {
         // We just finished unwinding.
         // Be sure to set the state before calling any other functions to avoid
         // possible infinite recursion here (For example in debug pthread builds
-        // the err() function itself can call back into WebAssembly to get the
+        // the dbg() function itself can call back into WebAssembly to get the
         // current pthread_self() pointer).
         Asyncify.state = Asyncify.State.Normal;
 #if ASYNCIFY_DEBUG
-        err('ASYNCIFY: stop unwind');
+        dbg('ASYNCIFY: stop unwind');
 #endif
         {{{ runtimeKeepalivePush(); }}}
         // Keep the runtime alive so that a re-wind can be done later.
-#if ASYNCIFY == 1
         runAndAbortIfError(_asyncify_stop_unwind);
-#endif
         if (typeof Fibers != 'undefined') {
           Fibers.trampoline();
         }
@@ -262,7 +256,7 @@ mergeInto(LibraryManager.library, {
     setDataRewindFunc: function(ptr) {
       var bottomOfCallStack = Asyncify.exportCallStack[0];
 #if ASYNCIFY_DEBUG >= 2
-      err('ASYNCIFY: setDataRewindFunc('+ptr+'), bottomOfCallStack is', bottomOfCallStack, new Error().stack);
+      dbg('ASYNCIFY: setDataRewindFunc('+ptr+'), bottomOfCallStack is', bottomOfCallStack, new Error().stack);
 #endif
       var rewindId = Asyncify.getCallStackId(bottomOfCallStack);
       {{{ makeSetValue('ptr', C_STRUCTS.asyncify_data_s.rewind_id, 'rewindId', 'i32') }}};
@@ -279,32 +273,21 @@ mergeInto(LibraryManager.library, {
       // Exported functions in side modules are not listed in `Module["asm"]`,
       // So we should use `resolveGlobalSymbol` helper function, which is defined in `library_dylink.js`.
       if (!func) {
-        func = resolveGlobalSymbol(name, false);
+        func = resolveGlobalSymbol(name, false).sym;
       }
 #endif
       return func;
     },
 
     doRewind: function(ptr) {
-#if ASYNCIFY == 2
-      // Resolve the promise. The VM will resume the wasm on the next event loop
-      // turn.
-      Asyncify.promiseResolve(Asyncify.handleSleepReturnValue);
-      setTimeout(() => {
-        // This timeout happens after the wasm has been resumed; we can stop
-        // artificially keeping the runtime alive at that time.
-        {{{ runtimeKeepalivePop(); }}}
-      });
-#else
       var start = Asyncify.getDataRewindFunc(ptr);
 #if ASYNCIFY_DEBUG
-      err('ASYNCIFY: start:', start);
+      dbg('ASYNCIFY: start:', start);
 #endif
       // Once we have rewound and the stack we no longer need to artificially
       // keep the runtime alive.
       {{{ runtimeKeepalivePop(); }}}
       return start();
-#endif
     },
 
     // This receives a function to call to start the async operation, and
@@ -316,7 +299,7 @@ mergeInto(LibraryManager.library, {
 #endif
       if (ABORT) return;
 #if ASYNCIFY_DEBUG
-      err('ASYNCIFY: handleSleep ' + Asyncify.state);
+      dbg('ASYNCIFY: handleSleep ' + Asyncify.state);
 #endif
       if (Asyncify.state === Asyncify.State.Normal) {
         // Prepare to sleep. Call startAsync, and see what happens:
@@ -325,12 +308,12 @@ mergeInto(LibraryManager.library, {
         // need to do anything.
         var reachedCallback = false;
         var reachedAfterCallback = false;
-        startAsync((handleSleepReturnValue) => {
+        startAsync((handleSleepReturnValue = 0) => {
 #if ASSERTIONS
           assert(!handleSleepReturnValue || typeof handleSleepReturnValue == 'number' || typeof handleSleepReturnValue == 'boolean'); // old emterpretify API supported other stuff
 #endif
           if (ABORT) return;
-          Asyncify.handleSleepReturnValue = handleSleepReturnValue || 0;
+          Asyncify.handleSleepReturnValue = handleSleepReturnValue;
           reachedCallback = true;
           if (!reachedAfterCallback) {
             // We are happening synchronously, so no need for async.
@@ -345,12 +328,10 @@ mergeInto(LibraryManager.library, {
           assert(!Asyncify.exportCallStack.length, 'Waking up (starting to rewind) must be done from JS, without compiled code on the stack.');
 #endif
 #if ASYNCIFY_DEBUG
-          err('ASYNCIFY: start rewind ' + Asyncify.currData);
+          dbg('ASYNCIFY: start rewind ' + Asyncify.currData);
 #endif
           Asyncify.state = Asyncify.State.Rewinding;
-#if ASYNCIFY == 1
           runAndAbortIfError(() => _asyncify_start_rewind(Asyncify.currData));
-#endif
           if (typeof Browser != 'undefined' && Browser.mainLoop.func) {
             Browser.mainLoop.resume();
           }
@@ -397,31 +378,20 @@ mergeInto(LibraryManager.library, {
           // TODO: reuse, don't alloc/free every sleep
           Asyncify.currData = Asyncify.allocateData();
 #if ASYNCIFY_DEBUG
-          err('ASYNCIFY: start unwind ' + Asyncify.currData);
+          dbg('ASYNCIFY: start unwind ' + Asyncify.currData);
 #endif
           if (typeof Browser != 'undefined' && Browser.mainLoop.func) {
             Browser.mainLoop.pause();
           }
-#if ASYNCIFY == 2
-          // Return a Promise to get the browser's stack switching logic to run.
-          return Asyncify.promise = new Promise((resolve, reject) => {
-            // Stash the resolve hook so we can call it at the proper time.
-            Asyncify.promiseResolve = resolve;
-            // TODO: handle rejection
-          });
-#else
           runAndAbortIfError(() => _asyncify_start_unwind(Asyncify.currData));
-#endif
         }
       } else if (Asyncify.state === Asyncify.State.Rewinding) {
         // Stop a resume.
 #if ASYNCIFY_DEBUG
-        err('ASYNCIFY: stop rewind');
+        dbg('ASYNCIFY: stop rewind');
 #endif
         Asyncify.state = Asyncify.State.Normal;
-#if ASYNCIFY == 1
         runAndAbortIfError(_asyncify_stop_rewind);
-#endif
         _free(Asyncify.currData);
         Asyncify.currData = null;
         // Call all sleep callbacks now that the sleep-resume is all done.
@@ -443,10 +413,50 @@ mergeInto(LibraryManager.library, {
         startAsync().then(wakeUp);
       });
     },
+
+#elif ASYNCIFY == 2
+    //
+    // JSPI implementation of Asyncify.
+    //
+    handleSleep: function(startAsync) {
+      {{{ runtimeKeepalivePush(); }}}
+      var promise = new Promise((resolve) => {
+        startAsync(resolve);
+      });
+      promise.finally(() => {
+        {{{ runtimeKeepalivePop(); }}}
+      });
+      return promise;
+    },
+    handleAsync: function(startAsync) {
+      return Asyncify.handleSleep((wakeUp) => {
+        // TODO: add error handling as a second param when handleSleep implements it.
+        startAsync().then(wakeUp);
+      });
+    },
+    makeAsyncFunction: function(original) {
+#if ASYNCIFY_DEBUG
+      dbg('asyncify: returnPromiseOnSuspend for', original);
+#endif
+      var type = WebAssembly.Function.type(original);
+      var parameters = type.parameters;
+      var results = type.results;
+#if ASSERTIONS
+      assert(results.length !== 0, 'There must be a return result')
+      assert(parameters[0] === 'externref', 'First param must be externref.');
+#endif
+      // Remove the extern ref.
+      parameters.shift();
+      return new WebAssembly.Function(
+        { parameters , results: ['externref'] },
+        original,
+        { promising : 'first' });
+    },
+#endif
   },
 
-  emscripten_sleep__sig: 'vi',
   emscripten_sleep__deps: ['$safeSetTimeout'],
+  emscripten_sleep__async: true,
   emscripten_sleep: function(ms) {
     // emscripten_sleep() does not return a value, but we still need a |return|
     // here for stack switching support (ASYNCIFY=2). In that mode this function
@@ -455,13 +465,13 @@ mergeInto(LibraryManager.library, {
     return Asyncify.handleSleep((wakeUp) => safeSetTimeout(wakeUp, ms));
   },
 
-  emscripten_wget__sig: 'vpp',
   emscripten_wget__deps: ['$Browser', '$PATH_FS', '$FS'],
+  emscripten_wget__async: true,
   emscripten_wget: function(url, file) {
     return Asyncify.handleSleep((wakeUp) => {
       var _url = UTF8ToString(url);
       var _file = UTF8ToString(file);
-      _file = PATH_FS.resolve(FS.cwd(), _file);
+      _file = PATH_FS.resolve(_file);
       var destinationDirectory = PATH.dirname(_file);
       FS.createPreloadedFile(
         destinationDirectory,
@@ -483,8 +493,8 @@ mergeInto(LibraryManager.library, {
     });
   },
 
-  emscripten_wget_data__sig: 'vpppp',
   emscripten_wget_data__deps: ['$asyncLoad', 'malloc'],
+  emscripten_wget_data__async: true,
   emscripten_wget_data: function(url, pbuffer, pnum, perror) {
     return Asyncify.handleSleep((wakeUp) => {
       asyncLoad(UTF8ToString(url), (byteArray) => {
@@ -502,8 +512,8 @@ mergeInto(LibraryManager.library, {
     });
   },
 
-  emscripten_scan_registers__sig: 'vp',
   emscripten_scan_registers__deps: ['$safeSetTimeout'],
+  emscripten_scan_registers__async: true,
   emscripten_scan_registers: function(func) {
     return Asyncify.handleSleep((wakeUp) => {
       // We must first unwind, so things are spilled to the stack. Then while
@@ -520,6 +530,7 @@ mergeInto(LibraryManager.library, {
   },
 
   emscripten_lazy_load_code__sig: 'v',
+  emscripten_lazy_load_code__async: true,
   emscripten_lazy_load_code: function() {
     return Asyncify.handleSleep((wakeUp) => {
       // Update the expected wasm binary file to be the lazy one.
@@ -531,6 +542,18 @@ mergeInto(LibraryManager.library, {
     });
   },
 
+  _load_secondary_module__sig: 'v',
+  _load_secondary_module: async function() {
+    // Mark the module as loading for the wasm module (so it doesn't try to load it again).
+    Module['asm']['load_secondary_module_status'].value = 1;
+    var imports = {'primary': Module['asm']};
+    // Replace '.wasm' suffix with '.deferred.wasm'.
+    var deferred = wasmBinaryFile.slice(0, -5) + '.deferred.wasm';
+    await new Promise((resolve) => {
+      instantiateAsync(null, deferred, imports, resolve);
+    });
+  },
+
   $Fibers__deps: ['$Asyncify'],
   $Fibers: {
     nextFiber: 0,
@@ -542,7 +565,7 @@ mergeInto(LibraryManager.library, {
           var fiber = Fibers.nextFiber;
           Fibers.nextFiber = 0;
 #if ASYNCIFY_DEBUG >= 2
-          err("ASYNCIFY/FIBER: trampoline jump into fiber", fiber, new Error().stack);
+          dbg("ASYNCIFY/FIBER: trampoline jump into fiber", fiber, new Error().stack);
 #endif
           Fibers.finishContextSwitch(fiber);
         } while (Fibers.nextFiber);
@@ -570,7 +593,7 @@ mergeInto(LibraryManager.library, {
         writeStackCookie();
 #endif
 #if ASYNCIFY_DEBUG
-        err('ASYNCIFY/FIBER: entering fiber', newFiber, 'for the first time');
+        dbg('ASYNCIFY/FIBER: entering fiber', newFiber, 'for the first time');
 #endif
         Asyncify.currData = null;
         {{{ makeSetValue('newFiber', C_STRUCTS.emscripten_fiber_s.entry, 0, 'i32') }}};
@@ -582,7 +605,7 @@ mergeInto(LibraryManager.library, {
         Asyncify.currData = asyncifyData;
 
 #if ASYNCIFY_DEBUG
-        err('ASYNCIFY/FIBER: start rewind', asyncifyData, '(resuming fiber', newFiber, ')');
+        dbg('ASYNCIFY/FIBER: start rewind', asyncifyData, '(resuming fiber', newFiber, ')');
 #endif
         Asyncify.state = Asyncify.State.Rewinding;
         _asyncify_start_rewind(asyncifyData);
@@ -591,38 +614,12 @@ mergeInto(LibraryManager.library, {
     },
   },
 
-  emscripten_fiber_init__sig: 'viiiiiii',
-  emscripten_fiber_init__deps: ['$Asyncify'],
-  emscripten_fiber_init: function(fiber, entryPoint, userData, cStack, cStackSize, asyncStack, asyncStackSize) {
-    var cStackBase = cStack + cStackSize;
-
-    {{{ makeSetValue('fiber', C_STRUCTS.emscripten_fiber_s.stack_base,  'cStackBase',  'i32') }}};
-    {{{ makeSetValue('fiber', C_STRUCTS.emscripten_fiber_s.stack_limit, 'cStack',      'i32') }}};
-    {{{ makeSetValue('fiber', C_STRUCTS.emscripten_fiber_s.stack_ptr,   'cStackBase',  'i32') }}};
-    {{{ makeSetValue('fiber', C_STRUCTS.emscripten_fiber_s.entry,       'entryPoint', 'i32') }}};
-    {{{ makeSetValue('fiber', C_STRUCTS.emscripten_fiber_s.user_data,   'userData',   'i32') }}};
-
-    var asyncifyData = fiber + {{{ C_STRUCTS.emscripten_fiber_s.asyncify_data }}};
-    Asyncify.setDataHeader(asyncifyData, asyncStack, asyncStackSize);
-  },
-
-  emscripten_fiber_init_from_current_context__sig: 'vii',
-  emscripten_fiber_init_from_current_context__deps: ['$Asyncify'],
-  emscripten_fiber_init_from_current_context: function(fiber, asyncStack, asyncStackSize) {
-    {{{ makeSetValue('fiber', C_STRUCTS.emscripten_fiber_s.stack_base,  '_emscripten_stack_get_base()', 'i32') }}};
-    {{{ makeSetValue('fiber', C_STRUCTS.emscripten_fiber_s.stack_limit, '_emscripten_stack_get_end()',  'i32') }}};
-    {{{ makeSetValue('fiber', C_STRUCTS.emscripten_fiber_s.entry,       0,            'i32') }}};
-
-    var asyncifyData = fiber + {{{ C_STRUCTS.emscripten_fiber_s.asyncify_data }}};
-    Asyncify.setDataHeader(asyncifyData, asyncStack, asyncStackSize);
-  },
-
-  emscripten_fiber_swap__sig: 'vii',
   emscripten_fiber_swap__deps: ["$Asyncify", "$Fibers"],
+  emscripten_fiber_swap__async: true,
   emscripten_fiber_swap: function(oldFiber, newFiber) {
     if (ABORT) return;
 #if ASYNCIFY_DEBUG
-    err('ASYNCIFY/FIBER: swap', oldFiber, '->', newFiber, 'state:', Asyncify.state);
+    dbg('ASYNCIFY/FIBER: swap', oldFiber, '->', newFiber, 'state:', Asyncify.state);
 #endif
     if (Asyncify.state === Asyncify.State.Normal) {
       Asyncify.state = Asyncify.State.Unwinding;
@@ -632,7 +629,7 @@ mergeInto(LibraryManager.library, {
       Asyncify.currData = asyncifyData;
 
 #if ASYNCIFY_DEBUG
-      err('ASYNCIFY/FIBER: start unwind', asyncifyData);
+      dbg('ASYNCIFY/FIBER: start unwind', asyncifyData);
 #endif
       _asyncify_start_unwind(asyncifyData);
 
@@ -645,7 +642,7 @@ mergeInto(LibraryManager.library, {
       assert(Asyncify.state === Asyncify.State.Rewinding);
 #endif
 #if ASYNCIFY_DEBUG
-      err('ASYNCIFY/FIBER: stop rewind');
+      dbg('ASYNCIFY/FIBER: stop rewind');
 #endif
       Asyncify.state = Asyncify.State.Normal;
       _asyncify_stop_rewind();
@@ -656,13 +653,13 @@ mergeInto(LibraryManager.library, {
   emscripten_sleep: function() {
     throw 'Please compile your program with async support in order to use asynchronous operations like emscripten_sleep';
   },
-  emscripten_wget: function() {
+  emscripten_wget: function(url, file) {
     throw 'Please compile your program with async support in order to use asynchronous operations like emscripten_wget';
   },
-  emscripten_wget_data: function() {
+  emscripten_wget_data: function(url, pbuffer, pnum, perror) {
     throw 'Please compile your program with async support in order to use asynchronous operations like emscripten_wget_data';
   },
-  emscripten_scan_registers: function() {
+  emscripten_scan_registers: function(func) {
     throw 'Please compile your program with async support in order to use asynchronous operations like emscripten_scan_registers';
   },
   emscripten_fiber_init: function() {
@@ -671,7 +668,7 @@ mergeInto(LibraryManager.library, {
   emscripten_fiber_init_from_current_context: function() {
     throw 'Please compile your program with async support in order to use asynchronous operations like emscripten_fiber_init_from_current_context';
   },
-  emscripten_fiber_swap: function() {
+  emscripten_fiber_swap: function(oldFiber, newFiber) {
     throw 'Please compile your program with async support in order to use asynchronous operations like emscripten_fiber_swap';
   },
 #endif // ASYNCIFY
