@@ -68,7 +68,9 @@ var LibraryGLFW = {
       this.windowRefreshFunc = null; // GLFWwindowrefreshfun
       this.windowFocusFunc = null; // GLFWwindowfocusfun
       this.windowIconifyFunc = null; // GLFWwindowiconifyfun
+      this.windowMaximizeFunc = null; // GLFWwindowmaximizefun
       this.framebufferSizeFunc = null; // GLFWframebuffersizefun
+      this.windowContentScaleFunc = null; // GLFWwindowcontentscalefun
       this.mouseButtonFunc = null; // GLFWmousebuttonfun
       this.cursorPosFunc = null; // GLFWcursorposfun
       this.cursorEnterFunc = null; // GLFWcursorenterfun
@@ -80,7 +82,7 @@ var LibraryGLFW = {
     },
 
   $GLFW__deps: ['emscripten_get_now', '$GL', '$Browser', '$GLFW_Window',
-    '$allocateUTF8',
+    '$stringToNewUTF8',
 #if FILESYSTEM
     '$FS',
 #endif
@@ -95,6 +97,7 @@ var LibraryGLFW = {
     errorFunc: null, // GLFWerrorfun
     monitorFunc: null, // GLFWmonitorfun
     active: null, // active window
+    scale: null,
     windows: null,
     monitors: null,
     monitorString: null,
@@ -108,6 +111,8 @@ var LibraryGLFW = {
       0x00020003:1, // GLFW_RESIZABLE
       0x00020004:1, // GLFW_VISIBLE
       0x00020005:1, // GLFW_DECORATED
+      0x0002000A:0, // GLFW_TRANSPARENT_FRAMEBUFFER
+      0x0002200C:0, // GLFW_SCALE_TO_MONITOR. can we emulate this?
 
       0x00021001:8, // GLFW_RED_BITS
       0x00021002:8, // GLFW_GREEN_BITS
@@ -350,6 +355,7 @@ var LibraryGLFW = {
       if (win.keys[341]) mod |= 0x0002; // GLFW_MOD_CONTROL
       if (win.keys[342]) mod |= 0x0004; // GLFW_MOD_ALT
       if (win.keys[343]) mod |= 0x0008; // GLFW_MOD_SUPER
+      // add caps and num lock keys? only if lock_key_mod is set
       return mod;
     },
 
@@ -364,7 +370,6 @@ var LibraryGLFW = {
 #if USE_GLFW == 2
       {{{ makeDynCall('vii', 'GLFW.active.charFunc') }}}(charCode, 1);
 #endif
-
 #if USE_GLFW == 3
       {{{ makeDynCall('vii', 'GLFW.active.charFunc') }}}(GLFW.active.id, charCode);
 #endif
@@ -381,16 +386,16 @@ var LibraryGLFW = {
 #endif
       GLFW.active.keys[key] = status;
       GLFW.active.domKeys[keyCode] = status;
-      if (!GLFW.active.keyFunc) return;
 
+      if (GLFW.active.keyFunc) {
 #if USE_GLFW == 2
-      {{{ makeDynCall('vii', 'GLFW.active.keyFunc') }}}(key, status);
+        {{{ makeDynCall('vii', 'GLFW.active.keyFunc') }}}(key, status);
 #endif
-
 #if USE_GLFW == 3
-      if (repeat) status = 2; // GLFW_REPEAT
-      {{{ makeDynCall('viiiii', 'GLFW.active.keyFunc') }}}(GLFW.active.id, key, keyCode, status, GLFW.getModBits(GLFW.active));
+        if (repeat) status = 2; // GLFW_REPEAT
+        {{{ makeDynCall('viiiii', 'GLFW.active.keyFunc') }}}(GLFW.active.id, key, keyCode, status, GLFW.getModBits(GLFW.active));
 #endif
+      }
     },
 
     onGamepadConnected: function(event) {
@@ -433,13 +438,14 @@ var LibraryGLFW = {
 
       if (event.target != Module["canvas"] || !GLFW.active.cursorPosFunc) return;
 
+      if (GLFW.active.cursorPosFunc) {
 #if USE_GLFW == 2
-      {{{ makeDynCall('vii', 'GLFW.active.cursorPosFunc') }}}(Browser.mouseX, Browser.mouseY);
+        {{{ makeDynCall('vii', 'GLFW.active.cursorPosFunc') }}}(Browser.mouseX, Browser.mouseY);
 #endif
-
 #if USE_GLFW == 3
-      {{{ makeDynCall('vidd', 'GLFW.active.cursorPosFunc') }}}(GLFW.active.id, Browser.mouseX, Browser.mouseY);
+        {{{ makeDynCall('vidd', 'GLFW.active.cursorPosFunc') }}}(GLFW.active.id, Browser.mouseX, Browser.mouseY);
 #endif
+      }
     },
 
     DOMToGLFWMouseButton: function(event) {
@@ -459,20 +465,24 @@ var LibraryGLFW = {
     onMouseenter: function(event) {
       if (!GLFW.active) return;
 
-      if (event.target != Module["canvas"] || !GLFW.active.cursorEnterFunc) return;
+      if (event.target != Module["canvas"]) return;
 
 #if USE_GLFW == 3
-      {{{ makeDynCall('vii', 'GLFW.active.cursorEnterFunc') }}}(GLFW.active.id, 1);
+      if (GLFW.active.cursorEnterFunc) {
+        {{{ makeDynCall('vii', 'GLFW.active.cursorEnterFunc') }}}(GLFW.active.id, 1);
+      }
 #endif
     },
 
     onMouseleave: function(event) {
       if (!GLFW.active) return;
 
-      if (event.target != Module["canvas"] || !GLFW.active.cursorEnterFunc) return;
+      if (event.target != Module["canvas"]) return;
 
 #if USE_GLFW == 3
-      {{{ makeDynCall('vii', 'GLFW.active.cursorEnterFunc') }}}(GLFW.active.id, 0);
+      if (GLFW.active.cursorEnterFunc) {
+        {{{ makeDynCall('vii', 'GLFW.active.cursorEnterFunc') }}}(GLFW.active.id, 0);
+      }
 #endif
     },
 
@@ -494,15 +504,14 @@ var LibraryGLFW = {
         GLFW.active.buttons &= ~(1 << eventButton);
       }
 
-      if (!GLFW.active.mouseButtonFunc) return;
-
+      if (GLFW.active.mouseButtonFunc) {
 #if USE_GLFW == 2
-      {{{ makeDynCall('vii', 'GLFW.active.mouseButtonFunc') }}}(eventButton, status);
+        {{{ makeDynCall('vii', 'GLFW.active.mouseButtonFunc') }}}(eventButton, status);
 #endif
-
 #if USE_GLFW == 3
-      {{{ makeDynCall('viiii', 'GLFW.active.mouseButtonFunc') }}}(GLFW.active.id, eventButton, status, GLFW.getModBits(GLFW.active));
+        {{{ makeDynCall('viiii', 'GLFW.active.mouseButtonFunc') }}}(GLFW.active.id, eventButton, status, GLFW.getModBits(GLFW.active));
 #endif
+      }
     },
 
     onMouseButtonDown: function(event) {
@@ -522,11 +531,9 @@ var LibraryGLFW = {
       GLFW.wheelPos += delta;
 
       if (!GLFW.active || !GLFW.active.scrollFunc || event.target != Module['canvas']) return;
-
 #if USE_GLFW == 2
       {{{ makeDynCall('vi', 'GLFW.active.scrollFunc') }}}(GLFW.wheelPos);
 #endif
-
 #if USE_GLFW == 3
       var sx = 0;
       var sy = delta;
@@ -588,24 +595,34 @@ var LibraryGLFW = {
     onWindowSizeChanged: function() {
       if (!GLFW.active) return;
 
-      if (!GLFW.active.windowSizeFunc) return;
-
+      if (GLFW.active.windowSizeFunc) {
 #if USE_GLFW == 2
-      {{{ makeDynCall('vii', 'GLFW.active.windowSizeFunc') }}}(GLFW.active.width, GLFW.active.height);
+        {{{ makeDynCall('vii', 'GLFW.active.windowSizeFunc') }}}(GLFW.active.width, GLFW.active.height);
 #endif
-
 #if USE_GLFW == 3
-      {{{ makeDynCall('viii', 'GLFW.active.windowSizeFunc') }}}(GLFW.active.id, GLFW.active.width, GLFW.active.height);
+        {{{ makeDynCall('viii', 'GLFW.active.windowSizeFunc') }}}(GLFW.active.id, GLFW.active.width, GLFW.active.height);
 #endif
+      }
     },
 
     onFramebufferSizeChanged: function() {
       if (!GLFW.active) return;
 
-      if (!GLFW.active.framebufferSizeFunc) return;
+#if USE_GLFW == 3
+      if (GLFW.active.framebufferSizeFunc) {
+        {{{ makeDynCall('viii', 'GLFW.active.framebufferSizeFunc') }}}(GLFW.active.id, GLFW.active.width, GLFW.active.height);
+      }
+#endif
+    },
+
+    onWindowContentScaleChanged: function(scale) {
+      GLFW.scale = scale;
+      if (!GLFW.active) return;
 
 #if USE_GLFW == 3
-      {{{ makeDynCall('viii', 'GLFW.active.framebufferSizeFunc') }}}(GLFW.active.id, GLFW.active.width, GLFW.active.height);
+      if (GLFW.active.windowContentScaleFunc) {
+        {{{ makeDynCall('viff', 'GLFW.active.windowContentScaleFunc') }}}(GLFW.active.id, GLFW.scale, GLFW.scale);
+      }
 #endif
     },
 
@@ -647,7 +664,7 @@ var LibraryGLFW = {
             if (!GLFW.joys[joy]) {
               out('glfw joystick connected:',joy);
               GLFW.joys[joy] = {
-                id: allocateUTF8(gamepad.id),
+                id: stringToNewUTF8(gamepad.id),
                 buttonsCount: gamepad.buttons.length,
                 axesCount: gamepad.axes.length,
                 buttons: _malloc(gamepad.buttons.length),
@@ -774,7 +791,7 @@ var LibraryGLFW = {
         };
         reader.readAsArrayBuffer(file);
 
-        var filename = allocateUTF8(path);
+        var filename = stringToNewUTF8(path);
         filenamesArray.push(filename);
         {{{ makeSetValue('filenames + i*4', '0', 'filename', POINTER_TYPE) }}};
       }
@@ -873,6 +890,14 @@ var LibraryGLFW = {
           out("glfwSetInputMode called with GLFW_STICKY_MOUSE_BUTTONS mode not implemented.");
           break;
         }
+        case 0x00033004: { // GLFW_LOCK_KEY_MODS
+          out("glfwSetInputMode called with GLFW_LOCK_KEY_MODS mode not implemented.");
+          break;
+        }
+        case 0x000330005: { // GLFW_RAW_MOUSE_MOTION
+          out("glfwSetInputMode called with GLFW_RAW_MOUSE_MOTION mode not implemented.");
+          break;
+        }
         default: {
           out("glfwSetInputMode called with unknown mode parameter value: " + mode + ".");
           break;
@@ -965,15 +990,14 @@ var LibraryGLFW = {
         }
       }
 
-      if (!win.windowSizeFunc) return;
-
+      if (win.windowSizeFunc) {
 #if USE_GLFW == 2
-      {{{ makeDynCall('vii', 'win.windowSizeFunc') }}}(width, height);
+        {{{ makeDynCall('vii', 'win.windowSizeFunc') }}}(width, height);
 #endif
-
 #if USE_GLFW == 3
-      {{{ makeDynCall('viii', 'win.windowSizeFunc') }}}(win.id, width, height);
+        {{{ makeDynCall('viii', 'win.windowSizeFunc') }}}(win.id, width, height);
 #endif
+      }
     },
 
     createWindow: function(width, height, title, monitor, share) {
@@ -1040,8 +1064,9 @@ var LibraryGLFW = {
       if (!win) return;
 
 #if USE_GLFW == 3
-      if (win.windowCloseFunc)
+      if (win.windowCloseFunc) {
         {{{ makeDynCall('vi', 'win.windowCloseFunc') }}}(win.id);
+      }
 #endif
 
       GLFW.windows[win.id - 1] = null;
@@ -1098,7 +1123,7 @@ var LibraryGLFW = {
 /*******************************************************************************
  * GLFW FUNCTIONS
  ******************************************************************************/
-  glfwInit__sig: 'i',
+  glfwInit__deps: ['emscripten_get_device_pixel_ratio', 'malloc', 'free'],
   glfwInit: function() {
     if (GLFW.windows) return 1; // GL_TRUE
 
@@ -1106,6 +1131,8 @@ var LibraryGLFW = {
     GLFW.hints = GLFW.defaultHints;
     GLFW.windows = new Array()
     GLFW.active = null;
+    GLFW.scale  = _emscripten_get_device_pixel_ratio();
+
 
     window.addEventListener("gamepadconnected", GLFW.onGamepadConnected, true);
     window.addEventListener("gamepaddisconnected", GLFW.onGamepadDisconnected, true);
@@ -1113,6 +1140,13 @@ var LibraryGLFW = {
     window.addEventListener("keypress", GLFW.onKeyPress, true);
     window.addEventListener("keyup", GLFW.onKeyup, true);
     window.addEventListener("blur", GLFW.onBlur, true);
+    // from https://stackoverflow.com/a/70514686/7484780 . maybe add this to browser.js?
+    // no idea how to remove this listener.
+    (function updatePixelRatio(){
+      window.matchMedia("(resolution: " + window.devicePixelRatio + "dppx)")
+      .addEventListener('change', updatePixelRatio, {once: true});
+      GLFW.onWindowContentScaleChanged(_emscripten_get_device_pixel_ratio());
+      })();
     Module["canvas"].addEventListener("touchmove", GLFW.onMousemove, true);
     Module["canvas"].addEventListener("touchstart", GLFW.onMouseButtonDown, true);
     Module["canvas"].addEventListener("touchcancel", GLFW.onMouseButtonUp, true);
@@ -1133,7 +1167,6 @@ var LibraryGLFW = {
     return 1; // GL_TRUE
   },
 
-  glfwTerminate__sig: 'v',
   glfwTerminate: function() {
     window.removeEventListener("gamepadconnected", GLFW.onGamepadConnected, true);
     window.removeEventListener("gamepaddisconnected", GLFW.onGamepadDisconnected, true);
@@ -1161,7 +1194,6 @@ var LibraryGLFW = {
     GLFW.active = null;
   },
 
-  glfwGetVersion__sig: 'viii',
   glfwGetVersion: function(major, minor, rev) {
 #if USE_GLFW == 2
     {{{ makeSetValue('major', '0', '2', 'i32') }}};
@@ -1176,24 +1208,19 @@ var LibraryGLFW = {
 #endif
   },
 
-  glfwPollEvents__sig: 'v',
   glfwPollEvents: function() {},
 
-  glfwWaitEvents__sig: 'v',
   glfwWaitEvents: function() {},
 
-  glfwGetTime__sig: 'd',
   glfwGetTime: function() {
     return GLFW.getTime() - GLFW.initialTime;
   },
 
-  glfwSetTime__sig: 'vd',
   glfwSetTime: function(time) {
     GLFW.initialTime = GLFW.getTime() - time;
   },
 
   glfwExtensionSupported__deps: ['glGetString'],
-  glfwExtensionSupported__sig: 'ii',
   glfwExtensionSupported: function(extension) {
     if (!GLFW.extensions) {
       GLFW.extensions = UTF8ToString(_glGetString(0x1F03)).split(' ');
@@ -1207,7 +1234,6 @@ var LibraryGLFW = {
   },
 
   glfwSwapInterval__deps: ['emscripten_set_main_loop_timing'],
-  glfwSwapInterval__sig: 'vi',
   glfwSwapInterval: function(interval) {
     interval = Math.abs(interval); // GLFW uses negative values to enable GLX_EXT_swap_control_tear, which we don't have, so just treat negative and positive the same.
     if (interval == 0) _emscripten_set_main_loop_timing(0/*EM_TIMING_SETTIMEOUT*/, 0);
@@ -1215,49 +1241,50 @@ var LibraryGLFW = {
   },
 
 #if USE_GLFW == 3
-  glfwGetVersionString__sig: 'i',
   glfwGetVersionString: function() {
     if (!GLFW.versionString) {
-      GLFW.versionString = allocateUTF8("3.2.1 JS WebGL Emscripten");
+      GLFW.versionString = stringToNewUTF8("3.2.1 JS WebGL Emscripten");
     }
     return GLFW.versionString;
   },
 
-  glfwSetErrorCallback__sig: 'ii',
   glfwSetErrorCallback: function(cbfun) {
     var prevcbfun = GLFW.errorFunc;
     GLFW.errorFunc = cbfun;
     return prevcbfun;
   },
 
-  glfwWaitEventsTimeout__sig: 'vd',
   glfwWaitEventsTimeout: function(timeout) {},
 
-  glfwPostEmptyEvent__sig: 'v',
   glfwPostEmptyEvent: function() {},
 
-  glfwGetMonitors__sig: 'ii',
+  glfwGetMonitors__deps: ['malloc'],
   glfwGetMonitors: function(count) {
     {{{ makeSetValue('count', '0', '1', 'i32') }}};
     if (!GLFW.monitors) {
-      GLFW.monitors = {{{ makeMalloc('glfwGetMonitors', Runtime.POINTER_SIZE) }}};
+      GLFW.monitors = _malloc({{{ POINTER_SIZE }}});
       {{{ makeSetValue('GLFW.monitors', '0', '1', 'i32') }}};
     }
     return GLFW.monitors;
   },
 
-  glfwGetPrimaryMonitor__sig: 'i',
   glfwGetPrimaryMonitor: function() {
     return 1;
   },
 
-  glfwGetMonitorPos__sig: 'viii',
   glfwGetMonitorPos: function(monitor, x, y) {
     {{{ makeSetValue('x', '0', '0', 'i32') }}};
     {{{ makeSetValue('y', '0', '0', 'i32') }}};
   },
 
-  glfwGetMonitorPhysicalSize__sig: 'viii',
+  glfwGetMonitorWorkarea: function(monitor, x, y, w, h) {
+    {{{ makeSetValue('x', '0', '0', 'i32') }}};
+    {{{ makeSetValue('y', '0', '0', 'i32') }}};
+
+    {{{ makeSetValue('w', '0', 'screen.availWidth', 'i32') }}};
+    {{{ makeSetValue('h', '0', 'screen.availHeight', 'i32') }}};
+  },
+
   glfwGetMonitorPhysicalSize: function(monitor, width, height) {
     // AFAIK there is no way to do this in javascript
     // Maybe with platform specific ccalls?
@@ -1267,15 +1294,18 @@ var LibraryGLFW = {
     {{{ makeSetValue('height', '0', '0', 'i32') }}};
   },
 
-  glfwGetMonitorName__sig: 'ii',
+  glfwGetMonitorContentScale: function(monitor, x, y) {
+    {{{ makeSetValue('x', '0', 'GLFW.scale', 'float') }}};
+    {{{ makeSetValue('y', '0', 'GLFW.scale', 'float') }}};
+  },
+
   glfwGetMonitorName: function(mon) {
     if (!GLFW.monitorString) {
-      GLFW.monitorString = allocateUTF8("HTML5 WebGL Canvas");
+      GLFW.monitorString = stringToNewUTF8("HTML5 WebGL Canvas");
     }
     return GLFW.monitorString;
   },
 
-  glfwSetMonitorCallback__sig: 'ii',
   glfwSetMonitorCallback: function(cbfun) {
     var prevcbfun = GLFW.monitorFunc;
     GLFW.monitorFunc = cbfun;
@@ -1283,90 +1313,79 @@ var LibraryGLFW = {
   },
 
   // TODO: implement
-  glfwGetVideoModes__sig: 'iii',
   glfwGetVideoModes: function(monitor, count) {
     {{{ makeSetValue('count', '0', '0', 'i32') }}};
     return 0;
   },
 
   // TODO: implement
-  glfwGetVideoMode__sig: 'ii',
   glfwGetVideoMode: function(monitor) { return 0; },
 
   // TODO: implement
-  glfwSetGamma__sig: 'vif',
   glfwSetGamma: function(monitor, gamma) { },
 
-  glfwGetGammaRamp__sig: 'ii',
   glfwGetGammaRamp: function(monitor) {
     throw "glfwGetGammaRamp not implemented.";
   },
 
-  glfwSetGammaRamp__sig: 'vii',
   glfwSetGammaRamp: function(monitor, ramp) {
     throw "glfwSetGammaRamp not implemented.";
   },
 
-  glfwDefaultWindowHints__sig: 'v',
   glfwDefaultWindowHints: function() {
     GLFW.hints = GLFW.defaultHints;
   },
 
-  glfwWindowHint__sig: 'vii',
   glfwWindowHint: function(target, hint) {
     GLFW.hints[target] = hint;
   },
 
-  glfwCreateWindow__sig: 'iiiiii',
+  glfwWindowHintString: function(hint, value) {
+    // from glfw docs -> we just ignore this.
+    // Some hints are platform specific.  These may be set on any platform but they
+    // will only affect their specific platform.  Other platforms will ignore them.
+  },
+
   glfwCreateWindow: function(width, height, title, monitor, share) {
     return GLFW.createWindow(width, height, title, monitor, share);
   },
 
-  glfwDestroyWindow__sig: 'vi',
   glfwDestroyWindow: function(winid) {
     return GLFW.destroyWindow(winid);
   },
 
-  glfwWindowShouldClose__sig: 'ii',
   glfwWindowShouldClose: function(winid) {
     var win = GLFW.WindowFromId(winid);
     if (!win) return 0;
     return win.shouldClose;
   },
 
-  glfwSetWindowShouldClose__sig: 'vii',
   glfwSetWindowShouldClose: function(winid, value) {
     var win = GLFW.WindowFromId(winid);
     if (!win) return;
     win.shouldClose = value;
   },
 
-  glfwSetWindowTitle__sig: 'vii',
   glfwSetWindowTitle: function(winid, title) {
     GLFW.setWindowTitle(winid, title);
   },
 
-  glfwGetWindowPos__sig: 'viii',
   glfwGetWindowPos: function(winid, x, y) {
     GLFW.getWindowPos(winid, x, y);
   },
 
-  glfwSetWindowPos__sig: 'viii',
   glfwSetWindowPos: function(winid, x, y) {
     GLFW.setWindowPos(winid, x, y);
   },
 
-  glfwGetWindowSize__sig: 'viii',
   glfwGetWindowSize: function(winid, width, height) {
     GLFW.getWindowSize(winid, width, height);
   },
 
-  glfwSetWindowSize__sig: 'viii',
   glfwSetWindowSize: function(winid, width, height) {
     GLFW.setWindowSize(winid, width, height);
   },
 
-  glfwGetFramebufferSize__sig: 'viii',
   glfwGetFramebufferSize: function(winid, width, height) {
     var ww = 0;
     var wh = 0;
@@ -1386,55 +1405,67 @@ var LibraryGLFW = {
     }
   },
 
-  glfwIconifyWindow__sig: 'vi',
+  glfwGetWindowContentScale: function(winid, x, y) {
+    // winid doesn't matter. all windows will use same scale anyway.
+    // hope i used this makeSetValue correctly
+    {{{ makeSetValue('x', '0', 'GLFW.scale', 'float') }}};
+    {{{ makeSetValue('y', '0', 'GLFW.scale', 'float') }}};
+  },
+
+  glfwGetWindowOpacity: function(winid) {
+    return 1.0;
+  },
+
+  glfwSetWindowOpacity: function(winid, opacity) {
+    // error
+  },
+
   glfwIconifyWindow: function(winid) {
 #if ASSERTIONS
     warnOnce('glfwIconifyWindow is not implemented');
 #endif
   },
 
-  glfwRestoreWindow__sig: 'vi',
   glfwRestoreWindow: function(winid) {
 #if ASSERTIONS
     warnOnce('glfwRestoreWindow is not implemented');
 #endif
   },
 
-  glfwShowWindow__sig: 'vi',
   glfwShowWindow: function(winid) {},
 
-  glfwHideWindow__sig: 'vi',
   glfwHideWindow: function(winid) {},
 
-  glfwGetWindowMonitor__sig: 'ii',
   glfwGetWindowMonitor: function(winid) {
     var win = GLFW.WindowFromId(winid);
     if (!win) return 0;
     return win.monitor;
   },
 
-  glfwGetWindowAttrib__sig: 'iii',
   glfwGetWindowAttrib: function(winid, attrib) {
     var win = GLFW.WindowFromId(winid);
     if (!win) return 0;
     return win.attributes[attrib];
   },
 
-  glfwSetWindowUserPointer__sig: 'vii',
+  glfwSetWindowAttrib: function(winid, attrib, value) {
+    var win = GLFW.WindowFromId(winid);
+    if (!win) return;
+    win.attributes[attrib] = value;
+  },
+
   glfwSetWindowUserPointer: function(winid, ptr) {
     var win = GLFW.WindowFromId(winid);
     if (!win) return;
     win.userptr = ptr;
   },
 
-  glfwGetWindowUserPointer__sig: 'ii',
   glfwGetWindowUserPointer: function(winid) {
     var win = GLFW.WindowFromId(winid);
     if (!win) return 0;
     return win.userptr;
   },
 
-  glfwSetWindowPosCallback__sig: 'iii',
   glfwSetWindowPosCallback: function(winid, cbfun) {
     var win = GLFW.WindowFromId(winid);
     if (!win) return null;
@@ -1443,22 +1474,18 @@ var LibraryGLFW = {
     return prevcbfun;
   },
 
-  glfwSetWindowSizeCallback__sig: 'iii',
   glfwSetWindowSizeCallback: function(winid, cbfun) {
     return GLFW.setWindowSizeCallback(winid, cbfun);
   },
 
-  glfwSetWindowCloseCallback__sig: 'iii',
   glfwSetWindowCloseCallback: function(winid, cbfun) {
     return GLFW.setWindowCloseCallback(winid, cbfun);
   },
 
-  glfwSetWindowRefreshCallback__sig: 'iii',
   glfwSetWindowRefreshCallback: function(winid, cbfun) {
     return GLFW.setWindowRefreshCallback(winid, cbfun);
   },
 
-  glfwSetWindowFocusCallback__sig: 'iii',
   glfwSetWindowFocusCallback: function(winid, cbfun) {
     var win = GLFW.WindowFromId(winid);
     if (!win) return null;
@@ -1467,7 +1494,6 @@ var LibraryGLFW = {
     return prevcbfun;
   },
 
-  glfwSetWindowIconifyCallback__sig: 'iii',
   glfwSetWindowIconifyCallback: function(winid, cbfun) {
     var win = GLFW.WindowFromId(winid);
     if (!win) return null;
@@ -1476,40 +1502,38 @@ var LibraryGLFW = {
     return prevcbfun;
   },
 
-  glfwSetWindowIcon__sig: 'viii',
+  glfwSetWindowMaximizeCallback: function(winid, cbfun) {
+    var win = GLFW.WindowFromId(winid);
+    if (!win) return null;
+    var prevcbfun = win.windowMaximizeFunc;
+    win.windowMaximizeFunc = cbfun;
+    return prevcbfun;
+  },
+
   glfwSetWindowIcon: function(winid, count, images) {},
 
-  glfwSetWindowSizeLimits__sig: 'viiiii',
   glfwSetWindowSizeLimits: function(winid, minwidth, minheight, maxwidth, maxheight) {},
 
-  glfwSetWindowAspectRatio__sig: 'viii',
   glfwSetWindowAspectRatio: function(winid, numer, denom) {},
 
-  glfwGetWindowFrameSize__sig: 'viiiii',
   glfwGetWindowFrameSize: function(winid, left, top, right, bottom) { throw "glfwGetWindowFrameSize not implemented."; },
 
-  glfwMaximizeWindow__sig: 'vi',
   glfwMaximizeWindow: function(winid) {},
 
-  glfwFocusWindow__sig: 'vi',
   glfwFocusWindow: function(winid) {},
 
-  glfwSetWindowMonitor__sig: 'viiiiiii',
+  glfwRequestWindowAttention: function(winid) {}, // maybe do window.focus()?
+
   glfwSetWindowMonitor: function(winid, monitor, xpos, ypos, width, height, refreshRate) { throw "glfwSetWindowMonitor not implemented."; },
 
-  glfwCreateCursor__sig: 'iiii',
   glfwCreateCursor: function(image, xhot, yhot) {},
 
-  glfwCreateStandardCursor__sig: 'ii',
   glfwCreateStandardCursor: function(shape) {},
 
-  glfwDestroyCursor__sig: 'vi',
   glfwDestroyCursor: function(cursor) {},
 
-  glfwSetCursor__sig: 'vii',
   glfwSetCursor: function(winid, cursor) {},
 
-  glfwSetFramebufferSizeCallback__sig: 'iii',
   glfwSetFramebufferSizeCallback: function(winid, cbfun) {
     var win = GLFW.WindowFromId(winid);
     if (!win) return null;
@@ -1518,7 +1542,14 @@ var LibraryGLFW = {
     return prevcbfun;
   },
 
-  glfwGetInputMode__sig: 'iii',
+  glfwSetWindowContentScaleCallback: function(winid, cbfun) {
+    var win = GLFW.WindowFromId(winid);
+    if (!win) return null;
+    var prevcbfun = win.windowContentScaleFunc;
+    win.windowContentScaleFunc = cbfun;
+    return prevcbfun;
+  },
+
   glfwGetInputMode: function(winid, mode) {
     var win = GLFW.WindowFromId(winid);
     if (!win) return;
@@ -1536,59 +1567,53 @@ var LibraryGLFW = {
     return win.inputModes[mode];
   },
 
-  glfwSetInputMode__sig: 'viii',
   glfwSetInputMode: function(winid, mode, value) {
     GLFW.setInputMode(winid, mode, value);
   },
 
-  glfwGetKey__sig: 'iii',
+  glfwRawMouseMotionSupported: function() {
+    return 0;
+  },
+
   glfwGetKey: function(winid, key) {
     return GLFW.getKey(winid, key);
   },
 
-  glfwGetKeyName__sig: 'iii',
   glfwGetKeyName: function(key, scancode) { throw "glfwGetKeyName not implemented."; },
 
-  glfwGetMouseButton__sig: 'iii',
+  glfwGetKeyScancode: function(key) { throw "glfwGetKeyScancode not implemented."; },
+
   glfwGetMouseButton: function(winid, button) {
     return GLFW.getMouseButton(winid, button);
   },
 
-  glfwGetCursorPos__sig: 'viii',
   glfwGetCursorPos: function(winid, x, y) {
     GLFW.getCursorPos(winid, x, y);
   },
 
   // I believe it is not possible to move the mouse with javascript
-  glfwSetCursorPos__sig: 'vidd',
   glfwSetCursorPos: function(winid, x, y) {
     GLFW.setCursorPos(winid, x, y);
   },
 
-  glfwSetKeyCallback__sig: 'iii',
   glfwSetKeyCallback: function(winid, cbfun) {
     return GLFW.setKeyCallback(winid, cbfun);
   },
 
-  glfwSetCharCallback__sig: 'iii',
   glfwSetCharCallback: function(winid, cbfun) {
     return GLFW.setCharCallback(winid, cbfun);
   },
 
-  glfwSetCharModsCallback__sig: 'iii',
   glfwSetCharModsCallback: function(winid, cbfun) { throw "glfwSetCharModsCallback not implemented."; },
 
-  glfwSetMouseButtonCallback__sig: 'iii',
   glfwSetMouseButtonCallback: function(winid, cbfun) {
     return GLFW.setMouseButtonCallback(winid, cbfun);
   },
 
-  glfwSetCursorPosCallback__sig: 'iii',
   glfwSetCursorPosCallback: function(winid, cbfun) {
     return GLFW.setCursorPosCallback(winid, cbfun);
   },
 
-  glfwSetCursorEnterCallback__sig: 'iii',
   glfwSetCursorEnterCallback: function(winid, cbfun) {
     var win = GLFW.WindowFromId(winid);
     if (!win) return null;
@@ -1597,47 +1622,30 @@ var LibraryGLFW = {
     return prevcbfun;
   },
 
-  glfwSetScrollCallback__sig: 'iii',
   glfwSetScrollCallback: function(winid, cbfun) {
     return GLFW.setScrollCallback(winid, cbfun);
   },
 
-  glfwVulkanSupported__sig: 'i',
   glfwVulkanSupported: function() {
     return 0;
   },
 
-  glfwSetDropCallback__sig: 'iii',
   glfwSetDropCallback: function(winid, cbfun) {
     return GLFW.setDropCallback(winid, cbfun);
   },
 
-  glfwGetTimerValue__sig: 'j',
   glfwGetTimerValue: function() { throw "glfwGetTimerValue is not implemented."; },
 
-  glfwGetTimerFrequency__sig: 'j',
   glfwGetTimerFrequency: function() { throw "glfwGetTimerFrequency is not implemented."; },
 
-  glfwGetRequiredInstanceExtensions__sig: 'ii',
   glfwGetRequiredInstanceExtensions: function(count) { throw "glfwGetRequiredInstanceExtensions is not implemented."; },
 
-  glfwGetInstanceProcAddress__sig: 'iii',
-  glfwGetInstanceProcAddress: function(instance, procname) { throw "glfwGetInstanceProcAddress is not implemented."; },
-
-  glfwGetPhysicalDevicePresentationSupport__sig: 'iiii',
-  glfwGetPhysicalDevicePresentationSupport: function(instance, device, queuefamily) { throw "glfwGetPhysicalDevicePresentationSupport is not implemented"; },
-
-  glfwCreateWindowSurface__sig: 'iiiii',
-  glfwCreateWindowSurface: function(instance, winid, allocator, surface) { throw "glfwCreateWindowSurface is not implemented."; },
-
-  glfwJoystickPresent__sig: 'ii',
   glfwJoystickPresent: function(joy) {
     GLFW.refreshJoysticks();
 
     return GLFW.joys[joy] !== undefined;
   },
 
-  glfwGetJoystickAxes__sig: 'iii',
   glfwGetJoystickAxes: function(joy, count) {
     GLFW.refreshJoysticks();
 
@@ -1651,7 +1659,6 @@ var LibraryGLFW = {
     return state.axes;
   },
 
-  glfwGetJoystickButtons__sig: 'iii',
   glfwGetJoystickButtons: function(joy, count) {
     GLFW.refreshJoysticks();
 
@@ -1665,7 +1672,10 @@ var LibraryGLFW = {
     return state.buttons;
   },
 
-  glfwGetJoystickName__sig: 'ii',
+  glfwGetJoystickHats: function(joy, count) {
+    throw "glfwGetJoystickHats is not implemented";
+  },
+
   glfwGetJoystickName: function(joy) {
     if (GLFW.joys[joy]) {
       return GLFW.joys[joy].id;
@@ -1673,33 +1683,41 @@ var LibraryGLFW = {
     return 0;
   },
 
-  glfwSetJoystickCallback__sig: 'ii',
+  glfwGetJoystickGUID: function(jid) {
+    throw "glfwGetJoystickGUID not implemented";
+  },
+
+  glfwSetJoystickUserPointer: function(jid, ptr) {
+    throw "glfwSetJoystickUserPointer not implemented";
+  },
+
+  glfwGetJoystickUserPointer: function(jid) {
+    throw "glfwGetJoystickUserPointer not implemented";
+  },
+
+  glfwJoystickIsGamepad: function(jid) {
+    throw "glfwJoystickIsGamepad not implemented";
+  },
+
   glfwSetJoystickCallback: function(cbfun) {
     GLFW.setJoystickCallback(cbfun);
   },
 
-  glfwSetClipboardString__sig: 'vii',
   glfwSetClipboardString: function(win, string) {},
 
-  glfwGetClipboardString__sig: 'ii',
   glfwGetClipboardString: function(win) {},
 
-  glfwMakeContextCurrent__sig: 'vi',
   glfwMakeContextCurrent: function(winid) {},
 
-  glfwGetCurrentContext__sig: 'i',
   glfwGetCurrentContext: function() {
     return GLFW.active ? GLFW.active.id : 0;
   },
 
-  glfwSwapBuffers__sig: 'vi',
   glfwSwapBuffers: function(winid) {
     GLFW.swapBuffers(winid);
   },
 
-#endif // GLFW 3
-
-#if USE_GLFW == 2
+#elif USE_GLFW == 2
   glfwOpenWindow: function(width, height, redbits, greenbits, bluebits, alphabits, depthbits, stencilbits, mode) {
     GLFW.hints[0x00021001] = redbits;     // GLFW_RED_BITS
     GLFW.hints[0x00021002] = greenbits;   // GLFW_GREEN_BITS
@@ -1720,39 +1738,35 @@ var LibraryGLFW = {
     GLFW.hints[target] = hint;
   },
 
-  glfwGetWindowSize: function(width, height) {
+  glfwGetWindowSize_v2: function(width, height) {
     GLFW.getWindowSize(GLFW.active.id, width, height);
   },
 
-  glfwSetWindowSize: function(width, height) {
+  glfwSetWindowSize_v2: function(width, height) {
     GLFW.setWindowSize(GLFW.active.id, width, height);
   },
 
-  glfwGetWindowPos: function(x, y) {
-    GLFW.getWindowPos(GLFW.active.id, x, y);
-  },
-
-  glfwSetWindowPos: function(x, y) {
+  glfwSetWindowPos_v2: function(x, y) {
     GLFW.setWindowPos(GLFW.active.id, x, y);
   },
 
-  glfwSetWindowTitle: function(title) {
+  glfwSetWindowTitle_v2: function(title) {
     GLFW.setWindowTitle(GLFW.active.id, title);
   },
 
-  glfwIconifyWindow: function() {
+  glfwIconifyWindow_v2: function() {
 #if ASSERTIONS
     warnOnce('glfwIconifyWindow is not implemented');
 #endif
   },
 
-  glfwRestoreWindow: function() {
+  glfwRestoreWindow_v2: function() {
 #if ASSERTIONS
     warnOnce('glfwRestoreWindow is not implemented');
 #endif
   },
 
-  glfwSwapBuffers: function() {
+  glfwSwapBuffers_v2: function() {
     GLFW.swapBuffers(GLFW.active.id);
   },
 
@@ -1761,23 +1775,23 @@ var LibraryGLFW = {
     return GLFW.hints[param];
   },
 
-  glfwSetWindowSizeCallback: function(cbfun) {
+  glfwSetWindowSizeCallback_v2: function(cbfun) {
     GLFW.setWindowSizeCallback(GLFW.active.id, cbfun);
   },
 
-  glfwSetWindowCloseCallback: function(cbfun) {
+  glfwSetWindowCloseCallback_v2: function(cbfun) {
     GLFW.setWindowCloseCallback(GLFW.active.id, cbfun);
   },
 
-  glfwSetWindowRefreshCallback: function(cbfun) {
+  glfwSetWindowRefreshCallback_v2: function(cbfun) {
     GLFW.setWindowRefreshCallback(GLFW.active.id, cbfun);
   },
 
-  glfwGetKey: function(key) {
+  glfwGetKey_v2: function(key) {
     return GLFW.getKey(GLFW.active.id, key);
   },
 
-  glfwGetMouseButton: function(button) {
+  glfwGetMouseButton_v2: function(button) {
     return GLFW.getMouseButton(GLFW.active.id, button);
   },
 
@@ -1796,15 +1810,15 @@ var LibraryGLFW = {
   glfwSetMouseWheel: function(pos) {
   },
 
-  glfwSetKeyCallback: function(cbfun) {
+  glfwSetKeyCallback_v2: function(cbfun) {
     GLFW.setKeyCallback(GLFW.active.id, cbfun);
   },
 
-  glfwSetCharCallback: function(cbfun) {
+  glfwSetCharCallback_v2: function(cbfun) {
     GLFW.setCharCallback(GLFW.active.id, cbfun);
   },
 
-  glfwSetMouseButtonCallback: function(cbfun) {
+  glfwSetMouseButtonCallback_v2: function(cbfun) {
     GLFW.setMouseButtonCallback(GLFW.active.id, cbfun);
   },
 
