@@ -11,22 +11,22 @@ mergeInto(LibraryManager.library, {
     isWindows: false,
     staticInit: () => {
       NODEFS.isWindows = !!process.platform.match(/^win/);
-      var flags = process["binding"]("constants");
+      var flags = process.binding("constants");
       // Node.js 4 compatibility: it has no namespaces for constants
       if (flags["fs"]) {
         flags = flags["fs"];
       }
       NODEFS.flagsForNodeMap = {
-        "{{{ cDefine('O_APPEND') }}}": flags["O_APPEND"],
-        "{{{ cDefine('O_CREAT') }}}": flags["O_CREAT"],
-        "{{{ cDefine('O_EXCL') }}}": flags["O_EXCL"],
-        "{{{ cDefine('O_NOCTTY') }}}": flags["O_NOCTTY"],
-        "{{{ cDefine('O_RDONLY') }}}": flags["O_RDONLY"],
-        "{{{ cDefine('O_RDWR') }}}": flags["O_RDWR"],
-        "{{{ cDefine('O_DSYNC') }}}": flags["O_SYNC"],
-        "{{{ cDefine('O_TRUNC') }}}": flags["O_TRUNC"],
-        "{{{ cDefine('O_WRONLY') }}}": flags["O_WRONLY"],
-        "{{{ cDefine('O_NOFOLLOW') }}}": flags["O_NOFOLLOW"],
+        "{{{ cDefs.O_APPEND }}}": flags["O_APPEND"],
+        "{{{ cDefs.O_CREAT }}}": flags["O_CREAT"],
+        "{{{ cDefs.O_EXCL }}}": flags["O_EXCL"],
+        "{{{ cDefs.O_NOCTTY }}}": flags["O_NOCTTY"],
+        "{{{ cDefs.O_RDONLY }}}": flags["O_RDONLY"],
+        "{{{ cDefs.O_RDWR }}}": flags["O_RDWR"],
+        "{{{ cDefs.O_DSYNC }}}": flags["O_SYNC"],
+        "{{{ cDefs.O_TRUNC }}}": flags["O_TRUNC"],
+        "{{{ cDefs.O_WRONLY }}}": flags["O_WRONLY"],
+        "{{{ cDefs.O_NOFOLLOW }}}": flags["O_NOFOLLOW"],
       };
 #if ASSERTIONS
       // The 0 define must match on both sides, as otherwise we would not
@@ -49,7 +49,7 @@ mergeInto(LibraryManager.library, {
     },
     createNode: (parent, name, mode, dev) => {
       if (!FS.isDir(mode) && !FS.isFile(mode) && !FS.isLink(mode)) {
-        throw new FS.ErrnoError({{{ cDefine('EINVAL') }}});
+        throw new FS.ErrnoError({{{ cDefs.EINVAL }}});
       }
       var node = FS.createNode(parent, name, mode);
       node.node_ops = NODEFS.node_ops;
@@ -84,11 +84,11 @@ mergeInto(LibraryManager.library, {
     // This maps the integer permission modes from http://linux.die.net/man/3/open
     // to node.js-specific file open permission strings at http://nodejs.org/api/fs.html#fs_fs_open_path_flags_mode_callback
     flagsForNode: (flags) => {
-      flags &= ~{{{ cDefine('O_PATH') }}}; // Ignore this flag from musl, otherwise node.js fails to open the file.
-      flags &= ~{{{ cDefine('O_NONBLOCK') }}}; // Ignore this flag from musl, otherwise node.js fails to open the file.
-      flags &= ~{{{ cDefine('O_LARGEFILE') }}}; // Ignore this flag from musl, otherwise node.js fails to open the file.
-      flags &= ~{{{ cDefine('O_CLOEXEC') }}}; // Some applications may pass it; it makes no sense for a single process.
-      flags &= ~{{{ cDefine('O_DIRECTORY') }}}; // Node.js doesn't need this passed in, it errors.
+      flags &= ~{{{ cDefs.O_PATH }}}; // Ignore this flag from musl, otherwise node.js fails to open the file.
+      flags &= ~{{{ cDefs.O_NONBLOCK }}}; // Ignore this flag from musl, otherwise node.js fails to open the file.
+      flags &= ~{{{ cDefs.O_LARGEFILE }}}; // Ignore this flag from musl, otherwise node.js fails to open the file.
+      flags &= ~{{{ cDefs.O_CLOEXEC }}}; // Some applications may pass it; it makes no sense for a single process.
+      flags &= ~{{{ cDefs.O_DIRECTORY }}}; // Node.js doesn't need this passed in, it errors.
       var newFlags = 0;
       for (var k in NODEFS.flagsForNodeMap) {
         if (flags & k) {
@@ -97,7 +97,7 @@ mergeInto(LibraryManager.library, {
         }
       }
       if (flags) {
-        throw new FS.ErrnoError({{{ cDefine('EINVAL') }}});
+        throw new FS.ErrnoError({{{ cDefs.EINVAL }}});
       }
       return newFlags;
     },
@@ -233,7 +233,7 @@ mergeInto(LibraryManager.library, {
           if (!e.code) throw e;
           // node under windows can return code 'UNKNOWN' here:
           // https://github.com/emscripten-core/emscripten/issues/15468
-          if (e.code === 'UNKNOWN') throw new FS.ErrnoError({{{ cDefine('EINVAL') }}});
+          if (e.code === 'UNKNOWN') throw new FS.ErrnoError({{{ cDefs.EINVAL }}});
           throw new FS.ErrnoError(NODEFS.convertNodeCode(e));
         }
       },
@@ -278,9 +278,9 @@ mergeInto(LibraryManager.library, {
       },
       llseek: (stream, offset, whence) => {
         var position = offset;
-        if (whence === {{{ cDefine('SEEK_CUR') }}}) {
+        if (whence === {{{ cDefs.SEEK_CUR }}}) {
           position += stream.position;
-        } else if (whence === {{{ cDefine('SEEK_END') }}}) {
+        } else if (whence === {{{ cDefs.SEEK_END }}}) {
           if (FS.isFile(stream.node.mode)) {
             try {
               var stat = fs.fstatSync(stream.nfd);
@@ -292,14 +292,14 @@ mergeInto(LibraryManager.library, {
         }
 
         if (position < 0) {
-          throw new FS.ErrnoError({{{ cDefine('EINVAL') }}});
+          throw new FS.ErrnoError({{{ cDefs.EINVAL }}});
         }
 
         return position;
       },
       mmap: (stream, length, position, prot, flags) => {
         if (!FS.isFile(stream.node.mode)) {
-          throw new FS.ErrnoError({{{ cDefine('ENODEV') }}});
+          throw new FS.ErrnoError({{{ cDefs.ENODEV }}});
         }
 
         var ptr = mmapAlloc(length);
