@@ -16,17 +16,16 @@ var WasiLibrary = {
 #endif
 
   proc_exit__nothrow: true,
-  proc_exit__sig: 'vi',
   proc_exit: function(code) {
 #if MINIMAL_RUNTIME
     throw 'exit(' + code + ')';
 #else
 #if RUNTIME_DEBUG
-    err('proc_exit: ' + code);
+    dbg('proc_exit: ' + code);
 #endif
     EXITSTATUS = code;
     if (!keepRuntimeAlive()) {
-#if USE_PTHREADS
+#if PTHREADS
       PThread.terminateAllThreads();
 #endif
 #if expectToReceiveOnModule('onExit')
@@ -35,7 +34,18 @@ var WasiLibrary = {
       ABORT = true;
     }
     quit_(code, new ExitStatus(code));
-#endif
+#endif // MINIMAL_RUNTIME
+  },
+
+  sched_yield__nothrow: true,
+  sched_yield: function() {
+    return 0;
+  },
+
+  random_get__deps: ['getentropy'],
+  random_get: function(buf, buf_len) {
+    _getentropy(buf, buf_len);
+    return 0;
   },
 
   $getEnvStrings__deps: ['$ENV', '$getExecutableName'],
@@ -50,6 +60,7 @@ var WasiLibrary = {
       var lang = 'C.UTF-8';
 #endif
       var env = {
+#if !PURE_WASI
         'USER': 'web_user',
         'LOGNAME': 'web_user',
         'PATH': '/',
@@ -57,6 +68,7 @@ var WasiLibrary = {
         'HOME': '/home/web_user',
         'LANG': lang,
         '_': getExecutableName()
+#endif
       };
       // Apply the user-provided values, if any.
       for (var x in ENV) {
@@ -77,7 +89,6 @@ var WasiLibrary = {
 
   environ_sizes_get__deps: ['$getEnvStrings'],
   environ_sizes_get__nothrow: true,
-  environ_sizes_get__sig: 'ipp',
   environ_sizes_get: function(penviron_count, penviron_buf_size) {
     var strings = getEnvStrings();
     {{{ makeSetValue('penviron_count', 0, 'strings.length', SIZE_TYPE) }}};
@@ -89,15 +100,14 @@ var WasiLibrary = {
     return 0;
   },
 
-  environ_get__deps: ['$getEnvStrings', '$writeAsciiToMemory'],
+  environ_get__deps: ['$getEnvStrings', '$stringToAscii'],
   environ_get__nothrow: true,
-  environ_get__sig: 'ipp',
   environ_get: function(__environ, environ_buf) {
     var bufSize = 0;
     getEnvStrings().forEach(function(string, i) {
       var ptr = environ_buf + bufSize;
-      {{{ makeSetValue('__environ', `i*${Runtime.POINTER_SIZE}`, 'ptr', POINTER_TYPE) }}};
-      writeAsciiToMemory(string, ptr);
+      {{{ makeSetValue('__environ', `i*${POINTER_SIZE}`, 'ptr', POINTER_TYPE) }}};
+      stringToAscii(string, ptr);
       bufSize += string.length + 1;
     });
     return 0;
@@ -107,7 +117,6 @@ var WasiLibrary = {
   // to main, and the `mainArgs` global does not exist.
 #if STANDALONE_WASM
   args_sizes_get__nothrow: true,
-  args_sizes_get__sig: 'ipp',
   args_sizes_get: function(pargc, pargv_buf_size) {
 #if MAIN_READS_PARAMS
     {{{ makeSetValue('pargc', 0, 'mainArgs.length', SIZE_TYPE) }}};
@@ -123,15 +132,14 @@ var WasiLibrary = {
   },
 
   args_get__nothrow: true,
-  args_get__sig: 'ipp',
-  args_get__deps: ['$writeAsciiToMemory'],
+  args_get__deps: ['$stringToAscii'],
   args_get: function(argv, argv_buf) {
 #if MAIN_READS_PARAMS
     var bufSize = 0;
     mainArgs.forEach(function(arg, i) {
       var ptr = argv_buf + bufSize;
-      {{{ makeSetValue('argv', `i*${Runtime.POINTER_SIZE}`, 'ptr', POINTER_TYPE) }}};
-      writeAsciiToMemory(arg, ptr);
+      {{{ makeSetValue('argv', `i*${POINTER_SIZE}`, 'ptr', POINTER_TYPE) }}};
+      stringToAscii(arg, ptr);
       bufSize += arg.length + 1;
     });
 #endif
@@ -140,10 +148,10 @@ var WasiLibrary = {
 #endif
 
   $checkWasiClock: function(clock_id) {
-    return clock_id == {{{ cDefine('__WASI_CLOCKID_REALTIME') }}} ||
-           clock_id == {{{ cDefine('__WASI_CLOCKID_MONOTONIC') }}} ||
-           clock_id == {{{ cDefine('__WASI_CLOCKID_PROCESS_CPUTIME_ID') }}} ||
-           clock_id == {{{ cDefine('__WASI_CLOCKID_THREAD_CPUTIME_ID') }}};
+    return clock_id == {{{ cDefs.__WASI_CLOCKID_REALTIME }}} ||
+           clock_id == {{{ cDefs.__WASI_CLOCKID_MONOTONIC }}} ||
+           clock_id == {{{ cDefs.__WASI_CLOCKID_PROCESS_CPUTIME_ID }}} ||
+           clock_id == {{{ cDefs.__WASI_CLOCKID_THREAD_CPUTIME_ID }}};
   },
 
   // TODO: the i64 in the API here must be legalized for this JS code to run,
@@ -151,20 +159,19 @@ var WasiLibrary = {
   // this is needed. To get this code to be usable as a JS shim we need to
   // either wait for BigInt support or to legalize on the client.
   clock_time_get__nothrow: true,
-  clock_time_get__sig: 'iijp',
   clock_time_get__deps: ['emscripten_get_now', '$nowIsMonotonic', '$checkWasiClock'],
   clock_time_get: function(clk_id, {{{ defineI64Param('ignored_precision') }}}, ptime) {
     if (!checkWasiClock(clk_id)) {
-      return {{{ cDefine('EINVAL') }}};
+      return {{{ cDefs.EINVAL }}};
     }
     var now;
     // all wasi clocks but realtime are monotonic
-    if (clk_id === {{{ cDefine('__WASI_CLOCKID_REALTIME') }}}) {
+    if (clk_id === {{{ cDefs.__WASI_CLOCKID_REALTIME }}}) {
       now = Date.now();
     } else if (nowIsMonotonic) {
       now = _emscripten_get_now();
     } else {
-      return {{{ cDefine('ENOSYS') }}};
+      return {{{ cDefs.ENOSYS }}};
     }
     // "now" is in ms, and wasi times are in ns.
     var nsec = Math.round(now * 1000 * 1000);
@@ -174,20 +181,19 @@ var WasiLibrary = {
   },
 
   clock_res_get__nothrow: true,
-  clock_res_get__sig: 'iip',
   clock_res_get__deps: ['emscripten_get_now', 'emscripten_get_now_res', '$nowIsMonotonic', '$checkWasiClock'],
   clock_res_get: function(clk_id, pres) {
     if (!checkWasiClock(clk_id)) {
-      return {{{ cDefine('EINVAL') }}};
+      return {{{ cDefs.EINVAL }}};
     }
     var nsec;
     // all wasi clocks but realtime are monotonic
-    if (clk_id === {{{ cDefine('CLOCK_REALTIME') }}}) {
+    if (clk_id === {{{ cDefs.CLOCK_REALTIME }}}) {
       nsec = 1000 * 1000; // educated guess that it's milliseconds
     } else if (nowIsMonotonic) {
       nsec = _emscripten_get_now_res();
     } else {
-      return {{{ cDefine('ENOSYS') }}};
+      return {{{ cDefs.ENOSYS }}};
     }
     {{{ makeSetValue('pres', 0, 'nsec >>> 0', 'i32') }}};
     {{{ makeSetValue('pres', 4, '(nsec / Math.pow(2, 32)) >>> 0', 'i32') }}};
@@ -206,6 +212,9 @@ var WasiLibrary = {
       if (curr < 0) return -1;
       ret += curr;
       if (curr < len) break; // nothing more to read
+      if (typeof offset !== 'undefined') {
+        offset += curr;
+      }
     }
     return ret;
   },
@@ -219,6 +228,9 @@ var WasiLibrary = {
       var curr = FS.write(stream, {{{ heapAndOffset('HEAP8', 'ptr') }}}, len, offset);
       if (curr < 0) return -1;
       ret += curr;
+      if (typeof offset !== 'undefined') {
+        offset += curr;
+      }
     }
     return ret;
   },
@@ -227,7 +239,7 @@ var WasiLibrary = {
   $printCharBuffers: [null, [], []], // 1 => stdout, 2 => stderr
   $printCharBuffers__internal: true,
   $printChar__internal: true,
-  $printChar__deps: ['$printCharBuffers'],
+  $printChar__deps: ['$printCharBuffers', '$UTF8ArrayToString'],
   $printChar: function(stream, curr) {
     var buffer = printCharBuffers[stream];
 #if ASSERTIONS
@@ -261,7 +273,6 @@ var WasiLibrary = {
 #else
   fd_write__deps: ['$printChar'],
 #endif
-  fd_write__sig: 'iippp',
   fd_write: function(fd, iov, iovcnt, pnum) {
 #if SYSCALLS_REQUIRE_FILESYSTEM
     var stream = SYSCALLS.getStreamFromFD(fd);
@@ -290,19 +301,18 @@ var WasiLibrary = {
   ].concat(i53ConversionDeps),
   fd_pwrite: function(fd, iov, iovcnt, {{{ defineI64Param('offset') }}}, pnum) {
 #if SYSCALLS_REQUIRE_FILESYSTEM
-    {{{ receiveI64ParamAsI53('offset', cDefine('EOVERFLOW')) }}}
+    {{{ receiveI64ParamAsI53('offset', cDefs.EOVERFLOW) }}}
     var stream = SYSCALLS.getStreamFromFD(fd)
     var num = doWritev(stream, iov, iovcnt, offset);
-    {{{ makeSetValue('pnum', 0, 'num', 'i32') }}};
+    {{{ makeSetValue('pnum', 0, 'num', SIZE_TYPE) }}};
     return 0;
 #elif ASSERTIONS
     abort('fd_pwrite called without SYSCALLS_REQUIRE_FILESYSTEM');
 #else
-    return {{{ cDefine('ENOSYS') }}};
+    return {{{ cDefs.ENOSYS }}};
 #endif
   },
 
-  fd_close__sig: 'ii',
   fd_close: function(fd) {
 #if SYSCALLS_REQUIRE_FILESYSTEM
     var stream = SYSCALLS.getStreamFromFD(fd);
@@ -318,11 +328,10 @@ var WasiLibrary = {
 #elif ASSERTIONS
     abort('fd_close called without SYSCALLS_REQUIRE_FILESYSTEM');
 #else
-    return {{{ cDefine('ENOSYS') }}};
+    return {{{ cDefs.ENOSYS }}};
 #endif // SYSCALLS_REQUIRE_FILESYSTEM
   },
 
-  fd_read__sig: 'iippp',
 #if SYSCALLS_REQUIRE_FILESYSTEM
   fd_read__deps: ['$doReadv'],
 #endif
@@ -330,12 +339,12 @@ var WasiLibrary = {
 #if SYSCALLS_REQUIRE_FILESYSTEM
     var stream = SYSCALLS.getStreamFromFD(fd);
     var num = doReadv(stream, iov, iovcnt);
-    {{{ makeSetValue('pnum', 0, 'num', 'i32') }}};
+    {{{ makeSetValue('pnum', 0, 'num', SIZE_TYPE) }}};
     return 0;
 #elif ASSERTIONS
     abort('fd_read called without SYSCALLS_REQUIRE_FILESYSTEM');
 #else
-    return {{{ cDefine('ENOSYS') }}};
+    return {{{ cDefs.ENOSYS }}};
 #endif // SYSCALLS_REQUIRE_FILESYSTEM
   },
 
@@ -344,58 +353,189 @@ var WasiLibrary = {
     '$doReadv',
 #endif
   ].concat(i53ConversionDeps),
-  fd_pread__sig: 'iippjp',
   fd_pread: function(fd, iov, iovcnt, {{{ defineI64Param('offset') }}}, pnum) {
 #if SYSCALLS_REQUIRE_FILESYSTEM
-    {{{ receiveI64ParamAsI53('offset', cDefine('EOVERFLOW')) }}}
+    {{{ receiveI64ParamAsI53('offset', cDefs.EOVERFLOW) }}}
     var stream = SYSCALLS.getStreamFromFD(fd)
     var num = doReadv(stream, iov, iovcnt, offset);
-    {{{ makeSetValue('pnum', 0, 'num', 'i32') }}};
+    {{{ makeSetValue('pnum', 0, 'num', SIZE_TYPE) }}};
     return 0;
 #elif ASSERTIONS
     abort('fd_pread called without SYSCALLS_REQUIRE_FILESYSTEM');
 #else
-    return {{{ cDefine('ENOSYS') }}};
+    return {{{ cDefs.ENOSYS }}};
 #endif
   },
 
-  fd_seek__sig: 'iijip',
   fd_seek__deps: i53ConversionDeps,
   fd_seek: function(fd, {{{ defineI64Param('offset') }}}, whence, newOffset) {
 #if SYSCALLS_REQUIRE_FILESYSTEM
-    {{{ receiveI64ParamAsI53('offset', cDefine('EOVERFLOW')) }}}
+    {{{ receiveI64ParamAsI53('offset', cDefs.EOVERFLOW) }}}
     var stream = SYSCALLS.getStreamFromFD(fd);
     FS.llseek(stream, offset, whence);
     {{{ makeSetValue('newOffset', '0', 'stream.position', 'i64') }}};
-    if (stream.getdents && offset === 0 && whence === {{{ cDefine('SEEK_SET') }}}) stream.getdents = null; // reset readdir state
+    if (stream.getdents && offset === 0 && whence === {{{ cDefs.SEEK_SET }}}) stream.getdents = null; // reset readdir state
     return 0;
 #else
-    return {{{ cDefine('ESPIPE') }}};
+    return {{{ cDefs.ESPIPE }}};
 #endif
   },
 
-  fd_fdstat_get__sig: 'iip',
-  fd_fdstat_get: function(fd, pbuf) {
-#if SYSCALLS_REQUIRE_FILESYSTEM
+  $wasiRightsToMuslOFlags: function(rights) {
+#if SYSCALL_DEBUG
+    dbg('wasiRightsToMuslOFlags: ' + rights);
+#endif
+    if ((rights & {{{ cDefs.__WASI_RIGHTS_FD_READ }}}) && (rights & {{{ cDefs.__WASI_RIGHTS_FD_WRITE }}})) {
+      return {{{ cDefs.O_RDWR }}};
+    }
+    if (rights & {{{ cDefs.__WASI_RIGHTS_FD_READ }}}) {
+      return {{{ cDefs.O_RDONLY }}};
+    }
+    if (rights & {{{ cDefs.__WASI_RIGHTS_FD_WRITE }}}) {
+      return {{{ cDefs.O_WRONLY }}};
+    }
+    throw new FS.ErrnoError({{{ cDefs.EINVAL }}});
+  },
+
+  $wasiOFlagsToMuslOFlags: function(oflags) {
+    var musl_oflags = 0;
+    if (oflags & {{{ cDefs.__WASI_OFLAGS_CREAT }}}) {
+      musl_oflags |= {{{ cDefs.O_CREAT }}};
+    }
+    if (oflags & {{{ cDefs.__WASI_OFLAGS_TRUNC }}}) {
+      musl_oflags |= {{{ cDefs.O_TRUNC }}};
+    }
+    if (oflags & {{{ cDefs.__WASI_OFLAGS_DIRECTORY }}}) {
+      musl_oflags |= {{{ cDefs.O_DIRECTORY }}};
+    }
+    if (oflags & {{{ cDefs.__WASI_OFLAGS_EXCL }}}) {
+      musl_oflags |= {{{ cDefs.O_EXCL }}};
+    }
+    return musl_oflags;
+  },
+
+#if PURE_WASI
+  // preopen maps open file descriptors to pathname.
+  // In emscripten we already have a VFS layer so (for now) we expose the entire
+  // VFS to the wasi API.
+  $preopens: "{3: '/'}",
+
+  path_open__sig: 'iiiiiiiiii',
+  path_open__deps: ['$wasiRightsToMuslOFlags', '$wasiOFlagsToMuslOFlags', '$preopens'],
+  path_open: function(fd, dirflags, path, path_len, oflags,
+                      fs_rights_base, fs_rights_inherting,
+                      fdflags, opened_fd) {
+    if (!(fd in preopens)) {
+      return {{{ cDefs.EBADF }}};
+    }
+    var pathname = UTF8ToString(path, path_len);
+    var musl_oflags = wasiRightsToMuslOFlags(Number(fs_rights_base));
+#if SYSCALL_DEBUG
+    dbg("oflags1: 0x" + musl_oflags.toString(16));
+#endif
+    musl_oflags |= wasiOFlagsToMuslOFlags(Number(oflags));
+#if SYSCALL_DEBUG
+    dbg("oflags2: 0x" + musl_oflags.toString(16));
+#endif
+    var stream = FS.open(pathname, musl_oflags);
+    {{{ makeSetValue('opened_fd', '0', 'stream.fd', 'i32') }}};
+    return 0;
+  },
+
+  fd_prestat_dir_name__deps: ['$preopens'],
+  fd_prestat_dir_name__sig: 'iiii',
+  fd_prestat_dir_name__nothrow: true,
+  fd_prestat_dir_name: function(fd, path, path_len) {
+    if (!(fd in preopens)) {
+      return {{{ cDefs.EBADF }}};
+    }
+    var preopen_path = preopens[fd];
+    stringToUTF8Array(preopen_path, HEAP8, path, path_len)
+#if SYSCALL_DEBUG
+    dbg('fd_prestat_dir_name -> "' + preopen_path + '"');
+#endif
+    return 0;
+  },
+
+  fd_prestat_get__deps: ['$preopens'],
+  fd_prestat_get__sig: 'iii',
+  fd_prestat_get__nothrow: true,
+  fd_prestat_get: function(fd, stat_buf) {
+    if (!(fd in preopens)) {
+      return {{{ cDefs.EBADF }}};
+    }
+    var preopen = preopens[fd];
+    {{{ makeSetValue('stat_buf', C_STRUCTS.__wasi_prestat_t.pr_type, cDefs.__WASI_PREOPENTYPE_DIR, 'i8') }}};
+    {{{ makeSetValue('stat_buf', C_STRUCTS.__wasi_prestat_t.u + C_STRUCTS.__wasi_prestat_dir_t.pr_name_len, 'preopen.length', 'i64') }}};
+    return 0;
+  },
+
+  fd_fdstat_set_flags__sig: 'iii',
+  fd_fdstat_set_flags: function(fd, flags) {
+    // TODO(sbc): implement
     var stream = SYSCALLS.getStreamFromFD(fd);
-    // All character devices are terminals (other things a Linux system would
-    // assume is a character device, like the mouse, we have special APIs for).
-    var type = stream.tty ? {{{ cDefine('__WASI_FILETYPE_CHARACTER_DEVICE') }}} :
-               FS.isDir(stream.mode) ? {{{ cDefine('__WASI_FILETYPE_DIRECTORY') }}} :
-               FS.isLink(stream.mode) ? {{{ cDefine('__WASI_FILETYPE_SYMBOLIC_LINK') }}} :
-               {{{ cDefine('__WASI_FILETYPE_REGULAR_FILE') }}};
-#else
-    // hack to support printf in SYSCALLS_REQUIRE_FILESYSTEM=0
-    var type = fd == 1 || fd == 2 ? {{{ cDefine('__WASI_FILETYPE_CHARACTER_DEVICE') }}} : abort();
-#endif
-    {{{ makeSetValue('pbuf', C_STRUCTS.__wasi_fdstat_t.fs_filetype, 'type', 'i8') }}};
-    // TODO {{{ makeSetValue('pbuf', C_STRUCTS.__wasi_fdstat_t.fs_flags, '?', 'i16') }}};
-    // TODO {{{ makeSetValue('pbuf', C_STRUCTS.__wasi_fdstat_t.fs_rights_base, '?', 'i64') }}};
-    // TODO {{{ makeSetValue('pbuf', C_STRUCTS.__wasi_fdstat_t.fs_rights_inheriting, '?', 'i64') }}};
     return 0;
   },
 
-  fd_sync__sig: 'ii',
+  fd_filestat_get__sig: 'iii',
+  fd_filestat_get: function(fd, stat_buf) {
+    // TODO(sbc): implement
+    var stream = SYSCALLS.getStreamFromFD(fd);
+    {{{ makeSetValue('stat_buf', C_STRUCTS.__wasi_filestat_t.dev, '0', 'i64') }}};
+    {{{ makeSetValue('stat_buf', C_STRUCTS.__wasi_filestat_t.ino, '0', 'i64') }}};
+    {{{ makeSetValue('stat_buf', C_STRUCTS.__wasi_filestat_t.filetype, '0', 'i8') }}};
+    {{{ makeSetValue('stat_buf', C_STRUCTS.__wasi_filestat_t.nlink, '0', 'i64') }}};
+    {{{ makeSetValue('stat_buf', C_STRUCTS.__wasi_filestat_t.size, '0', 'i64') }}};
+    {{{ makeSetValue('stat_buf', C_STRUCTS.__wasi_filestat_t.atim, '0', 'i64') }}};
+    {{{ makeSetValue('stat_buf', C_STRUCTS.__wasi_filestat_t.mtim, '0', 'i64') }}};
+    {{{ makeSetValue('stat_buf', C_STRUCTS.__wasi_filestat_t.ctim, '0', 'i64') }}};
+    return 0;
+  },
+#endif
+
+#if PURE_WASI
+  fd_fdstat_get__deps: ['$preopens'],
+#endif
+  fd_fdstat_get: function(fd, pbuf) {
+    var rightsBase = 0;
+    var rightsInheriting = 0;
+    var flags = 0;
+#if PURE_WASI
+    if (fd in preopens) {
+      var type = {{{ cDefs.__WASI_FILETYPE_DIRECTORY }}};
+      rightsBase =  {{{ cDefs.__WASI_RIGHTS_PATH_CREATE_FILE |
+                             cDefs.__WASI_RIGHTS_PATH_OPEN }}};
+      rightsInheriting =  {{{ cDefs.__WASI_RIGHTS_FD_READ |
+                                   cDefs.__WASI_RIGHTS_FD_WRITE }}}
+    } else
+#endif
+    {
+#if SYSCALLS_REQUIRE_FILESYSTEM
+      var stream = SYSCALLS.getStreamFromFD(fd);
+      // All character devices are terminals (other things a Linux system would
+      // assume is a character device, like the mouse, we have special APIs for).
+      var type = stream.tty ? {{{ cDefs.__WASI_FILETYPE_CHARACTER_DEVICE }}} :
+                 FS.isDir(stream.mode) ? {{{ cDefs.__WASI_FILETYPE_DIRECTORY }}} :
+                 FS.isLink(stream.mode) ? {{{ cDefs.__WASI_FILETYPE_SYMBOLIC_LINK }}} :
+                 {{{ cDefs.__WASI_FILETYPE_REGULAR_FILE }}};
+#else
+      // hack to support printf in SYSCALLS_REQUIRE_FILESYSTEM=0
+      var type = fd == 0 || fd == 1 || fd == 2 ? {{{ cDefs.__WASI_FILETYPE_CHARACTER_DEVICE }}} : abort();
+      if (fd == 0) {
+        rightsBase = {{{ cDefs.__WASI_RIGHTS_FD_READ }}};
+      } else if (fd == 1 || fd == 2) {
+        rightsBase = {{{ cDefs.__WASI_RIGHTS_FD_WRITE }}};
+      }
+      flags = {{{ cDefs.__WASI_FDFLAGS_APPEND }}};
+#endif
+    }
+    {{{ makeSetValue('pbuf', C_STRUCTS.__wasi_fdstat_t.fs_filetype, 'type', 'i8') }}};
+    {{{ makeSetValue('pbuf', C_STRUCTS.__wasi_fdstat_t.fs_flags, 'flags', 'i16') }}};
+    {{{ makeSetValue('pbuf', C_STRUCTS.__wasi_fdstat_t.fs_rights_base, 'rightsBase', 'i64') }}};
+    {{{ makeSetValue('pbuf', C_STRUCTS.__wasi_fdstat_t.fs_rights_inheriting, 'rightsInheriting', 'i64') }}};
+    return 0;
+  },
+
   fd_sync: function(fd) {
 #if SYSCALLS_REQUIRE_FILESYSTEM
     var stream = SYSCALLS.getStreamFromFD(fd);
@@ -409,7 +549,7 @@ var WasiLibrary = {
       }
       mount.type.syncfs(mount, false, function(err) {
         if (err) {
-          wakeUp(function() { return {{{ cDefine('EIO') }}} });
+          wakeUp(function() { return {{{ cDefs.EIO }}} });
           return;
         }
         wakeUp(0);
@@ -417,14 +557,14 @@ var WasiLibrary = {
     });
 #else
     if (stream.stream_ops && stream.stream_ops.fsync) {
-      return -stream.stream_ops.fsync(stream);
+      return stream.stream_ops.fsync(stream);
     }
     return 0; // we can't do anything synchronously; the in-memory FS is already synced to
 #endif // ASYNCIFY
 #elif ASSERTIONS
     abort('fd_sync called without SYSCALLS_REQUIRE_FILESYSTEM');
 #else
-    return {{{ cDefine('ENOSYS') }}};
+    return {{{ cDefs.ENOSYS }}};
 #endif // SYSCALLS_REQUIRE_FILESYSTEM
   },
 };
