@@ -5,86 +5,83 @@
  */
 
 mergeInto(LibraryManager.library, {
-  // TODO: Generate these ID pools from a common utility.
-  $wasmfsOPFSDirectoryHandles: {
-    allocated: [],
-    free: [],
-    get: function(i) {
-#if ASSERTIONS
-      assert(this.allocated[i] !== undefined);
-#endif
-      return this.allocated[i];
+  $wasmfsOPFSDirectoryHandles__deps: ['$HandleAllocator'],
+  $wasmfsOPFSDirectoryHandles: "new HandleAllocator()",
+  $wasmfsOPFSFileHandles__deps: ['$HandleAllocator'],
+  $wasmfsOPFSFileHandles: "new HandleAllocator()",
+  $wasmfsOPFSAccessHandles__deps: ['$HandleAllocator'],
+  $wasmfsOPFSAccessHandles: "new HandleAllocator()",
+  $wasmfsOPFSBlobs__deps: ["$HandleAllocator"],
+  $wasmfsOPFSBlobs: "new HandleAllocator()",
+
+#if !USE_PTHREADS
+  // OPFS will only be used on modern browsers that supports JS classes.
+  $FileSystemAsyncAccessHandle: class FileSystemAsyncAccessHandle {
+    // This class implements the same interface as the sync version, but has
+    // async reads and writes. Hopefully this will one day be implemented by the
+    // platform so we can remove it.
+    constructor(handle) {
+      this.handle = handle;
+    }
+    async close() {}
+    async flush() {}
+    async getSize() {
+      let file = await this.handle.getFile();
+      return file.size;
+    }
+    async read(buffer, options = { at: 0 }) {
+      let file = await this.handle.getFile();
+      // The end position may be past the end of the file, but slice truncates
+      // it.
+      let slice = await file.slice(options.at, options.at + buffer.length);
+      let fileBuffer = await slice.arrayBuffer();
+      let array = new Uint8Array(fileBuffer);
+      buffer.set(array);
+      return array.length;
+    }
+    async write(buffer, options = { at: 0 }) {
+      let writable = await this.handle.createWritable({keepExistingData: true});
+      await writable.write({ type: 'write', position: options.at, data: buffer });
+      await writable.close();
+      return buffer.length;
+    }
+    async truncate(size) {
+      let writable = await this.handle.createWritable({keepExistingData: true});
+      await writable.truncate(size);
+      await writable.close();
     }
   },
 
-  $wasmfsOPFSFileHandles: {
-    allocated: [],
-    free: [],
-    get: function(i) {
-#if ASSERTIONS
-      assert(this.allocated[i] !== undefined);
+  $wasmfsOPFSCreateAsyncAccessHandle__deps: ['$FileSystemAsyncAccessHandle'],
+  $wasmfsOPFSCreateAsyncAccessHandle: function(fileHandle) {
+    return new FileSystemAsyncAccessHandle(fileHandle);
+  },
 #endif
-      return this.allocated[i];
-    }
-  },
 
-  $wasmfsOPFSAccessHandles: {
-    allocated: [],
-    free: [],
-    get: function(i) {
-#if ASSERTIONS
-      assert(this.allocated[i] !== undefined);
+  $wasmfsOPFSProxyFinish: function(ctx) {
+    // When using pthreads the proxy needs to know when the work is finished.
+    // When used with JSPI the work will be executed in an async block so there
+    // is no need to notify when done.
+#if USE_PTHREADS
+    _emscripten_proxy_finish(ctx);
 #endif
-      return this.allocated[i];
-    }
   },
 
-  $wasmfsOPFSBlobs: {
-    allocated: [],
-    free: [],
-    get: function(i) {
-#if ASSERTIONS
-      assert(this.allocated[i] !== undefined);
-#endif
-      return this.allocated[i];
-    }
-  },
-
-  $wasmfsOPFSAllocate: function(ids, handle) {
-    let id;
-    if (ids.free.length > 0) {
-      id = ids.free.pop();
-      ids.allocated[id] = handle;
-    } else {
-      id = ids.allocated.length;
-      ids.allocated.push(handle);
-    }
-    return id;
-  },
-
-  $wasmfsOPFSFree: function(ids, id) {
-#if ASSERTIONS
-    assert(ids.allocated[id] !== undefined);
-#endif
-    delete ids.allocated[id];
-    ids.free.push(id);
-  },
-
-  _wasmfs_opfs_init_root_directory__deps: ['$wasmfsOPFSDirectoryHandles'],
+  _wasmfs_opfs_init_root_directory__sig: 'vp',
+  _wasmfs_opfs_init_root_directory__deps: ['$wasmfsOPFSDirectoryHandles', '$wasmfsOPFSProxyFinish'],
   _wasmfs_opfs_init_root_directory: async function(ctx) {
-    if (wasmfsOPFSDirectoryHandles.allocated.length == 0) {
-      // Directory 0 is reserved as the root
+    // allocated.length starts off as 1 since 0 is a reserved handle
+    if (wasmfsOPFSDirectoryHandles.allocated.length == 1) {
       let root = await navigator.storage.getDirectory();
       wasmfsOPFSDirectoryHandles.allocated.push(root);
     }
-    _emscripten_proxy_finish(ctx);
+    wasmfsOPFSProxyFinish(ctx);
   },
 
   // Return the file ID for the file with `name` under `parent`, creating it if
   // it doesn't exist and `create` or otherwise return a negative error code
   // corresponding to the error.
-  $wasmfsOPFSGetOrCreateFile__deps: ['$wasmfsOPFSAllocate',
-                                     '$wasmfsOPFSDirectoryHandles',
+  $wasmfsOPFSGetOrCreateFile__deps: ['$wasmfsOPFSDirectoryHandles',
                                      '$wasmfsOPFSFileHandles'],
   $wasmfsOPFSGetOrCreateFile: async function(parent, name, create) {
     let parentHandle = wasmfsOPFSDirectoryHandles.get(parent);
@@ -93,24 +90,23 @@ mergeInto(LibraryManager.library, {
       fileHandle = await parentHandle.getFileHandle(name, {create: create});
     } catch (e) {
       if (e.name === "NotFoundError") {
-        return -{{{ cDefine('EEXIST') }}};
+        return -{{{ cDefs.EEXIST }}};
       }
       if (e.name === "TypeMismatchError") {
-        return -{{{ cDefine('EISDIR') }}};
+        return -{{{ cDefs.EISDIR }}};
       }
 #if ASSERTIONS
       err('unexpected error:', e, e.stack);
 #endif
-      return -{{{ cDefine('EIO') }}};
+      return -{{{ cDefs.EIO }}};
     }
-    return wasmfsOPFSAllocate(wasmfsOPFSFileHandles, fileHandle);
+    return wasmfsOPFSFileHandles.allocate(fileHandle);
   },
 
   // Return the file ID for the directory with `name` under `parent`, creating
   // it if it doesn't exist and `create` or otherwise return a negative error
   // code corresponding to the error.
-  $wasmfsOPFSGetOrCreateDir__deps: ['$wasmfsOPFSAllocate',
-                                    '$wasmfsOPFSDirectoryHandles'],
+  $wasmfsOPFSGetOrCreateDir__deps: ['$wasmfsOPFSDirectoryHandles'],
   $wasmfsOPFSGetOrCreateDir: async function(parent, name, create) {
     let parentHandle = wasmfsOPFSDirectoryHandles.get(parent);
     let childHandle;
@@ -119,36 +115,38 @@ mergeInto(LibraryManager.library, {
           await parentHandle.getDirectoryHandle(name, {create: create});
     } catch (e) {
       if (e.name === "NotFoundError") {
-        return -{{{ cDefine('EEXIST') }}};
+        return -{{{ cDefs.EEXIST }}};
       }
       if (e.name === "TypeMismatchError") {
-        return -{{{ cDefine('ENOTDIR') }}};
+        return -{{{ cDefs.ENOTDIR }}};
       }
 #if ASSERTIONS
       err('unexpected error:', e, e.stack);
 #endif
-      return -{{{ cDefine('EIO') }}};
+      return -{{{ cDefs.EIO }}};
     }
-    return wasmfsOPFSAllocate(wasmfsOPFSDirectoryHandles, childHandle);
+    return wasmfsOPFSDirectoryHandles.allocate(childHandle);
   },
 
+  _wasmfs_opfs_get_child__sig: 'vpippp',
   _wasmfs_opfs_get_child__deps: ['$wasmfsOPFSGetOrCreateFile',
-                                 '$wasmfsOPFSGetOrCreateDir'],
+                                 '$wasmfsOPFSGetOrCreateDir', '$wasmfsOPFSProxyFinish'],
   _wasmfs_opfs_get_child:
       async function(ctx, parent, namePtr, childTypePtr, childIDPtr) {
     let name = UTF8ToString(namePtr);
     let childType = 1;
     let childID = await wasmfsOPFSGetOrCreateFile(parent, name, false);
-    if (childID == -{{{ cDefine('EISDIR') }}}) {
+    if (childID == -{{{ cDefs.EISDIR }}}) {
       childType = 2;
       childID = await wasmfsOPFSGetOrCreateDir(parent, name, false);
     }
     {{{ makeSetValue('childTypePtr', 0, 'childType', 'i32') }}};
     {{{ makeSetValue('childIDPtr', 0, 'childID', 'i32') }}};
-    _emscripten_proxy_finish(ctx);
+    wasmfsOPFSProxyFinish(ctx);
   },
 
-  _wasmfs_opfs_get_entries__deps: [],
+  _wasmfs_opfs_get_entries__sig: 'vpipp',
+  _wasmfs_opfs_get_entries__deps: ['$wasmfsOPFSProxyFinish'],
   _wasmfs_opfs_get_entries: async function(ctx, dirID, entriesPtr, errPtr) {
     let dirHandle = wasmfsOPFSDirectoryHandles.get(dirID);
 
@@ -158,7 +156,7 @@ mergeInto(LibraryManager.library, {
       for (let entry; entry = await iter.next(), !entry.done;) {
         let [name, child] = entry.value;
         withStackSave(() => {
-          let namePtr = allocateUTF8OnStack(name);
+          let namePtr = stringToUTF8OnStack(name);
           let type = child.kind == "file" ?
               {{{ cDefine('File::DataFileKind') }}} :
           {{{ cDefine('File::DirectoryKind') }}};
@@ -166,77 +164,87 @@ mergeInto(LibraryManager.library, {
         });
       }
     } catch {
-      let err = -{{{ cDefine('EIO') }}};
+      let err = -{{{ cDefs.EIO }}};
       {{{ makeSetValue('errPtr', 0, 'err', 'i32') }}};
     }
-    _emscripten_proxy_finish(ctx);
+    wasmfsOPFSProxyFinish(ctx);
   },
 
-  _wasmfs_opfs_insert_file__deps: ['$wasmfsOPFSGetOrCreateFile'],
+  _wasmfs_opfs_insert_file__sig: 'vpipp',
+  _wasmfs_opfs_insert_file__deps: ['$wasmfsOPFSGetOrCreateFile', '$wasmfsOPFSProxyFinish'],
   _wasmfs_opfs_insert_file: async function(ctx, parent, namePtr, childIDPtr) {
     let name = UTF8ToString(namePtr);
     let childID = await wasmfsOPFSGetOrCreateFile(parent, name, true);
     {{{ makeSetValue('childIDPtr', 0, 'childID', 'i32') }}};
-    _emscripten_proxy_finish(ctx);
+    wasmfsOPFSProxyFinish(ctx);
   },
 
-  _wasmfs_opfs_insert_directory__deps: ['$wasmfsOPFSGetOrCreateDir'],
+  _wasmfs_opfs_insert_directory__sig: 'vpipp',
+  _wasmfs_opfs_insert_directory__deps: ['$wasmfsOPFSGetOrCreateDir', '$wasmfsOPFSProxyFinish'],
   _wasmfs_opfs_insert_directory:
       async function(ctx, parent, namePtr, childIDPtr) {
     let name = UTF8ToString(namePtr);
     let childID = await wasmfsOPFSGetOrCreateDir(parent, name, true);
     {{{ makeSetValue('childIDPtr', 0, 'childID', 'i32') }}};
-    _emscripten_proxy_finish(ctx);
+    wasmfsOPFSProxyFinish(ctx);
   },
 
-  _wasmfs_opfs_move__deps: ['$wasmfsOPFSFileHandles',
-                            '$wasmfsOPFSDirectoryHandles'],
-  _wasmfs_opfs_move: async function(ctx, fileID, newDirID, namePtr, errPtr) {
+  _wasmfs_opfs_move_file__sig: 'vpiipp',
+  _wasmfs_opfs_move_file__deps: ['$wasmfsOPFSFileHandles',
+                                 '$wasmfsOPFSDirectoryHandles',
+                                 '$wasmfsOPFSProxyFinish'],
+  _wasmfs_opfs_move_file: async function(ctx, fileID, newParentID, namePtr, errPtr) {
     let name = UTF8ToString(namePtr);
     let fileHandle = wasmfsOPFSFileHandles.get(fileID);
-    let newDirHandle = wasmfsOPFSDirectoryHandles.get(newDirID);
+    let newDirHandle = wasmfsOPFSDirectoryHandles.get(newParentID);
     try {
       await fileHandle.move(newDirHandle, name);
     } catch {
-      let err = -{{{ cDefine('EIO') }}};
+      let err = -{{{ cDefs.EIO }}};
       {{{ makeSetValue('errPtr', 0, 'err', 'i32') }}};
     }
-    _emscripten_proxy_finish(ctx);
+    wasmfsOPFSProxyFinish(ctx);
   },
 
-  _wasmfs_opfs_remove_child__deps: ['$wasmfsOPFSFree',
-                                    '$wasmfsOPFSDirectoryHandles'],
+  _wasmfs_opfs_remove_child__sig: 'vpipp',
+  _wasmfs_opfs_remove_child__deps: ['$wasmfsOPFSDirectoryHandles', '$wasmfsOPFSProxyFinish'],
   _wasmfs_opfs_remove_child: async function(ctx, dirID, namePtr, errPtr) {
     let name = UTF8ToString(namePtr);
     let dirHandle = wasmfsOPFSDirectoryHandles.get(dirID);
     try {
       await dirHandle.removeEntry(name);
     } catch {
-      let err = -{{{ cDefine('EIO') }}};
+      let err = -{{{ cDefs.EIO }}};
       {{{ makeSetValue('errPtr', 0, 'err', 'i32') }}};
     }
-    _emscripten_proxy_finish(ctx);
+    wasmfsOPFSProxyFinish(ctx);
   },
 
-  _wasmfs_opfs_free_file__deps: ['$wasmfsOPFSFree', '$wasmfsOPFSFileHandles'],
+  _wasmfs_opfs_free_file__sig: 'vi',
+  _wasmfs_opfs_free_file__deps: ['$wasmfsOPFSFileHandles'],
   _wasmfs_opfs_free_file: function(fileID) {
-    wasmfsOPFSFree(wasmfsOPFSFileHandles, fileID);
+    wasmfsOPFSFileHandles.free(fileID);
   },
 
-  _wasmfs_opfs_free_directory__deps: ['$wasmfsOPFSFree',
-                                      '$wasmfsOPFSDirectoryHandles'],
+  _wasmfs_opfs_free_directory__sig: 'vi',
+  _wasmfs_opfs_free_directory__deps: ['$wasmfsOPFSDirectoryHandles'],
   _wasmfs_opfs_free_directory: function(dirID) {
-    wasmfsOPFSFree(wasmfsOPFSDirectoryHandles, dirID);
+    wasmfsOPFSDirectoryHandles.free(dirID);
   },
 
-  _wasmfs_opfs_open_access__deps: ['$wasmfsOPFSAllocate',
-                                   '$wasmfsOPFSFileHandles',
-                                   '$wasmfsOPFSAccessHandles'],
+  _wasmfs_opfs_open_access__sig: 'vpip',
+  _wasmfs_opfs_open_access__deps: ['$wasmfsOPFSFileHandles',
+                                   '$wasmfsOPFSAccessHandles', '$wasmfsOPFSProxyFinish',
+#if !USE_PTHREADS
+                                   '$wasmfsOPFSCreateAsyncAccessHandle'
+#endif
+                                  ],
   _wasmfs_opfs_open_access: async function(ctx, fileID, accessIDPtr) {
     let fileHandle = wasmfsOPFSFileHandles.get(fileID);
     let accessID;
     try {
       let accessHandle;
+#if USE_PTHREADS
       // TODO: Remove this once the Access Handles API has settled.
       if (FileSystemFileHandle.prototype.createSyncAccessHandle.length == 0) {
         accessHandle = await fileHandle.createSyncAccessHandle();
@@ -244,84 +252,89 @@ mergeInto(LibraryManager.library, {
         accessHandle = await fileHandle.createSyncAccessHandle(
             {mode: "in-place"});
       }
-      accessID = wasmfsOPFSAllocate(wasmfsOPFSAccessHandles, accessHandle);
+#else
+      accessHandle = await wasmfsOPFSCreateAsyncAccessHandle(fileHandle);
+#endif
+      accessID = wasmfsOPFSAccessHandles.allocate(accessHandle);
     } catch (e) {
       // TODO: Presumably only one of these will appear in the final API?
       if (e.name === "InvalidStateError" ||
           e.name === "NoModificationAllowedError") {
-        accessID = -{{{ cDefine('EACCES') }}};
+        accessID = -{{{ cDefs.EACCES }}};
       } else {
 #if ASSERTIONS
         err('unexpected error:', e, e.stack);
 #endif
-        accessID = -{{{ cDefine('EIO') }}};
+        accessID = -{{{ cDefs.EIO }}};
       }
     }
     {{{ makeSetValue('accessIDPtr', 0, 'accessID', 'i32') }}};
-    _emscripten_proxy_finish(ctx);
+    wasmfsOPFSProxyFinish(ctx);
   },
 
-  _wasmfs_opfs_open_blob__deps: ['$wasmfsOPFSAllocate',
-                                 '$wasmfsOPFSFileHandles',
-                                 '$wasmfsOPFSBlobs'],
+  _wasmfs_opfs_open_blob__sig: 'vpip',
+  _wasmfs_opfs_open_blob__deps: ['$wasmfsOPFSFileHandles',
+                                 '$wasmfsOPFSBlobs', '$wasmfsOPFSProxyFinish'],
   _wasmfs_opfs_open_blob: async function(ctx, fileID, blobIDPtr) {
     let fileHandle = wasmfsOPFSFileHandles.get(fileID);
     let blobID;
     try {
       let blob = await fileHandle.getFile();
-      blobID = wasmfsOPFSAllocate(wasmfsOPFSBlobs, blob);
+      blobID = wasmfsOPFSBlobs.allocate(blob);
     } catch (e) {
       if (e.name === "NotAllowedError") {
-        blobID = -{{{ cDefine('EACCES') }}};
+        blobID = -{{{ cDefs.EACCES }}};
       } else {
 #if ASSERTIONS
         err('unexpected error:', e, e.stack);
 #endif
-        blobID = -{{{ cDefine('EIO') }}};
+        blobID = -{{{ cDefs.EIO }}};
       }
     }
     {{{ makeSetValue('blobIDPtr', 0, 'blobID', 'i32') }}};
-    _emscripten_proxy_finish(ctx);
+    wasmfsOPFSProxyFinish(ctx);
   },
 
-  _wasmfs_opfs_close_access__deps: ['$wasmfsOPFSFree',
-                                    '$wasmfsOPFSAccessHandles'],
+  _wasmfs_opfs_close_access__sig: 'vpip',
+  _wasmfs_opfs_close_access__deps: ['$wasmfsOPFSAccessHandles', '$wasmfsOPFSProxyFinish'],
   _wasmfs_opfs_close_access: async function(ctx, accessID, errPtr) {
     let accessHandle = wasmfsOPFSAccessHandles.get(accessID);
     try {
       await accessHandle.close();
     } catch {
-      let err = -{{{ cDefine('EIO') }}};
+      let err = -{{{ cDefs.EIO }}};
       {{{ makeSetValue('errPtr', 0, 'err', 'i32') }}};
     }
-    wasmfsOPFSFree(wasmfsOPFSAccessHandles, accessID);
-    _emscripten_proxy_finish(ctx);
+    wasmfsOPFSAccessHandles.free(accessID);
+    wasmfsOPFSProxyFinish(ctx);
   },
 
-  _wasmfs_opfs_close_blob__deps: ['$wasmfsOPFSFree',
-                                  '$wasmfsOPFSBlobs'],
+  _wasmfs_opfs_close_blob__sig: 'vi',
+  _wasmfs_opfs_close_blob__deps: ['$wasmfsOPFSBlobs'],
   _wasmfs_opfs_close_blob: function(blobID) {
-    wasmfsOPFSFree(wasmfsOPFSBlobs, blobID);
+    wasmfsOPFSBlobs.free(blobID);
   },
 
+  _wasmfs_opfs_read_access__sig: 'iipii',
   _wasmfs_opfs_read_access__deps: ['$wasmfsOPFSAccessHandles'],
-  _wasmfs_opfs_read_access: function(accessID, bufPtr, len, pos) {
+  _wasmfs_opfs_read_access: {{{ asyncIf(!USE_PTHREADS) }}}  function(accessID, bufPtr, len, pos) {
     let accessHandle = wasmfsOPFSAccessHandles.get(accessID);
     let data = HEAPU8.subarray(bufPtr, bufPtr + len);
     try {
-      return accessHandle.read(data, {at: pos});
+      return {{{ awaitIf(!USE_PTHREADS) }}} accessHandle.read(data, {at: pos});
     } catch (e) {
       if (e.name == "TypeError") {
-        return -{{{ cDefine('EINVAL') }}};
+        return -{{{ cDefs.EINVAL }}};
       }
 #if ASSERTIONS
       err('unexpected error:', e, e.stack);
 #endif
-      return -{{{ cDefine('EIO') }}};
+      return -{{{ cDefs.EIO }}};
     }
   },
 
-  _wasmfs_opfs_read_blob__deps: ['$wasmfsOPFSBlobs'],
+  _wasmfs_opfs_read_blob__sig: 'ipipiip',
+  _wasmfs_opfs_read_blob__deps: ['$wasmfsOPFSBlobs', '$wasmfsOPFSProxyFinish'],
   _wasmfs_opfs_read_blob: async function(ctx, blobID, bufPtr, len, pos, nreadPtr) {
     let blob = wasmfsOPFSBlobs.get(blobID);
     let slice = blob.slice(pos, pos + len);
@@ -337,69 +350,74 @@ mergeInto(LibraryManager.library, {
       nread += data.length;
     } catch (e) {
       if (e instanceof RangeError) {
-        nread = -{{{ cDefine('EFAULT') }}};
+        nread = -{{{ cDefs.EFAULT }}};
       } else {
 #if ASSERTIONS
         err('unexpected error:', e, e.stack);
 #endif
-        nread = -{{{ cDefine('EIO') }}};
+        nread = -{{{ cDefs.EIO }}};
       }
     }
 
     {{{ makeSetValue('nreadPtr', 0, 'nread', 'i32') }}};
-    _emscripten_proxy_finish(ctx);
+    wasmfsOPFSProxyFinish(ctx);
   },
 
+  _wasmfs_opfs_write_access__sig: 'iipii',
   _wasmfs_opfs_write_access__deps: ['$wasmfsOPFSAccessHandles'],
-  _wasmfs_opfs_write_access: function(accessID, bufPtr, len, pos) {
+  _wasmfs_opfs_write_access: {{{ asyncIf(!USE_PTHREADS) }}} function(accessID, bufPtr, len, pos) {
     let accessHandle = wasmfsOPFSAccessHandles.get(accessID);
     let data = HEAPU8.subarray(bufPtr, bufPtr + len);
     try {
-      return accessHandle.write(data, {at: pos});
+      return {{{ awaitIf(!USE_PTHREADS) }}} accessHandle.write(data, {at: pos});
     } catch (e) {
       if (e.name == "TypeError") {
-        return -{{{ cDefine('EINVAL') }}};
+        return -{{{ cDefs.EINVAL }}};
       }
 #if ASSERTIONS
       err('unexpected error:', e, e.stack);
 #endif
-      return -{{{ cDefine('EIO') }}};
+      return -{{{ cDefs.EIO }}};
     }
   },
 
-  _wasmfs_opfs_get_size_access__deps: ['$wasmfsOPFSAccessHandles'],
+  _wasmfs_opfs_get_size_access__sig: 'vpip',
+  _wasmfs_opfs_get_size_access__deps: ['$wasmfsOPFSAccessHandles', '$wasmfsOPFSProxyFinish'],
   _wasmfs_opfs_get_size_access: async function(ctx, accessID, sizePtr) {
     let accessHandle = wasmfsOPFSAccessHandles.get(accessID);
     let size;
     try {
       size = await accessHandle.getSize();
     } catch {
-      size = -{{{ cDefine('EIO') }}};
+      size = -{{{ cDefs.EIO }}};
     }
     {{{ makeSetValue('sizePtr', 0, 'size', 'i64') }}};
-    _emscripten_proxy_finish(ctx);
+    wasmfsOPFSProxyFinish(ctx);
   },
 
+  _wasmfs_opfs_get_size_blob__sig: 'ii',
   _wasmfs_opfs_get_size_blob__deps: ['$wasmfsOPFSBlobs'],
   _wasmfs_opfs_get_size_blob: function(blobID) {
     // This cannot fail.
     return wasmfsOPFSBlobs.get(blobID).size;
   },
 
-  _wasmfs_opfs_get_size_file__deps: ['$wasmfsOPFSFileHandles'],
+  _wasmfs_opfs_get_size_file__sig: 'vpip',
+  _wasmfs_opfs_get_size_file__deps: ['$wasmfsOPFSFileHandles', '$wasmfsOPFSProxyFinish'],
   _wasmfs_opfs_get_size_file: async function(ctx, fileID, sizePtr) {
     let fileHandle = wasmfsOPFSFileHandles.get(fileID);
     let size;
     try {
       size = (await fileHandle.getFile()).size;
     } catch {
-      size = -{{{ cDefine('EIO') }}};
+      size = -{{{ cDefs.EIO }}};
     }
     {{{ makeSetValue('sizePtr', 0, 'size', 'i64') }}};
-    _emscripten_proxy_finish(ctx);
+    wasmfsOPFSProxyFinish(ctx);
   },
 
-  _wasmfs_opfs_set_size_access__deps: ['$wasmfsOPFSAccessHandles'],
+  _wasmfs_opfs_set_size_access__sig: 'vpijp',
+  _wasmfs_opfs_set_size_access__deps: ['$wasmfsOPFSAccessHandles', '$wasmfsOPFSProxyFinish'],
   _wasmfs_opfs_set_size_access: async function(ctx, accessID,
                                                {{{ defineI64Param('size') }}},
                                                errPtr) {
@@ -408,13 +426,14 @@ mergeInto(LibraryManager.library, {
     try {
       await accessHandle.truncate(size);
     } catch {
-      let err = -{{{ cDefine('EIO') }}};
+      let err = -{{{ cDefs.EIO }}};
       {{{ makeSetValue('errPtr', 0, 'err', 'i32') }}};
     }
-    _emscripten_proxy_finish(ctx);
+    wasmfsOPFSProxyFinish(ctx);
   },
 
-  _wasmfs_opfs_set_size_file__deps: ['$wasmfsOPFSFileHandles'],
+  _wasmfs_opfs_set_size_file__sig: 'vpijp',
+  _wasmfs_opfs_set_size_file__deps: ['$wasmfsOPFSFileHandles', '$wasmfsOPFSProxyFinish'],
   _wasmfs_opfs_set_size_file: async function(ctx, fileID,
                                              {{{ defineI64Param('size') }}},
                                              errPtr) {
@@ -425,21 +444,22 @@ mergeInto(LibraryManager.library, {
       await writable.truncate(size);
       await writable.close();
     } catch {
-      let err = -{{{ cDefine('EIO') }}};
+      let err = -{{{ cDefs.EIO }}};
       {{{ makeSetValue('errPtr', 0, 'err', 'i32') }}};
     }
-    _emscripten_proxy_finish(ctx);
+    wasmfsOPFSProxyFinish(ctx);
   },
 
-  _wasmfs_opfs_flush_access__deps: ['$wasmfsOPFSAccessHandles'],
+  _wasmfs_opfs_flush_access__sig: 'vpip',
+  _wasmfs_opfs_flush_access__deps: ['$wasmfsOPFSAccessHandles', '$wasmfsOPFSProxyFinish'],
   _wasmfs_opfs_flush_access: async function(ctx, accessID, errPtr) {
     let accessHandle = wasmfsOPFSAccessHandles.get(accessID);
     try {
       await accessHandle.flush();
     } catch {
-      let err = -{{{ cDefine('EIO') }}};
+      let err = -{{{ cDefs.EIO }}};
       {{{ makeSetValue('errPtr', 0, 'err', 'i32') }}};
     }
-    _emscripten_proxy_finish(ctx);
+    wasmfsOPFSProxyFinish(ctx);
   }
 });
