@@ -11,8 +11,6 @@
 global.FOUR_GB = 4 * 1024 * 1024 * 1024;
 const FLOAT_TYPES = new Set(['float', 'double']);
 
-let currentlyParsedFilename = '';
-
 // Does simple 'macro' substitution, using Django-like syntax,
 // {{{ code }}} will be replaced with |eval(code)|.
 // NOTE: Be careful with that ret check. If ret is |0|, |ret ? ret.toString() : ''| would result in ''!
@@ -45,6 +43,8 @@ function preprocess(filename) {
       .replace(/\bimport\.meta\b/g, 'EMSCRIPTEN$IMPORT$META')
       .replace(/\bawait import\b/g, 'EMSCRIPTEN$AWAIT$IMPORT');
   }
+  // Remove windows line endings, if any
+  text = text.replace(/\r\n/g, '\n');
 
   const IGNORE = 0;
   const SHOW = 1;
@@ -55,7 +55,8 @@ function preprocess(filename) {
   const showStack = [];
   const showCurrentLine = () => showStack.every((x) => x == SHOW);
 
-  currentlyParsedFilename = filename;
+  const oldFilename = currentFile;
+  currentFile = filename;
   const fileExt = filename.split('.').pop().toLowerCase();
   const isHtml = (fileExt === 'html' || fileExt === 'htm') ? true : false;
   let inStyle = false;
@@ -65,88 +66,87 @@ function preprocess(filename) {
 
   try {
     for (let [i, line] of lines.entries()) {
-      if (line[line.length - 1] === '\r') {
-        line = line.substr(0, line.length - 1); // Windows will have '\r' left over from splitting over '\r\n'
-      }
-      if (isHtml && line.includes('<style') && !inStyle) {
-        inStyle = true;
-      }
-      if (isHtml && line.includes('</style') && inStyle) {
-        inStyle = false;
+      if (isHtml) {
+        if (line.includes('<style') && !inStyle) {
+          inStyle = true;
+        }
+        if (line.includes('</style') && inStyle) {
+          inStyle = false;
+        }
+        if (inStyle) {
+          if (showCurrentLine()) {
+            ret += line + '\n';
+          }
+          continue;
+        }
       }
 
-      if (!inStyle) {
-        const trimmed = line.trim();
-        if (trimmed.startsWith('#')) {
-          const first = trimmed.split(' ', 1)[0];
-          if (first == '#if' || first == '#ifdef' || first == '#elif') {
-            if (first == '#ifdef') {
-              warn('use of #ifdef in js library.  Use #if instead.');
-            }
-            if (first == '#elif') {
-              const curr = showStack.pop();
-              if (curr == SHOW || curr == IGNORE_ALL) {
-                // If we showed to previous block we enter the IGNORE_ALL state
-                // and stay there until endif is seen
-                showStack.push(IGNORE_ALL);
-                continue;
-              }
-            }
-            const after = trimmed.substring(trimmed.indexOf(' '));
-            const truthy = !!vm.runInThisContext(after, { filename, lineOffset: i, columnOffset: line.indexOf(after) });
-            showStack.push(truthy ? SHOW : IGNORE);
-          } else if (first === '#include') {
-            if (showCurrentLine()) {
-              let filename = line.substr(line.indexOf(' ') + 1);
-              if (filename.startsWith('"')) {
-                filename = filename.substr(1, filename.length - 2);
-              }
-              const result = preprocess(filename);
-              if (result) {
-                ret += `// include: ${filename}\n`;
-                ret += result;
-                ret += `// end include: ${filename}\n`;
-              }
-            }
-          } else if (first === '#else') {
-            assert(showStack.length > 0);
-            const curr = showStack.pop();
-            if (curr == IGNORE) {
-              showStack.push(SHOW);
-            } else {
-              showStack.push(IGNORE);
-            }
-          } else if (first === '#endif') {
-            assert(showStack.length > 0);
-            showStack.pop();
-          } else if (first === '#warning') {
-            if (showCurrentLine()) {
-              printErr(`${filename}:${i + 1}: #warning ${trimmed.substring(trimmed.indexOf(' ')).trim()}`);
-            }
-          } else if (first === '#error') {
-            if (showCurrentLine()) {
-              error(`${filename}:${i + 1}: #error ${trimmed.substring(trimmed.indexOf(' ')).trim()}`);
-            }
-          } else {
-            throw new Error(`${filename}:${i + 1}: Unknown preprocessor directive ${first}`);
+      const trimmed = line.trim();
+      if (trimmed.startsWith('#')) {
+        const first = trimmed.split(' ', 1)[0];
+        if (first == '#if' || first == '#ifdef' || first == '#elif') {
+          if (first == '#ifdef') {
+            warn('use of #ifdef in js library.  Use #if instead.');
           }
-        } else {
-          if (showCurrentLine()) {
-            // Never emit more than one empty line at a time.
-            if (emptyLine && !line) {
+          if (first == '#elif') {
+            const curr = showStack.pop();
+            if (curr == SHOW || curr == IGNORE_ALL) {
+              // If we showed to previous block we enter the IGNORE_ALL state
+              // and stay there until endif is seen
+              showStack.push(IGNORE_ALL);
               continue;
             }
-            ret += line + '\n';
-            if (!line) {
-              emptyLine = true;
-            } else {
-              emptyLine = false;
+          }
+          const after = trimmed.substring(trimmed.indexOf(' '));
+          const truthy = !!vm.runInThisContext(after, { filename, lineOffset: i, columnOffset: line.indexOf(after) });
+          showStack.push(truthy ? SHOW : IGNORE);
+        } else if (first === '#include') {
+          if (showCurrentLine()) {
+            let filename = line.substr(line.indexOf(' ') + 1);
+            if (filename.startsWith('"')) {
+              filename = filename.substr(1, filename.length - 2);
+            }
+            const result = preprocess(filename);
+            if (result) {
+              ret += `// include: ${filename}\n`;
+              ret += result;
+              ret += `// end include: ${filename}\n`;
             }
           }
+        } else if (first === '#else') {
+          assert(showStack.length > 0);
+          const curr = showStack.pop();
+          if (curr == IGNORE) {
+            showStack.push(SHOW);
+          } else {
+            showStack.push(IGNORE);
+          }
+        } else if (first === '#endif') {
+          assert(showStack.length > 0);
+          showStack.pop();
+        } else if (first === '#warning') {
+          if (showCurrentLine()) {
+            printErr(`${filename}:${i + 1}: #warning ${trimmed.substring(trimmed.indexOf(' ')).trim()}`);
+          }
+        } else if (first === '#error') {
+          if (showCurrentLine()) {
+            error(`${filename}:${i + 1}: #error ${trimmed.substring(trimmed.indexOf(' ')).trim()}`);
+          }
+        } else {
+          error(`${filename}:${i + 1}: Unknown preprocessor directive ${first}`);
         }
-      } else { // !inStyle
+      } else {
         if (showCurrentLine()) {
+          // Never emit more than one empty line at a time.
+          if (emptyLine && !line) {
+            continue;
+          }
           ret += line + '\n';
+          if (!line) {
+            emptyLine = true;
+          } else {
+            emptyLine = false;
+          }
         }
       }
     }
@@ -154,7 +154,7 @@ function preprocess(filename) {
 no matching #endif found (${showStack.length$}' unmatched preprocessing directives on stack)`);
     return ret;
   } finally {
-    currentlyParsedFilename = null;
+    currentFile = oldFilename;
   }
 }
 
@@ -190,22 +190,6 @@ function isPointerType(type) {
   return type[type.length - 1] == '*';
 }
 
-function isIntImplemented(type) {
-  return type[0] == 'i' || type[0] == 'u' || isPointerType(type);
-}
-
-// Note: works for iX types and structure types, not pointers (even though they are implemented as ints)
-function getBits(type, allowPointers) {
-  if (allowPointers && isPointerType(type)) return POINTER_SIZE;
-  if (!type) return 0;
-  if (type[0] == 'i' || type[0] == 'u') {
-    const left = type.substr(1);
-    if (!isNumber(left)) return 0;
-    return parseInt(left);
-  }
-  return 0;
-}
-
 // Given an expression like (VALUE=VALUE*2,VALUE<10?VALUE:t+1) , this will
 // replace VALUE with value. If value is not a simple identifier of a variable,
 // value will be replaced with tempVar.
@@ -221,13 +205,13 @@ function makeInlineCalculation(expression, value, tempVar) {
 
 // Splits a number (an integer in a double, possibly > 32 bits) into an i64
 // value, represented by a low and high i32 pair.
-// Will suffer from rounding.
+// Will suffer from rounding and truncation.
 function splitI64(value) {
   // general idea:
   //
   //  $1$0 = ~~$d >>> 0;
   //  $1$1 = Math.abs($d) >= 1 ? (
-  //     $d > 0 ? Math.min(Math.floor(($d)/ 4294967296.0), 4294967295.0)
+  //     $d > 0 ? Math.floor(($d)/ 4294967296.0) >>> 0
   //            : Math.ceil(Math.min(-4294967296.0, $d - $1$0)/ 4294967296.0)
   //  ) : 0;
   //
@@ -239,13 +223,13 @@ function splitI64(value) {
   // For negatives, we need to ensure a -1 if the value is overall negative,
   // even if not significant negative component
 
+  assert(!WASM_BIGINT, 'splitI64 should not be used when WASM_BIGINT is enabled');
   const low = value + '>>>0';
   const high = makeInlineCalculation(
       asmCoercion('Math.abs(VALUE)', 'double') + ' >= ' + asmEnsureFloat('1', 'double') + ' ? ' +
         '(VALUE > ' + asmEnsureFloat('0', 'double') + ' ? ' +
-        asmCoercion('Math.min(' + asmCoercion('Math.floor((VALUE)/' +
-        asmEnsureFloat(4294967296, 'double') + ')', 'double') + ', ' +
-        asmEnsureFloat(4294967295, 'double') + ')', 'i32') + '>>>0' +
+        asmCoercion('Math.floor((VALUE)/' +
+        asmEnsureFloat(4294967296, 'double') + ')', 'double') + '>>>0' +
         ' : ' +
         asmFloatToInt(asmCoercion('Math.ceil((VALUE - +((' + asmFloatToInt('VALUE') + ')>>>0))/' +
         asmEnsureFloat(4294967296, 'double') + ')', 'double')) + '>>>0' +
@@ -275,12 +259,13 @@ function indentify(text, indent) {
 // Correction tools
 
 function getHeapOffset(offset, type) {
-  if (!WASM_BIGINT && Runtime.getNativeFieldSize(type) > 4 && type == 'i64') {
-    // we emulate 64-bit integer values as 32 in asmjs-unknown-emscripten, but not double
+  if (type == 'i64' && !WASM_BIGINT) {
+    // We are foreced to use the 32-bit heap for 64-bit values when we don't
+    // have WASM_BIGINT.
     type = 'i32';
   }
 
-  const sz = Runtime.getNativeTypeSize(type);
+  const sz = getNativeTypeSize(type);
   const shifts = Math.log(sz) / Math.LN2;
   return `((${offset})>>${shifts})`;
 }
@@ -362,51 +347,41 @@ function makeGetValue(ptr, pos, type, noNeedFirst, unsigned, ignore, align) {
 }
 
 /**
- * @param {nunber} ptr The pointer. Used to find both the slab and the offset in that slab. If the pointer
+ * @param {number} ptr The pointer. Used to find both the slab and the offset in that slab. If the pointer
  *            is just an integer, then this is almost redundant, but in general the pointer type
  *            may in the future include information about which slab as well. So, for now it is
  *            possible to put |0| here, but if a pointer is available, that is more future-proof.
- * @param {nunber} pos The position in that slab - the offset. Added to any offset in the pointer itself.
+ * @param {number} pos The position in that slab - the offset. Added to any offset in the pointer itself.
  * @param {number} value The value to set.
  * @param {string} type A string defining the type. Used to find the slab (HEAPU8, HEAP16, HEAPU32, etc.).
  *             which means we should write to all slabs, ignore type differences if any on reads, etc.
- * @param {bool} noNeedFirst Whether to ignore the offset in the pointer itself.
- * @param {bool} ignore: legacy, ignored.
- * @param {number} align: legacy, ignored.
- * @param {string} sep: TODO
- * @return {TODO}
+ * @return {string} JS code for performing the memory set operation
  */
-function makeSetValue(ptr, pos, value, type, noNeedFirst, ignore, align, sep = ';') {
-  assert(typeof align === 'undefined', 'makeSetValue no longer supports align parameter');
-  assert(typeof noNeedFirst === 'undefined', 'makeSetValue no longer supports noNeedFirst parameter');
-  if (type == 'i64' && (!WASM_BIGINT || !MEMORY64)) {
-    // If we lack either BigInt support or Memory64 then we must fall back to an
-    // unaligned read of a 64-bit value: without BigInt we do not have HEAP64,
-    // and without Memory64 i64 fields are not guaranteed to be aligned to 64
-    // bits, so HEAP64[ptr>>3] might be broken.
-    return '(tempI64 = [' + splitI64(value) + '],' +
-            makeSetValue(ptr, pos, 'tempI64[0]', 'i32', noNeedFirst, ignore, align, ',') + ',' +
-            makeSetValue(ptr, getFastValue(pos, '+', Runtime.getNativeTypeSize('i32')), 'tempI64[1]', 'i32', noNeedFirst, ignore, align, ',') + ')';
+function makeSetValue(ptr, pos, value, type) {
+  var rtn = makeSetValueImpl(ptr, pos, value, type);
+  if (ASSERTIONS == 2 && (type.startsWith('i') || type.startsWith('u'))) {
+    const width = getBitWidth(type);
+    const assertion = `checkInt${width}(${value})`;
+    rtn += ';' + assertion
   }
+  return rtn;
+}
 
-  const bits = getBits(type);
-  const needSplitting = bits > 0 && !isPowerOfTwo(bits); // an unnatural type like i24
-  if (needSplitting) {
-    // Alignment is important here, or we need to split this up for other reasons.
-    const bytes = Runtime.getNativeTypeSize(type);
-    if (needSplitting) {
-      let ret = '';
-      ret += 'tempBigInt=' + value + sep;
-      for (let i = 0; i < bytes; i++) {
-        ret += makeSetValue(ptr, getFastValue(pos, '+', i), 'tempBigInt&0xff', 'i8', noNeedFirst, ignore, 1);
-        if (i < bytes - 1) ret += sep + 'tempBigInt = tempBigInt>>8' + sep;
-      }
-      return ret;
-    }
+function makeSetValueImpl(ptr, pos, value, type) {
+  if (type == 'i64' && !WASM_BIGINT) {
+    // If we lack BigInt support we must fall back to an reading a pair of I32
+    // values.
+    return '(tempI64 = [' + splitI64(value) + '], ' +
+            makeSetValueImpl(ptr, pos, 'tempI64[0]', 'i32') + ',' +
+            makeSetValueImpl(ptr, getFastValue(pos, '+', getNativeTypeSize('i32')), 'tempI64[1]', 'i32') + ')';
   }
 
   const offset = calcFastOffset(ptr, pos);
 
+  if (type === 'i53') {
+    return `writeI53ToI64(${offset}, ${value})`;
+  }
+
   const slab = getHeapForType(type);
   if (slab == 'HEAPU64' || slab == 'HEAP64') {
     value = `BigInt(${value})`;
@@ -414,51 +389,19 @@ function makeSetValue(ptr, pos, value, type, noNeedFirst, ignore, align, sep = '
   return slab + '[' + getHeapOffset(offset, type) + '] = ' + value;
 }
 
-const UNROLL_LOOP_MAX = 8;
-
-function makeCopyValues(dest, src, num, type, modifier, align, sep = ';') {
-  assert(typeof align === 'undefined');
-  function unroll(type, num, jump = 1) {
-    const setValues = range(num).map((i) => makeSetValue(dest, i * jump, makeGetValue(src, i * jump, type), type));
-    return setValues.join(sep);
-  }
-  // If we don't know how to handle this at compile-time, or handling it is best
-  // done in a large amount of code, call memcpy
-  if (!isNumber(num)) num = stripCorrections(num);
-  if (!isNumber(align)) align = stripCorrections(align);
-  if (!isNumber(num) || (parseInt(num) / align >= UNROLL_LOOP_MAX)) {
-    return '(_memcpy(' + dest + ', ' + src + ', ' + num + ')|0)';
-  }
-  num = parseInt(num);
-  // remove corrections, since we will be correcting after we add anyhow,
-  dest = stripCorrections(dest);
-  src = stripCorrections(src);
-  // and in the heap assignment expression
-  const ret = [];
-  [4, 2, 1].forEach((possibleAlign) => {
-    if (num == 0) return;
-    if (align >= possibleAlign) {
-      ret.push(unroll('i' + (possibleAlign * 8), Math.floor(num / possibleAlign), possibleAlign));
-      src = getFastValue(src, '+', Math.floor(num / possibleAlign) * possibleAlign);
-      dest = getFastValue(dest, '+', Math.floor(num / possibleAlign) * possibleAlign);
-      num %= possibleAlign;
-    }
-  });
-  return ret.join(sep);
-}
-
 function makeHEAPView(which, start, end) {
   const size = parseInt(which.replace('U', '').replace('F', '')) / 8;
   const mod = size == 1 ? '' : ('>>' + Math.log2(size));
   return `HEAP${which}.subarray((${start})${mod}, (${end})${mod})`;
 }
 
-const TWO_TWENTY = Math.pow(2, 20);
-
 // Given two values and an operation, returns the result of that operation.
 // Tries to do as much as possible at compile time.
-// Leaves overflows etc. unhandled, *except* for integer multiply, in order to be efficient with Math.imul
-function getFastValue(a, op, b, type) {
+function getFastValue(a, op, b) {
+  // In the past we supported many operations, but today we only use addition.
+  assert(op == '+');
+
+  // Convert 'true' and 'false' to '1' and '0'.
   a = a === 'true' ? '1' : (a === 'false' ? '0' : a);
   b = b === 'true' ? '1' : (b === 'false' ? '0' : b);
 
@@ -467,38 +410,23 @@ function getFastValue(a, op, b, type) {
   if (typeof a == 'number') {
     aNumber = a;
     a = a.toString();
-  } else if (isNumber(a)) aNumber = parseFloat(a);
+  } else if (isNumber(a)) {
+    aNumber = parseFloat(a);
+  }
   if (typeof b == 'number') {
     bNumber = b;
     b = b.toString();
-  } else if (isNumber(b)) bNumber = parseFloat(b);
+  } else if (isNumber(b)) {
+    bNumber = parseFloat(b);
+  }
 
+  // First check if we can do the addition at compile time
   if (aNumber !== null && bNumber !== null) {
-    switch (op) {
-      case '+': return (aNumber + bNumber).toString();
-      case '-': return (aNumber - bNumber).toString();
-      case '*': return (aNumber * bNumber).toString();
-      case '/': {
-        if (type[0] === 'i') {
-          return ((aNumber / bNumber) | 0).toString();
-        }
-        return (aNumber / bNumber).toString();
-      }
-      case '%': return (aNumber % bNumber).toString();
-      case '|': return (aNumber | bNumber).toString();
-      case '>>>': return (aNumber >>> bNumber).toString();
-      case '&': return (aNumber & bNumber).toString();
-      case 'pow': return Math.pow(aNumber, bNumber).toString();
-      default: assert(false, 'need to implement getFastValue pn ' + op);
-    }
+    return (aNumber + bNumber).toString();
   }
-  if (op === 'pow') {
-    if (a === '2' && isIntImplemented(type)) {
-      return `(1 << (${b}))`;
-    }
-    return `Math.pow(${a}, ${b})`;
-  }
-  if ((op === '+' || op === '*') && aNumber !== null) { // if one of them is a number, keep it last
+
+  // If one of them is a number, keep it last
+  if (aNumber !== null) {
     const c = b;
     b = a;
     a = c;
@@ -506,56 +434,29 @@ function getFastValue(a, op, b, type) {
     bNumber = aNumber;
     aNumber = cNumber;
   }
-  if (op === '*') {
-    // We can't eliminate where a or b are 0 as that would break things for creating
-    // a negative 0.
-    if ((aNumber === 0 || bNumber === 0) && !FLOAT_TYPES.has(type)) {
-      return '0';
-    } else if (aNumber === 1) {
-      return b;
-    } else if (bNumber === 1) {
-      return a;
-    } else if (bNumber !== null && type && isIntImplemented(type) && Runtime.getNativeTypeSize(type) <= 32) {
-      const shifts = Math.log(bNumber) / Math.LN2;
-      if (shifts % 1 === 0) {
-        return `(${a}<<${shifts})`;
-      }
-    }
-    if (!FLOAT_TYPES.has(type)) {
-      // if guaranteed small enough to not overflow into a double, do a normal multiply
-      // default is 32-bit multiply for things like getelementptr indexes
-      const bits = getBits(type) || 32;
-      // Note that we can emit simple multiple in non-asm.js mode, but asm.js
-      // will not parse "16-bit" multiple, so must do imul there
-      if ((aNumber !== null && Math.abs(a) < TWO_TWENTY) || (bNumber !== null && Math.abs(b) < TWO_TWENTY)) {
-        // keep a non-eliminatable coercion directly on this
-        return `(((${a})*(${b}))&${(Math.pow(2, bits) - 1) | 0})`;
-      }
-      return `(Math.imul(${a}, ${b})|0)`;
-    }
-  } else if (op === '/') {
-    // careful on floats, since 0*NaN is not 0
-    if (a === '0' && !FLOAT_TYPES.has(type)) {
-      return '0';
-    } else if (b === 1) {
-      return a;
-    } // Doing shifts for division is problematic, as getting the rounding right on negatives is tricky
-  } else if (op === '+' || op === '-') {
-    if (b[0] === '-') {
-      op = op === '+' ? '-' : '+';
-      b = b.substr(1);
-    }
-    if (aNumber === 0) {
-      return op === '+' ? b : `(-${b})`;
-    } else if (bNumber === 0) {
-      return a;
-    }
+
+  if (aNumber === 0) {
+    return b;
+  } else if (bNumber === 0) {
+    return a;
   }
+
+  if (b[0] === '-') {
+    op = '-'
+    b = b.substr(1);
+  }
+
   return `(${a})${op}(${b})`;
 }
 
 function calcFastOffset(ptr, pos) {
-  return getFastValue(ptr, '+', pos, 'i32');
+  return getFastValue(ptr, '+', pos);
+}
+
+function getBitWidth(type) {
+  if (type == 'i53' || type == 'u53')
+    return 53;
+  return getNativeTypeSize(type) * 8;
 }
 
 function getHeapForType(type) {
@@ -595,59 +496,26 @@ function makeReturn64(value) {
   return `(setTempRet0(${pair[1]}), ${pair[0]})`;
 }
 
-function makeThrow(what) {
+function makeThrow(excPtr) {
   if (ASSERTIONS && DISABLE_EXCEPTION_CATCHING) {
-    what += ' + " - Exception catching is disabled, this exception cannot be caught. Compile with -sNO_DISABLE_EXCEPTION_CATCHING or -sEXCEPTION_CATCHING_ALLOWED=[..] to catch."';
+    var assertInfo = 'Exception thrown, but exception catching is not enabled. Compile with -sNO_DISABLE_EXCEPTION_CATCHING or -sEXCEPTION_CATCHING_ALLOWED=[..] to catch.';
     if (MAIN_MODULE) {
-      what += ' + " (note: in dynamic linking, if a side module wants exceptions, the main module must be built with that support)"';
+      assertInfo += ' (note: in dynamic linking, if a side module wants exceptions, the main module must be built with that support)';
     }
+    return `assert(false, '${assertInfo}');`;
   }
-  return `throw ${what};`;
+  return `throw ${excPtr};`;
 }
 
-function stripCorrections(param) {
-  let m;
-  while (true) {
-    if (m = /^\((.*)\)$/.exec(param)) {
-      param = m[1];
-      continue;
-    }
-    if (m = /^\(([$_\w]+)\)&\d+$/.exec(param)) {
-      param = m[1];
-      continue;
-    }
-    if (m = /^\(([$_\w()]+)\)\|0$/.exec(param)) {
-      param = m[1];
-      continue;
-    }
-    if (m = /^\(([$_\w()]+)\)\>>>0$/.exec(param)) {
-      param = m[1];
-      continue;
-    }
-    if (m = /CHECK_OVERFLOW\(([^,)]*),.*/.exec(param)) {
-      param = m[1];
-      continue;
-    }
-    break;
-  }
-  return param;
+function storeException(varName, excPtr) {
+  var exceptionToStore = EXCEPTION_STACK_TRACES ? `new CppException(${excPtr})` : `${excPtr}`;
+  return `${varName} = ${exceptionToStore};`;
 }
 
 function charCode(char) {
   return char.charCodeAt(0);
 }
 
-function getTypeFromHeap(suffix) {
-  switch (suffix) {
-    case '8': return 'i8';
-    case '16': return 'i16';
-    case '32': return 'i32';
-    case 'F32': return 'float';
-    case 'F64': return 'double';
-  }
-  assert(false, 'bad type suffix: ' + suffix);
-}
-
 function ensureValidFFIType(type) {
   return type === 'float' ? 'double' : type; // ffi does not tolerate float XXX
 }
@@ -696,7 +564,7 @@ function makeDynCall(sig, funcPtr) {
 
 
   if (funcPtr === undefined) {
-    warn(`${currentlyParsedFilename}: \
+    warn(`
 Legacy use of {{{ makeDynCall("${sig}") }}}(funcPtr, arg1, arg2, ...). \
 Starting from Emscripten 2.0.2 (Aug 31st 2020), syntax for makeDynCall has changed. \
 New syntax is {{{ makeDynCall("${sig}", "funcPtr") }}}(arg1, arg2, ...). \
@@ -781,7 +649,7 @@ function addAtExit(code) {
 }
 
 function makeRetainedCompilerSettings() {
-  const ignore = new Set(['STRUCT_INFO']);
+  const ignore = new Set();
   if (STRICT) {
     for (const setting of LEGACY_SETTINGS) {
       ignore.add(setting);
@@ -836,11 +704,11 @@ function modifyFunction(text, func) {
 }
 
 function runIfMainThread(text) {
-  if (WASM_WORKERS && USE_PTHREADS) {
+  if (WASM_WORKERS && PTHREADS) {
     return 'if (!ENVIRONMENT_IS_WASM_WORKER && !ENVIRONMENT_IS_PTHREAD) { ' + text + ' }';
   } else if (WASM_WORKERS) {
     return 'if (!ENVIRONMENT_IS_WASM_WORKER) { ' + text + ' }';
-  } else if (USE_PTHREADS) {
+  } else if (PTHREADS) {
     return 'if (!ENVIRONMENT_IS_PTHREAD) { ' + text + ' }';
   } else {
     return text;
@@ -864,7 +732,7 @@ function makeRemovedModuleAPIAssert(moduleName, localName) {
 function checkReceiving(name) {
   // ALL_INCOMING_MODULE_JS_API contains all valid incoming module API symbols
   // so calling makeModuleReceive* with a symbol not in this list is an error
-  assert(ALL_INCOMING_MODULE_JS_API.includes(name));
+  assert(ALL_INCOMING_MODULE_JS_API.has(name));
 }
 
 // Make code to receive a value on the incoming Module object.
@@ -970,7 +838,7 @@ function hasExportedSymbol(sym) {
 // it is a BigInt. Otherwise, we legalize into pairs of i32s.
 function defineI64Param(name) {
   if (WASM_BIGINT) {
-    return `/** @type {!BigInt} */ ${name}`;
+    return name;
   }
   return `${name}_low, ${name}_high`;
 }
@@ -1047,19 +915,12 @@ function addReadyPromiseAssertions(promise) {
 });`;
 }
 
-function makeMalloc(source, param) {
-  if (hasExportedSymbol('malloc')) {
-    return `_malloc(${param})`;
-  }
-  // It should be impossible to call some functions without malloc being
-  // included, unless we have a deps_info.json bug. To let closure not error
-  // on `_malloc` not being present, they don't call malloc and instead abort
-  // with an error at runtime.
-  // TODO: A more comprehensive deps system could catch this at compile time.
-  if (!ASSERTIONS) {
-    return 'abort();';
-  }
-  return `abort('malloc was not included, but is needed in ${source}. Adding "_malloc" to EXPORTED_FUNCTIONS should fix that. This may be a bug in the compiler, please file an issue.');`;
+function asyncIf(condition) {
+  return condition ? 'async' : '';
+}
+
+function awaitIf(condition) {
+  return condition ? 'await' : '';
 }
 
 // Adds a call to runtimeKeepalivePush, if needed by the current build
@@ -1067,7 +928,7 @@ function makeMalloc(source, param) {
 // We skip this completely in MINIMAL_RUNTIME and also in builds that
 // don't ever need to exit the runtime.
 function runtimeKeepalivePush() {
-  if (MINIMAL_RUNTIME || (EXIT_RUNTIME == 0 && USE_PTHREADS == 0)) return '';
+  if (MINIMAL_RUNTIME || (EXIT_RUNTIME == 0 && PTHREADS == 0)) return '';
   return 'runtimeKeepalivePush();';
 }
 
@@ -1076,7 +937,7 @@ function runtimeKeepalivePush() {
 // We skip this completely in MINIMAL_RUNTIME and also in builds that
 // don't ever need to exit the runtime.
 function runtimeKeepalivePop() {
-  if (MINIMAL_RUNTIME || (EXIT_RUNTIME == 0 && USE_PTHREADS == 0)) return '';
+  if (MINIMAL_RUNTIME || (EXIT_RUNTIME == 0 && PTHREADS == 0)) return '';
   return 'runtimeKeepalivePop();';
 }
 
@@ -1122,7 +983,14 @@ function getEntryFunction() {
 function preJS() {
   let result = '';
   for (const fileName of PRE_JS_FILES) {
-    result += preprocess(fileName);
+    result += read(fileName);
   }
   return result;
 }
+
+function formattedMinNodeVersion() {
+  var major = MIN_NODE_VERSION / 10000
+  var minor = (MIN_NODE_VERSION / 100) % 100
+  var rev = MIN_NODE_VERSION % 100
+  return `v${major}.${minor}.${rev}`;
+}
