@@ -5,8 +5,10 @@
  * found in the LICENSE file.
  */
 
+#define _GNU_SOURCE
 #include <assert.h>
 #include <errno.h>
+#include <signal.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/mman.h>
@@ -17,10 +19,12 @@
 
 #include <emscripten.h>
 #include <emscripten/heap.h>
+#include <emscripten/console.h>
 #include <wasi/api.h>
 #include <wasi/wasi-helpers.h>
 
 #include "lock.h"
+#include "emscripten_internal.h"
 
 /*
  * WASI support code. These are compiled with the program, and call out
@@ -66,8 +70,13 @@ int clock_getres(clockid_t clk_id, struct timespec *tp) {
 
 // Mark these as weak so that wasmfs does not collide with it. That is, if
 // wasmfs is in use, we want to use that and not this.
-__attribute__((__weak__)) intptr_t _mmap_js(
-  size_t length, int prot, int flags, int fd, size_t offset, int* allocated) {
+__attribute__((__weak__)) int _mmap_js(size_t length,
+                                       int prot,
+                                       int flags,
+                                       int fd,
+                                       size_t offset,
+                                       int* allocated,
+                                       void** addr) {
   return -ENOSYS;
 }
 
@@ -115,8 +124,6 @@ int getentropy(void* buffer, size_t length) {
 
 // Emscripten additions
 
-extern void emscripten_notify_memory_growth(size_t memory_index);
-
 // Should never be called in standalone mode
 void emscripten_memcpy_big(void *restrict dest, const void *restrict src, size_t n) {
   __builtin_unreachable();
@@ -135,7 +142,10 @@ int emscripten_resize_heap(size_t size) {
   assert(old_size < size);
   ssize_t diff = (size - old_size + WASM_PAGE_SIZE - 1) / WASM_PAGE_SIZE;
   size_t result = __builtin_wasm_memory_grow(0, diff);
-  if (result != (size_t)-1) {
+  // Its seems v8 has a bug in memory.grow that causes it to return 
+  // (uint32_t)-1 even with memory64:
+  // https://bugs.chromium.org/p/v8/issues/detail?id=13948
+  if (result != (uint32_t)-1 && result != (size_t)-1) {
     // Success, update JS (see https://github.com/WebAssembly/WASI/issues/82)
     emscripten_notify_memory_growth(0);
     return 1;
@@ -145,29 +155,18 @@ int emscripten_resize_heap(size_t size) {
 }
 
 double emscripten_get_now(void) {
-  return (1000 * clock()) / (double)CLOCKS_PER_SEC;
+  return (1000ll * clock()) / (double)CLOCKS_PER_SEC;
 }
 
 // C++ ABI
 
-// Emscripten disables exception catching by default, but not throwing. That
-// allows users to see a clear error if a throw happens, and 99% of the
-// overhead is in the catching, so this is a reasonable tradeoff.
-// For now, in a standalone build just terminate. TODO nice error message
-//
-// Define these symbols as weak so that when we build with exceptions
-// enabled (using wasm-eh) we get the real versions of these functions
-// as defined in libc++abi.
-
-__attribute__((__weak__))
+#if EMSCRIPTEN_NOCATCH
+// When exception catching is disabled, we stub out calls to `__cxa_throw`.
+// Otherwise, `__cxa_throw` will be imported from the host.
 void __cxa_throw(void* ptr, void* type, void* destructor) {
   abort();
 }
-
-__attribute__((__weak__))
-void* __cxa_allocate_exception(size_t thrown_size) {
-  abort();
-}
+#endif
 
 // WasmFS integration. We stub out file preloading and such, that are not
 // expected to work anyhow.
@@ -180,7 +179,7 @@ int _wasmfs_get_preloaded_file_size(int index) { return 0; }
 
 int _wasmfs_get_preloaded_file_mode(int index) { return 0; }
 
-size_t _wasmfs_copy_preloaded_file_data(int index, void* buffer) { return 0; }
+void _wasmfs_copy_preloaded_file_data(int index, void* buffer) {}
 
 void _wasmfs_get_preloaded_parent_path(int index, void* buffer) {}
 
@@ -201,7 +200,7 @@ imported__wasi_fd_write(__wasi_fd_t fd,
                         __wasi_size_t* nwritten);
 
 // Write a buffer + a newline.
-static void wasi_writeln(__wasi_fd_t fd, char* buffer) {
+static void wasi_writeln(__wasi_fd_t fd, const char* buffer) {
   struct __wasi_ciovec_t iovs[2];
   iovs[0].buf = (uint8_t*)buffer;
   iovs[0].buf_len = strlen(buffer);
@@ -211,6 +210,20 @@ static void wasi_writeln(__wasi_fd_t fd, char* buffer) {
   imported__wasi_fd_write(fd, iovs, 2, &nwritten);
 }
 
-void _emscripten_out(char* text) { wasi_writeln(1, text); }
+void _emscripten_out(const char* text) { wasi_writeln(1, text); }
 
-void _emscripten_err(char* text) { wasi_writeln(2, text); }
+void _emscripten_err(const char* text) { wasi_writeln(2, text); }
+
+// In the non-standalone build we define this helper function in JS to avoid
+// signture mismatch issues.
+// See: https://github.com/emscripten-core/posixtestsuite/issues/6
+void __call_sighandler(sighandler_t handler, int sig) {
+  handler(sig);
+}
+
+int _setitimer_js(int which, double timeout) {
+  // There is no API to let us set timers in standalone mode atm. Return an
+  // error.
+  errno = ENOTSUP;
+  return -1;
+}
