@@ -48,7 +48,7 @@ void Directory::Handle::cacheChild(const std::string& name,
   // backend.
   if (kind == DCacheKind::Mount || !getDir()->maintainsFileIdentity()) {
     auto& dcache = getDir()->dcache;
-    auto [_, inserted] = dcache.insert({name, {kind, child}});
+    [[maybe_unused]] auto [_, inserted] = dcache.insert({name, {kind, child}});
     assert(inserted && "inserted child already existed!");
   }
   // Set the child's parent.
@@ -178,12 +178,13 @@ int Directory::Handle::insertMove(const std::string& name,
       it->second.file->locked().setParent(nullptr);
       it->second = entry;
     }
-    file->locked().setParent(getDir());
   } else {
     // This backend doesn't use the dcache.
     assert(getDir()->maintainsFileIdentity());
   }
 
+  file->locked().setParent(getDir());
+
   // TODO: Moving mount points probably shouldn't update the mtime.
   auto now = time(NULL);
   oldParent->locked().setMTime(now);
