@@ -73,6 +73,7 @@ import shutil
 import sys
 from subprocess import PIPE
 from textwrap import dedent
+from typing import List
 
 __scriptdir__ = os.path.dirname(os.path.abspath(__file__))
 __rootdir__ = os.path.dirname(__scriptdir__)
@@ -93,7 +94,7 @@ DDS_HEADER_SIZE = 128
 # to work around silly av false positives
 AV_WORKAROUND = 0
 
-excluded_patterns = []
+excluded_patterns: List[str] = []
 new_data_files = []
 
 
@@ -344,9 +345,9 @@ def generate_object_file(data_files):
     target = 'wasm64-unknown-emscripten'
   else:
     target = 'wasm32-unknown-emscripten'
-  shared.check_call([shared.LLVM_MC,
-                     '-filetype=obj',
-                     '-triple=' + target,
+  shared.check_call([shared.EMCC,
+                     '-c',
+                     '--target=' + target,
                      '-o', options.obj_output,
                      asm_file])
 
@@ -409,8 +410,7 @@ def main():
       options.from_emcc = True
       leading = ''
     elif arg.startswith('--plugin'):
-      with open(arg.split('=', 1)[1]) as f:
-        plugin = f.read()
+      plugin = utils.read_file(arg.split('=', 1)[1])
       eval(plugin) # should append itself to plugins
       leading = ''
     elif leading == 'preload' or leading == 'embed':
@@ -474,11 +474,11 @@ def main():
   # even if we cd'd into a symbolic link.
   curr_abspath = os.path.abspath(os.getcwd())
 
-  if not file_.explicit_dst_path:
-    for file_ in data_files:
+  for file_ in data_files:
+    if not file_.explicit_dst_path:
       # This file was not defined with src@dst, so we inferred the destination
-      # from the source. In that case, we require that the destination not be
-      # under the current location
+      # from the source. In that case, we require that the destination be
+      # within the current working directory.
       path = file_.dstpath
       # Use os.path.realpath to resolve any symbolic links to hard paths,
       # to match the structure in curr_abspath.
@@ -486,18 +486,19 @@ def main():
       if DEBUG:
         err(path, abspath, curr_abspath)
       if not abspath.startswith(curr_abspath):
-        err('Error: Embedding "%s" which is below the current directory '
-            '"%s". This is invalid since the current directory becomes the '
-            'root that the generated code will see' % (path, curr_abspath))
+        err('Error: Embedding "%s" which is not contained within the current directory '
+            '"%s".  This is invalid since the current directory becomes the '
+            'root that the generated code will see.  To include files outside of the current '
+            'working directoty you can use the `--preload-file srcpath@dstpath` syntax to '
+            'explicitly specify the target location.' % (path, curr_abspath))
         sys.exit(1)
       file_.dstpath = abspath[len(curr_abspath) + 1:]
       if os.path.isabs(path):
         err('Warning: Embedding an absolute file/directory name "%s" to the '
             'virtual filesystem. The file will be made available in the '
-            'relative path "%s". You can use the explicit syntax '
-            '--preload-file srcpath@dstpath to explicitly specify the target '
-            'location the absolute source path should be directed to.'
-            % (path, file_.dstpath))
+            'relative path "%s". You can use the `--preload-file srcpath@dstpath` '
+            'syntax to explicitly specify the target location the absolute source '
+            'path should be directed to.' % (path, file_.dstpath))
 
   for file_ in data_files:
     # name in the filesystem, native and emulated
@@ -520,6 +521,10 @@ def main():
     seen.add(name)
     return False
 
+  # The files are sorted by the dstpath to make the order of files reproducible
+  # across file systems / operating systems (os.walk does not produce the same
+  # file order on different file systems / operating systems)
+  data_files = sorted(data_files, key=lambda file_: file_.dstpath)
   data_files = [file_ for file_ in data_files if not was_seen(file_.dstpath)]
 
   if AV_WORKAROUND:
@@ -550,17 +555,13 @@ def main():
       # differs from the current generated one, otherwise leave the file
       # untouched preserving its old timestamp
       if os.path.isfile(options.jsoutput):
-        with open(options.jsoutput) as f:
-          old = f.read()
+        old = utils.read_file(options.jsoutput)
         if old != ret:
-          with open(options.jsoutput, 'w') as f:
-            f.write(ret)
+          utils.write_file(options.jsoutput, ret)
       else:
-        with open(options.jsoutput, 'w') as f:
-          f.write(ret)
+        utils.write_file(options.jsoutput, ret)
       if options.separate_metadata:
-        with open(options.jsoutput + '.metadata', 'w') as f:
-          json.dump(metadata, f, separators=(',', ':'))
+        utils.write_file(options.jsoutput + '.metadata', json.dumps(metadata, separators=(',', ':')))
 
   return 0
 
@@ -581,9 +582,8 @@ def generate_js(data_target, data_files, metadata):
 
   Module.expectedDataFileDownloads++;
   (function() {
-    // When running as a pthread, FS operations are proxied to the main thread, so we don't need to
-    // fetch the .data bundle on the worker
-    if (Module['ENVIRONMENT_IS_PTHREAD']) return;
+    // Do not attempt to redownload the virtual filesystem data when in a pthread or a Wasm Worker context.
+    if (Module['ENVIRONMENT_IS_PTHREAD'] || Module['$ww']) return;
     var loadPackage = function(metadata) {\n'''
 
   code = '''
@@ -612,8 +612,7 @@ def generate_js(data_target, data_files, metadata):
     with open(data_target, 'wb') as data:
       for file_ in data_files:
         file_.data_start = start
-        with open(file_.srcpath, 'rb') as f:
-          curr = f.read()
+        curr = utils.read_binary(file_.srcpath)
         file_.data_end = start + len(curr)
         if AV_WORKAROUND:
             curr += '\x00'
@@ -630,11 +629,7 @@ def generate_js(data_target, data_files, metadata):
           Module['FS_createPreloadedFile'](this.name, null, byteArray, true, true, function() {
             Module['removeRunDependency']('fp ' + that.name);
           }, function() {
-            if (that.audio) {
-              Module['removeRunDependency']('fp ' + that.name); // workaround for chromium bug 124926 (still no audio with this, but at least we don't hang)
-            } else {
-              err('Preloading file ' + that.name + ' failed');
-            }
+            err('Preloading file ' + that.name + ' failed');
           }, false, true); // canOwn this data in the filesystem, it is a slide into the heap that will never change\n'''
     create_data = '''// canOwn this data in the filesystem, it is a slide into the heap that will never change
           Module['FS_createDataFile'](this.name, null, byteArray, true, true, true);
@@ -991,7 +986,7 @@ def generate_js(data_target, data_files, metadata):
     code += '''
       function processPackageData(arrayBuffer) {
         assert(arrayBuffer, 'Loading data file failed.');
-        assert(arrayBuffer instanceof ArrayBuffer, 'bad input to processPackageData');
+        assert(arrayBuffer.constructor.name === ArrayBuffer.name, 'bad input to processPackageData');
         var byteArray = new Uint8Array(arrayBuffer);
         var curr;
         %s
